#include "ControlAlgorithm.h"
#include "sfhMath.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ControlAlgorithm Constructor
//==================================================================================================
///		This constructor performs all initial setup for a Control algorithm simobject.
///		Reading in parameters, setting up communication interface i.e. output ports,
///		input ports, and states, plus additional 'one time only' resource setup.
///	
/// \param [in]		simObjectName	The name of the simobject. Used primarily 
///									by superclass constructor
///
/// \param [in]		creator			Retrive parameters. Register states, ports
///									and shared resources
//
////////////////////////////////////////////////////////////////////////////////////////////////////
ControlAlgorithm::ControlAlgorithm(const string& simObjectName, ISimObjectCreator* const creator) : SimObject(simObjectName){
	creator->AddInport("VesselPosition",6,&inputEta);															// Input port, position
	creator->AddInport("VesselVelocityBODY",6,&inputNu);														// Input port, velocity

	creator->AddInport("VesselPositionDesired",6,&inputEtaDesired);												// Input port, position
	creator->AddInport("VesselVelocityBODYDesired",6,&inputNuDesired);											// Input port, velocity
	creator->AddInport("VesselAccelerationDesired",6,&inputNuDotDesired);										// Input port, velocity


	creator->AddOutport("ForcesAndMomentsBODY",6,PORT_FUNCTION(ControlAlgorithm::ForcesAndMomentsBODY));	// Output port, Forces and moments in body coordinates
	creator->AddOutport("Error",6,PORT_FUNCTION(ControlAlgorithm::Error));								// Output port, Error
	creator->GetDoubleParamArray("TauSaturationLimitMax",tauLimitMax,6);
	creator->GetDoubleParamArray("TauSaturationLimitMin",tauLimitMin,6);
	double ones[] = {1,1,1,1,1,1};
	creator->GetDoubleParamArray("ControllableStates",controllableStates,6,ones);
}

void ControlAlgorithm::OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep){

}
//double ControlAlgorithm::Reminder(double a, double b){
//
// int numberOfB = static_cast<int>( a / b );
// double result = a - (numberOfB* b);
// return result;
// }

//int ControlAlgorithm::sign(double value){
//	int sign;
//	if (value>0)
//		sign = 1;
//	else if (value < 0)
//		sign = -1;
//	else 
//		sign = 0;
//	
//	return sign;
//}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		RAD3PIPI
//==================================================================================================
///		
///		Converts and angle given the range \f$-\infty\f$- \f$\infty\f$ to \f$-\pi\f$-\f$\pi\f$. 
///	
/// \param [in]		angleIn			Angle in radians to convert
/// \return							angle
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
double ControlAlgorithm::RAD2PIPI(double angleIn){
	double argument = angleIn + sfh::Sign(angleIn)*M_PI;
	double temp = fmod(argument,2*M_PI) - sfh::Sign(angleIn)*M_PI;
	
	
	//double temp = fmod(angleIn,2*M_PI);
	//if (abs(temp) > M_PI){
	//	//double angle = Reminder(angleIn,2*M_PI);
	//	int signAngle = sign(temp);
	//	temp = -signAngle*(2*M_PI - signAngle*temp);
	//}
	//double r = Modulus(angleIn+sign(angleIn)*M_PI,2*M_PI);
	//double s = sign(sign(angleIn)+2*(sign(abs(Modulus(angleIn+M_PI,2*M_PI)/(2*M_PI)))-1));
	//double result = r-s*M_PI;
	return temp;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ShortestError
//==================================================================================================
///		
///		Method for calculating the shortest rotation error if the angles are defined in the range \f$-\pi\f$-\f$\pi\f$.
///	
/// \param [in]		angle			angle in radians
/// \param [in]		desiredAngle	Angle in radians
///
////////////////////////////////////////////////////////////////////////////////////////////////////
double ControlAlgorithm::ShortestError(double angle, double desiredAngle){

	if (sfh::Sign(angle) != sfh::Sign(desiredAngle))
		if (angle <0)
			angle = angle + (2*M_PI);
		else
			desiredAngle = desiredAngle+(2*M_PI);
	return RAD2PIPI(angle-desiredAngle);
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		RMatrix
//==================================================================================================
///		
///		Calculates the rotation matrix in 3DOF using Euler angles.
///	
/// \param [in]		eta			Array containing the 6DOF position.
///
/// \return						Rotation matrix as a 3\f$\times\f$3 Eigen Matrix
///
////////////////////////////////////////////////////////////////////////////////////////////////////
Eigen::Matrix<double,3,3> ControlAlgorithm::RMatrix(const double* const eta){
	Matrix3d rotationMatrix;
	
	double cPhi = cos(eta[3]);
	double sPhi = sin(eta[3]);
	
	double cTheta = cos(eta[4]);
	double sTheta = sin(eta[4]);
	
	double cPsi = cos(eta[5]);
	double sPsi = sin(eta[5]);

	rotationMatrix.topRows(1) << cPsi*cTheta, -sPsi*cPhi + cPsi*sTheta*sPhi, sPsi*sPhi + cPsi*cPhi*sTheta;
	rotationMatrix.middleRows<1>(1) << sPsi*cTheta, cPsi*cPhi + sPhi*sTheta*sPsi, -cPsi*sPhi + sTheta*sPsi*cPhi;
	rotationMatrix.bottomRows(1) << -sTheta, cTheta*sPhi, cTheta*cPhi;
	
	return rotationMatrix;
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		TMatrix
//==================================================================================================
///		
///		Calculates the transformation matrix in 3DOF using Euler angles.
///	
/// \param [in]		eta			Array containing the 6DOF position.
///
/// \return			Transformation matrix as a 3\f$\times\f$3 Eigen Matrix
///
////////////////////////////////////////////////////////////////////////////////////////////////////
Eigen::Matrix<double,3,3> ControlAlgorithm::TMatrix(const double* const eta){
	Eigen::Matrix3d transformationMatrix;

	transformationMatrix.row(0) << 1,sin(eta[3])*tan(eta[4]),cos(eta[3])*tan(eta[4]);
	transformationMatrix.row(1) << 0, cos(eta[3]), -sin(eta[3]);
	transformationMatrix.row(2) << 0, sin(eta[3])/cos(eta[4]), cos(eta[3])/cos(eta[4]);
	return transformationMatrix;
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		JMatrix
//==================================================================================================
///		
///		Calculates the rotation matrix in 6DOF using Euler angles.
///	
/// \param [in]		eta			Array containing the 6DOF position.
///
/// \return			Rotation matrix as a 6\f$\times\f$6  Eigen Matrix.
///
////////////////////////////////////////////////////////////////////////////////////////////////////
Eigen::Matrix<double,6,6> ControlAlgorithm::JMatrix(const double* const eta){

	Eigen::Matrix<double,3,3> R = RMatrix(eta);
	Eigen::Matrix<double,3,3> T = TMatrix(eta);

	Eigen::Matrix<double,6,6> J = Eigen::Matrix<double,6,6>::Zero();

	J.topLeftCorner(3,3) = R;
	J.bottomRightCorner(3,3) = T;

	return J;

}