/**
\class ControlAlgorithm
\par User Documentation:

Superclass for motion controllers
\par Parameters
<table >
<tr><td><b>Parameter<td>Size<td>Comment</b>
<tr><td>TauSaturationLimitMax<td>6<td>Maximum available force in all 6DOF
<tr><td>TauSaturationLimitMin<td>6<td>Minimum available force in all 6DOF
<tr><td>ControllableStates<td>6<td>1 or 0, indicating if the state is controllable
</table>

\par Input ports
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>VesselPosition<td>6<td>Position and attitude of the vessel.
<tr><td>VesselVelocityBODY<td>6<td>Velocity of the vessel in BODY-frame
<tr><td>VesselPositionDesired<td>6<td>Desired position vessel
<tr><td>VesselVelocityBODYDesired<td>6<td>Desired velocity of the vessel in BODY-frame
<tr><td>VesselAccelerationDesired<td>6<td>Desired acceleration of the vessel in BODY-frame
</table>

\par Output ports
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>ForcesAndMomentsBODY<td>6<td>Forces and moments in body frame
<tr><td>Error<td>6<td>Position and attitude error
</table>

*/
#ifndef CONTROLALGORITHM
#define CONTROLALGORITHM

#include "SimObject.h"
#include <string>
#include <Eigen/Eigen>
#include <math.h>
using namespace Eigen;
/**
\brief	Motion controller superclass

\author	Mats Nvik Hval

Superclass for motion control system.es

\par Revision history:
	06.06.2012 MNH: Initial version
*/
class ControlAlgorithm : public SimObject
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW ///< enables the use of eigen types as members variables
	ControlAlgorithm(const string& simObjectName, ISimObjectCreator* const creator);							///< Reads parameters, registers states, input/output ports and shared resources
	virtual void OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep);///< Returns object derivatives as a function of time, states and input ports
	virtual const double* const ForcesAndMomentsBODY(const double T, const double* const X) = 0;				///< Returns the 6DOF forces and moments calculated by the controller
	virtual const double* const Error(const double T, const double* const X) = 0;								///< Returns the error in the position in 6DOF
		
#ifdef FH_VISUALIZATION
	/// The rendering is not used.
	virtual void RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator){}

	/// The rendering is not used.
	virtual void RenderUpdate(const double T, const double* const X){}

#endif		

protected:
	double ShortestError(double eta, double etaRef);			///< Calculate shortest angle error in range \f$-\pi \f$-\f$\pi\f$.
	double RAD2PIPI(double angleIn);							///< Converts angle given the range \f$- \infty \f$- \f$ \infty \f$ to \f$-\pi\f$-\f$\pi\f$.
	//double Reminder(double value1,double value2);
	//int sign(double value);
	Eigen::Matrix<double,3,3> RMatrix(const double* const eta);	///< Calculates the 3DOF rotation matrix.
	Eigen::Matrix<double,3,3> TMatrix(const double* const eta);	///< Calculates the 3DOF transformation matrix
	Eigen::Matrix<double,6,6> JMatrix(const double* const eta);	///< Calculates the 6DOF transformation matrix
	
	ISignalPort* inputEta;						///< Input port for the vessel NED position
	ISignalPort* inputEtaDesired;				///< Input port for the desired NED position
	ISignalPort* inputNu;						///< Input port for the vessel BODY velocity
	ISignalPort* inputNuDesired;				///< Input port for the desired BODY velocity
	ISignalPort* inputNuDotDesired;				///< Input port for the desired BODY acceleration
	
	double tauLimitMin[6];						///< Saturation-limits for positive thrust
	double tauLimitMax[6];						///< Saturation-limits for negative thrust
	double controllableStates[6];				///< Array holding a flag for the controllable states
private:
};
#endif

