#include "GuidanceSystem.h"
#include "HelpFunctions.h"
#include <Eigen/StdVector>
#include <Eigen/Dense>
#include <sfhMath.h>
#include <sfhGeneral.h>

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		GuidanceSystem constructor
//==================================================================================================
///		This constructor performs all initial setup for a guidance system simobject.
///		Reading in parameters, setting up communication interface i.e. output ports,
///		input ports, and states, plus additional 'one time only' resource setup.
///	
/// \param [in]		simObjectName	The name of the simobject. Used primarily 
///									by superclass constructor
///
/// \param [in]		creator			Retrive parameters. Register states, ports
///									and shared resources
///
////////////////////////////////////////////////////////////////////////////////////////////////////
GuidanceSystem::GuidanceSystem(const string& simObjectName, ISimObjectCreator* const creator) : SimObject(simObjectName)
{
	creator->AddInport	("PositionIn",	6,	&etaInputPort);
	creator->AddInport	("VelocityIn",	6,	&nuInputPort);
	
	creator->AddOutport("NuDotDesired",6,PORT_FUNCTION(GuidanceSystem::NuDotDesired));
	creator->AddOutport("NuDesired",6,PORT_FUNCTION(GuidanceSystem::NuDesired));
	creator->AddOutport("EtaDesired",6,PORT_FUNCTION(GuidanceSystem::EtaDesired));
	creator->AddOutport("EtaReference",6,PORT_FUNCTION(GuidanceSystem::EtaReference));
	
	std::string referenceGeneratorType;
	creator->GetStringParam("GuidanceGenerator",referenceGeneratorType,"FilterBased");
	filterBased = false;
	if (referenceGeneratorType.compare("FilterBased") == 0 || referenceGeneratorType.compare("filterBased")  == 0|| referenceGeneratorType.compare("FILTERBASED")  == 0){
		filterBased = true;
		string saturatedValues = "No";
		creator->GetStringParam("Saturated",saturatedValues,saturatedValues);
		if (saturatedValues.compare("Yes")  == 0|| saturatedValues.compare("yes") == 0 || saturatedValues.compare("YES") == 0){}
			saturated = true;
			creator->GetDoubleParamArray("MaxVelocities",maxVelocities,6);
			creator->GetDoubleParamArray("SafetyMarginsOnMaxVelocities",safetyMarginsVel,6);
			creator->GetDoubleParamArray("MaxAcceleration",maxAcceleration,6);
			creator->GetDoubleParamArray("SafetyMarginsOnMaxAccelerations",safetyMarginsAcc,6);
		
		double delta[6];
		double omega[6];
		creator->GetDoubleParamArray("RelativeDampingRatiosFilter",delta,6);
		creator->GetDoubleParamArray("OmegaFilter",omega,6);
		twoDeltaPlusIdentity = Eigen::Matrix<double,6,6>::Zero();
		OMEGA = Eigen::Matrix<double,6,6>::Zero();
		for (size_t i = 0 ; i<6 ; i++){
			twoDeltaPlusIdentity(i,i)  = 2*delta[i] +1;
			OMEGA(i,i) = omega[i];
		}
	}

	double zero[] = {0,0,0,0,0,0};
	
	//indexJerk =			creator->AddState("Jerk",6,zero);
	indexEtaDDot	=	creator->AddState("EtaDDot",6,zero);
	indexEtaDot		=	creator->AddState("EtaDot",6,zero);
	indexEta		=	creator->AddState("Eta",6);

}
void GuidanceSystem::OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep){
	/// implements a third order filter based reference model.
	if (filterBased){
		const double* a_eta = etaInputPort->GetPortValue(T,X);
		//Eigen::Matrix<double,6,6> OMEGA = Eigen::Matrix<double,6,6>::Zero();
		Eigen::Matrix<double,6,1> etaDesired, etaDotDesired, etaDDotDesired;
		Eigen::Matrix<double,6,1> etaD_temp, etaDotD_temp, etaDDotD_temp;
		Eigen::Matrix<double,6,1> jerk = Eigen::Matrix<double,6,1>::Zero();
		Eigen::Matrix<double,6,1> eta_ref = Eigen::Matrix<double,6,1>::Zero();
		
		eta_ref << etaReference[0],etaReference[1],etaReference[2],etaReference[3],etaReference[4],etaReference[5];
		
		/*if (saturated){
			for (size_t i = 0;i<6;i++){
				double m = etaReference[i] = startEta[i];
				double mag = abs(m);
				double time_t = 2*mag/(maxVelocities[i]*safetyMarginsVel[i]);
				if (OMEGA(i,i) >= (2*M_PI)/(time_t) && time_t !=0)
					OMEGA(i,i) = (2*M_PI/time_t);
				else 
					OMEGA(i,i) = 1;
				if (m<0)
					OMEGA(i,i) = OMEGA(i,i)/2;
			}
		}*/
			
		for (size_t i = 0;i<6;i++){
			etaDesired(i) = X[indexEta+i];
			etaDotDesired(i) = X[indexEtaDot+i];
			etaDDotDesired(i) = X[indexEtaDDot+i];
			//if (operationFinished)
				//etaDotDesired(i) = 0;
		}
		Eigen::Matrix<double,6,6> omega3 = OMEGA*OMEGA*OMEGA;
		Eigen::Matrix<double,6,6> matEtaDD = twoDeltaPlusIdentity*OMEGA;
		Eigen::Matrix<double,6,6> matEtaD = twoDeltaPlusIdentity*OMEGA*OMEGA;
		//etaDotDesired = SaturateVector(etaDotDesired,maxVelocities);
		//etaDDotDesired = SaturateVector(etaDDotDesired,maxAcceleration);
		//Eigen::Matrix<double,6,1> etaDDTerm = matEtaDD*etaDDotDesired;
		//Eigen::Matrix<double,6,1> etaDTerm = matEtaD*etaDotDesired;
		Eigen::Matrix<double,6,6> Jm = Jmatrix(a_eta);
		Eigen::Matrix<double,6,6> Jinv = Jm.inverse();
		Eigen::Matrix<double,6,1> nu = Jinv*etaDotDesired;
		Eigen::Matrix<double,6,1> nuDot = Jinv*etaDDotDesired;

		Eigen::Matrix<double,6,1> nuSat = SaturateVector(nu,maxVelocities);
		Eigen::Matrix<double,6,1> nuDotSat = SaturateVector(nuDot,maxAcceleration);

		Eigen::Matrix<double,6,1> etaDotSat = Jm*nuSat;
		Eigen::Matrix<double,6,1> etaDDotSat = Jm*nuDotSat;

		jerk = omega3*eta_ref - matEtaDD*etaDDotDesired - matEtaD*etaDotDesired - (omega3)*etaDesired;
		
		
		for(size_t i =0;i<6;i++){
			XDot[indexEtaDDot+i] = jerk(i);
			XDot[indexEtaDot+i] = etaDDotDesired(i);
			XDot[indexEta+i] = etaDotDesired(i);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		Saturate
//==================================================================================================
//
///	Saturates the input value
///	
/// \param [in]		value				value to saturate
///
/// \param [in]		min					minimum value
///
/// \param [in]		max					maximum  value
///
/// \return								saturated value
///
////////////////////////////////////////////////////////////////////////////////////////////////////
Eigen::Matrix<double,6,1> GuidanceSystem::SaturateVector(Eigen::Matrix<double,6,1> &inputVector,double max[6]){

	Eigen::Matrix<double,6,1> saturated = Eigen::Matrix<double,6,1>::Zero();

	for (size_t i = 0;i<6;i++){
		if (abs(inputVector(i)) > max[i])
			saturated(i) = sign(inputVector(i))*max[i];
		else
			saturated(i) = inputVector(i);
	}

	return saturated;

}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		EtaDesired
//==================================================================================================
//
///	Returns the desired eta
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					current state
///
/// \return								array of desired eta
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const GuidanceSystem::EtaDesired(const double T, const double* const X){
	const double *eta = etaInputPort->GetPortValue(T,X);
	for (size_t i =0;i<6;i++){
		a_etaDesired[i] = X[indexEta+i];
	}
	a_etaDesired[5] = RAD2PIPI(a_etaDesired[5]);
	return a_etaDesired;

}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		NuDesired
//==================================================================================================
//
///	Returns the desired eta
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					current state
///
/// \return								array of desired nu
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const GuidanceSystem::NuDesired(const double T, const double* const X){
	const double *eta_ = etaInputPort->GetPortValue(T,X);
	Eigen::Matrix<double,6,6> Jmat= Jmatrix(eta_).inverse();
	Eigen::Matrix<double,6,1> etaDott;
	etaDott << X[indexEtaDot],X[indexEtaDot+1],X[indexEtaDot+2],X[indexEtaDot+3],X[indexEtaDot+4],X[indexEtaDot+5];
	Eigen::Matrix<double,6,1> nuD  = Jmat*etaDott;
		
	for (size_t i =0;i<6;i++){
		a_nuDesired[i] = nuD(i);
	}
	return a_nuDesired;

}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		NuDotDesired
//==================================================================================================
//
///	Returns the desired eta
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					current state
///
/// \return								array of desired nuDot
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const GuidanceSystem::NuDotDesired(const double T, const double* const X){
	const double *eta_ = etaInputPort->GetPortValue(T,X);
	Eigen::Matrix<double,6,6> Jmat= Jmatrix(eta_).inverse();
	Eigen::Matrix<double,6,1> etaDDott;
	etaDDott << X[indexEtaDDot],X[indexEtaDDot+1],X[indexEtaDDot+2],X[indexEtaDDot+3],X[indexEtaDDot+4],X[indexEtaDDot+5];
	Eigen::Matrix<double,6,1> nuDD  = Jmat*etaDDott;
	
	for (size_t i =0;i<6;i++){
		a_nuDotDesired[i] = nuDD(i);//X[indexEtaDDot+i];
	}
	return a_nuDotDesired;
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		EtaReference
//==================================================================================================
//
///	Returns the desired eta
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					current state
///
/// \return								array of reference eta
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const GuidanceSystem::EtaReference(const double T, const double* const X){
	return etaReference;
}