////////////////////////////////////////////////////////////////////////////////////////////////////
/** \class							GuidanceSystem
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// \brief			Superclass of guidance systems for imput to the motion control system
Simulation model of a mass point in three degrees of freedom.

\par Parameters
<table >
<tr><td><b>Parameter<td>Default<td>Comment</b>
<tr><td>Mass<td>-<td>The mass of the mass object.
<tr><td>Material<td>"DiffuseSpecular"<td>The visualization material of the mass, specified by its name.
<tr><td>Radius<td>-<td>Visual radius of sphere
<tr><td>Color<td>-<td>Primary color of sphere
</table>

\par Input ports
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>PositionIn<td>3<td>The current position of the vessel in NED-frame
<tr><td>VelocityIn<td>3<td>The current velocity of the vessel in the BODY frame
</table>

\par Output ports
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>EtaDesired<td>3<td>The desired position and attitude in NED
<tr><td>NuDesired<td>3<td>The desired velocity in BODY
<tr><td>NuDotDesired<td>3<td>The desired acceleration in BODY
<tr><td>EtaReference<td>3<td>The reference point in NED
</table>

\par States
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>Eta<td>3<td>The desired position.
<tr><td>EtaDot<td>3<td>The desired velocity.
<tr><td>EtadDot<td>3<td>The desired acceleration.
</table>
 \author			Mats N{\aa}vik Hval
 \par Revision history:
					07.05.2012 MNH: Initial version.
*/
////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef GUIDANCESYSTEM_H
#define GUIDANCESYSTEM_H

#include "SimObject.h"
#include <string>
#include <Eigen/StdVector>

class GuidanceSystem : public SimObject
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW; ///< enables the use of eigen types as member variables
	GuidanceSystem(const string& simObjectName, ISimObjectCreator* const creator); ///< reads parameters, registers states, input/output ports and shared resources
	void OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep); ///< returns object derivatives as a function of time, states and input ports
	const double* const EtaDesired(const double T, const double* const X);						///< method for returning desired eta (6DOF)
	const double* const NuDesired(const double T, const double* const X);						///< method for returning desired nu (6DOF)
	const double* const NuDotDesired(const double T, const double* const X);					///< method for returning desired nuDot	(6DOF)
	const double* const EtaReference(const double T, const double* const X);					///< method for returning the reference points (6DOF)

#ifdef FH_VISUALIZATION
	virtual void	RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator){} ///< sets up visualization resources
	virtual void	RenderUpdate(const double T, const double* const X){} ///< updates visualization resources to reflect current state
#endif
protected:
	ISignalPort* etaInputPort;						///< input port. Current position
	ISignalPort* nuInputPort;						///< input port. Current linear velocity
	double eta[6];									///< Position and attitude for the vessel at current time
	double etaReference[6];							///< Input reference position and attitude
	double startEta[6];								///< array holding eta at the time a new setpoint is set
///	bool operationFinished;		

private:
	Eigen::Matrix<double,6,1> SaturateVector(Eigen::Matrix<double,6,1> &inputVector,double max[6]); ///< method for saturating a vector.
	

	double a_nuDotDesired[6];						///< return array for desired acceleration in BODY
	double a_etaDesired[6];							///< return array for desired position in NED
	double a_nuDesired[6];							///< return array for desired nu in BODY

	
	//----------------------
	//
	//	Variables specifically for the filter approach
	//
	//----------------------
	bool filterBased;								///< flag for filterbased reference model
	bool saturated;									///< flag telling if the values are to be saturated
	double maxVelocities[6];						///< maximum velocities of the vessel in each dof
	double maxAcceleration[6];						///< maximum acceleration of the vessel in each dof
	double safetyMarginsVel[6];						///< safety margins in the velocity in case of saturation
	double safetyMarginsAcc[6];						///< safety margins in the acceleration in case of saturation
	Eigen::Matrix<double,6,6> twoDeltaPlusIdentity;	
	Eigen::Matrix<double,6,6> OMEGA;				///< 6DOF matrix holding the natural frequency of the movement for the referencemodel
	
	int indexEtaDDot;								///< index of the desired acceleration state in the state vector
	int indexEtaDot;								///< index of the desired velocity state in the state vector
	int indexEta;									///< index of the desired position state in the state vector

};

#endif
