#include <assert.h>
#include "NonlinearPIDcontroller.h"

NonlinearPIDController::NonlinearPIDController(const string& simObjectName, ISimObjectCreator* const creator) : PIDController(simObjectName,creator)
{
	double satCoeffs[]= {0.8,0.8,0.8,0.8,0.8,0.8};
	creator->GetDoubleParamArray("SaturationFactor_TauIntegral",integralSaturationCoeffs,6,satCoeffs);
	creator->GetDoubleParamArray("RigidBodyandAddedMassMatrixDiagonal",massDiagonal,6);
	creator->GetDoubleParamArray("LinearDampingMatrixDiagonal",dampingLinearDiagonal,6);
	creator->GetDoubleParamArray("NonlinearDampingMatrixDiagonal",dampingNonlinearDiagonal,6);
	creator->GetDoubleParamArray("IntegralLimits",integralLimits,6);
	creator->GetDoubleParamArray("AntiWindupConstants",antiWindupConstants,6);
	creator->GetDoubleParamArray("TauSaturationLimitMin",tauSatMin,6);
	creator->GetDoubleParamArray("TauSaturationLimitMax",tauSatMax,6);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ForcesAndMomentsBODY
//==================================================================================================
//
///	Returns the forces and moments desired from the controller
/// Only the yaw-torque is implemented.
///	
/// \param [in]		T						current simulation time
///
/// \param [in]		X						current simulation state
///
/// \return									desired forces and momemnts
///
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////

const double* const NonlinearPIDController::ForcesAndMomentsBODY(const double T, const double* const X){
	CalculateTauP();
	CalculateTauI(T,X);
	CalculateTauD();
	CalculateTauFF(T,X);
	const double *eta = inputEta->GetPortValue(T,X);
	
	Eigen::Matrix<double,6,1> forcesNED;
	for (size_t i =0;i<6;i++)
		forcesNED(i) = a_tauP[i] + a_tauI[i] + a_tauD[i];
	force = -ControlAlgorithm::JMatrix(eta).inverse() * (forcesNED);
	for (size_t i =0;i<6;i++){
		force(i) = force(i) + a_tauFF[i];
		forcesAndMomentsBODY[i] = Saturate(force(i),tauSatMin[i],tauSatMax[i]);
	}
	/*for (size_t i =0;i<6;i++)
		if (force(i) >= 0){
			if( force(i) > tauLimitMax[i])
				forcesAndMomentsBODY[i] = tauLimitMax[i];
			else
				forcesAndMomentsBODY[i] = force(i);
		}
		else{
			if( force(i) < tauLimitMin[i])
				forcesAndMomentsBODY[i] = tauLimitMin[i];
			else
				forcesAndMomentsBODY[i] = force(i);
		}
	*/
	return forcesAndMomentsBODY; 
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		OdeFnc
//==================================================================================================
//
///	Calculates the error in position and velocity, and assignes the position error to the state vector.
/// The error is defined as value - desired_value.
///	
/// \param [in]		T					current simulation time
///
/// \param [in]		X					current simulation state
///
/// \param [in]		XDot				pointer to the integrand
///
/// \param [in]		bIsMajorTimeStep	is this a major time step
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
void NonlinearPIDController::OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep){
	const double *a_eta = inputEta->GetPortValue(T,X);
	const double *a_etaDesired = inputEtaDesired->GetPortValue(T,X);
	const double *a_nu = inputNu->GetPortValue(T,X);
	const double *a_nuDesired = inputNuDesired->GetPortValue(T,X);

	for (size_t i=0;i<6;i++){
		error(i) = a_eta[i] - a_etaDesired[i];
		errorDot(i) = a_nu[i] - a_nuDesired[i];
		errorIntegral(i) = X[indexErrorIntegral+i];	
	}

	for (size_t i = 0;i<3;i++){
		a_error[i] = error(i);
		a_error[i+3] = ShortestError(a_eta[i+3], a_etaDesired[i+3]);
		error(i+3) = a_error[i+3];
	}
	
	for(size_t i = 0 ; i < 6 ; i++){
		if (abs(a_error[i]) < integralLimits[i]){
			XDot[indexErrorIntegral+i] = a_error[i];
		}
		else{
		//	a_error[i] = 0;
			XDot[indexErrorIntegral+i] = 0;
		}
	}
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		TauFF
//==================================================================================================
//
///	Calculates the feed-forward term based on the model parameters
///	
/// \param [in]		T					current simulation time
///
/// \param [in]		X					current simulation state
///
////////////////////////////////////////////////////////////////////////////////////////////////////
void NonlinearPIDController::CalculateTauFF(const double T,const double* const X){

	const double* accDesired = inputNuDotDesired->GetPortValue(T,X);
	const double* velDesired = inputNuDesired->GetPortValue(T,X);
	for (size_t i = 0;i<6; i++){
		a_tauFF[i] =controllableStates[i] * (massDiagonal[i] * accDesired[i] + dampingLinearDiagonal[i]*velDesired[i] + dampingNonlinearDiagonal[i]*abs(velDesired[i])*velDesired[i]);
	}
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		TauI
//==================================================================================================
//
///	Calculates the integral term with saturation
///	
/// \param [in]		T					current simulation time
///
/// \param [in]		X					current simulation state
///
////////////////////////////////////////////////////////////////////////////////////////////////////
void NonlinearPIDController::CalculateTauI(const double T,const double* const X){
	for (size_t i = 0;i<6; i++){
		double temp_tauI = Ki(i,i) * X[indexErrorIntegral+i];
		double temp_sat = antiWindupConstants[i] * (Saturate((Ki(i,i) * X[indexErrorIntegral+i]),tauSatMin[i]*integralSaturationCoeffs[i],tauSatMax[i]*integralSaturationCoeffs[i]) - temp_tauI);
		a_tauI[i] = controllableStates[i] * Saturate(temp_tauI,tauSatMin[i]*integralSaturationCoeffs[i],tauSatMax[i]*integralSaturationCoeffs[i]);//, + temp_sat;	
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		Saturate
//==================================================================================================
//
///	Saturates the input value
///	
/// \param [in]		value				value to saturate
///
/// \param [in]		min					minimum value
///
/// \param [in]		max					maximum  value
///
/// \return			saturated value
///
////////////////////////////////////////////////////////////////////////////////////////////////////
double NonlinearPIDController::Saturate(double value, double min, double max){
	
	double saturated = value;
	if (value >= max)
		saturated = max;
	if (value <= min)
		saturated = min;
	
	return saturated;
}