////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class							NonlinearPassiveObserver
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// \author			Mats N{\aa}vik Hval
///
///	Nonlinear Passive Observer for a 6DOF vessel
///
/// \par Revision history:
///					06.06.2012 MNH: Initial version.
///
////////////////////////////////////////////////////////////////////////////////////////////////////
#include "NonlinearPassiveObserver.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		Nonlinear Passive Observer constructor
//==================================================================================================
///		This constructor performs all initial setup for a nonlinear passive observer simobject.
///		Reading in parameters, setting up communication interface i.e. output ports,
///		input ports, and states, plus additional 'one time only' resource setup.
///	
/// \param [in]		simObjectName	The name of the simobject. Used primarily 
///									by superclass constructor
///
/// \param [in]		creator			Retrive parameters. Register states, ports
///									and shared resources
///
////////////////////////////////////////////////////////////////////////////////////////////////////
NonlinearPassiveObserver::NonlinearPassiveObserver(const string& simObjectName, ISimObjectCreator* const creator) : SimObject(simObjectName)
{
	creator->AddInport	("StateEta",					6,	&EtaInputPort);
	creator->AddInport	("ThrusterForcesAndMoments",	6,	&ThrusterInputPort);
	creator->AddInport	("MooringForcesAndMoments",		6,	&MooringForcesInputPort);
	creator->AddOutport ("AccelerationBODY",			6,  PORT_FUNCTION(NonlinearPassiveObserver::AccelerationBODY));	
	creator->AddOutport ("VelocityBODY",				6,  PORT_FUNCTION(NonlinearPassiveObserver::VelocityBODY));	
	creator->AddOutport ("VelocityNED",					6,  PORT_FUNCTION(NonlinearPassiveObserver::VelocityNED));	
	creator->AddOutport ("PositionNED",					6,	PORT_FUNCTION(NonlinearPassiveObserver::PositionNED));
	creator->AddOutport ("PositionNEDwithWaves",		6,	PORT_FUNCTION(NonlinearPassiveObserver::PositionNEDwithWaves));
	creator->AddOutport ("BiasEstimate",				6,	PORT_FUNCTION(NonlinearPassiveObserver::BiasEstimate));
	creator->AddOutport ("Yerror",						6,	PORT_FUNCTION(NonlinearPassiveObserver::YError));

	double zero6[] = {0,0,0,0,0,0};
	creator->GetBoolParam("WaveFiltering", &waveFiltering,true);
	if (waveFiltering){
		creator->GetDoubleParamArray("WaveModel_OMEGA",waveModel_omega, 6);
		creator->GetDoubleParamArray("WaveModel_DELTA",waveModel_delta, 6);
		//creator->GetDoubleParam("MainWaveNumber",&waveNumber);
		creator->GetDoubleParamArray("K1_elements",aK1,12);
		double zero12[] = {0,0,0,0,0,0,0,0,0,0,0,0};
		indexWaveEstimate = creator->AddState("WaveEstimate",6,zero12);
		waveNumber = pow(waveModel_omega[1],2) * 9.81; 
	}

	creator->GetDoubleParamArray("BiasTimeConstants",bias_timeConstants,6);
	double expMax[] = {1000,1000,1000,1000,1000,1000};
	double expTime[] = {1,1,1,1,1,1};
	creator->GetDoubleParamArray("BiasExponentialTermMax",biasExponentialTermMax,6,expMax);
	creator->GetDoubleParamArray("BiasExponentialTermTimeConstants",biasExponentialTermTimeConstants,6,expTime);


	creator->GetDoubleParamArray("RigidBodyMassMatrixDiagonal",massMatrixDiagonal,6);
	
	double zeroFreqAddedMassDiagonal[6];
	double zeroFrequencyAddedMassMatrix[36];
	bool addedmassDiag = true;
		creator->GetDoubleParamArray("ZeroFrequencyAddedMassDiagonal",zeroFreqAddedMassDiagonal,6,zero6);
	if ( zeroFreqAddedMassDiagonal[0] == 0){
		addedmassDiag = false;
		creator->GetDoubleParamArray("ZeroFrequencyAddedMassMatrix",zeroFrequencyAddedMassMatrix,36);
	}

	creator->GetDoubleParamArray("LinearDampingMatrixDiagonal",linearDampingDiagonal,6);
	creator->GetDoubleParamArray("NonlinearDampingMatrixDiagonal",nonlinearDampingDiagonal,6,zero6);
	double ones[]= {1,1,1,1,1,1};
	/*creator->GetDoubleParamArray("WhichEtaMeasurementsAvailable",etaAvailableMeasurement,6,ones);
	creator->GetDoubleParamArray("WhichNuMeasurementsAvailable",nuAvailableMeasurement,6,ones);
	creator->GetDoubleParamArray("WhichNuDotMeasurementsAvailable",nuDotAvailableMeasurement,6,ones);
*/
	creator->AddInport("EtaMeasurementsAvailable",6,&StateMeasurementEta);
	//creator->AddInport("NuMeasurementAvailable",6,&StateMeasurementEta);
	//creator->AddInport("NuDotMeasurementAvailable",6,&StateMeasurementEta);
	
	creator->GetDoubleParamArray("K2_elements",aK2,6);
	creator->GetDoubleParamArray("K3_elements",aK3,6);
	creator->GetDoubleParamArray("K4_elements",aK4,6);

	indexBiasEstimate = creator->AddState("BiasEstimate",6,zero6);
	indexNuEstimate = creator->AddState("NuEstimate",6,zero6);
	indexEtaEstimate = creator->AddState("EtaEstimate",6);

	K1 = Eigen::Matrix<double,12,6>::Zero();
	K2 = Eigen::Matrix<double,6,6>::Zero();
	K3 = Eigen::Matrix<double,6,6>::Zero();
	K4 = Eigen::Matrix<double,6,6>::Zero();
	Aw = Eigen::Matrix<double,12,12>::Zero();
	Cw = Eigen::Matrix<double,6,12>::Zero();
	Tinv = Eigen::Matrix<double,6,6>::Zero();
	Dlinear = Eigen::Matrix<double,6,6>::Zero();
	Dnonlinear = Eigen::Matrix<double,6,6>::Zero();
	D = Eigen::Matrix<double,6,6>::Zero();
	Minv = Eigen::Matrix<double,6,6>::Zero();
	waveOmega = Eigen::Matrix<double,6,6>::Zero();
	waveDelta = Eigen::Matrix<double,6,6>::Zero();
	
	Eigen::Matrix<double,6,6> temp66 = Eigen::Matrix<double,6,6>::Zero();
	Y.setZero();
	VbiasEstimate.setZero();
	VwaveEstimate.setZero();
	VnuEstimate.setZero();
	VetaEstimate.setZero();

	for (size_t i = 0; i<6; i++){
		if(waveFiltering){
			K1(i,i) = aK1[i];
			K1(i+3,i) = aK1[i+3];
			waveOmega(i,i) = waveModel_omega[i];
			waveDelta(i,i) = waveModel_delta[i];
		}
		K2(i,i) = aK2[i];
		K3(i,i) = aK3[i];
		K4(i,i) = aK4[i];
		Tinv(i,i) = bias_timeConstants[i];
		Dlinear(i,i) = linearDampingDiagonal[i];
		Dnonlinear(i,i) = nonlinearDampingDiagonal[i];
		Minv(i,i) = massMatrixDiagonal[i];// + zeroFrequencyAddedMassMatrixDiagonal[i];
		if (addedmassDiag)
			Minv(i,i) = Minv(i,i) + zeroFreqAddedMassDiagonal[i];
		else 
			for (size_t j=0;j<6;j++)
				Minv(i,j) = Minv(i,j) + zeroFrequencyAddedMassMatrix[i*6+j];
		accelerationBODY[i] = 0;
		velocityBODY[i] = 0;
		velocityNED[i] = 0;
		positionNED[i] = 0;
		positionNEDwithWaves[i] = 0;
	}

	temp66 = Minv.inverse();
	Minv = temp66;

	temp66 = Tinv.inverse();
	Tinv = temp66;

	if (waveFiltering){
		Aw.topRightCorner(6,6) = Eigen::Matrix<double,6,6>::Identity();
		Aw.bottomLeftCorner(6,6) = -waveOmega * waveOmega;
		Aw.bottomRightCorner(6,6) = -2*waveDelta*waveOmega;
		Cw.bottomRightCorner(6,6) = Eigen::Matrix<double,6,6>::Identity();
	}
}

void NonlinearPassiveObserver::OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep){
	const double* thrusterForces = ThrusterInputPort->GetPortValue(T,X);
	const double* mooring = MooringForcesInputPort->GetPortValue(T,X);
	Eigen::Matrix<double,6,6> exponentialTermBias = Eigen::Matrix<double,6,6>::Identity();
	for (size_t i = 0;i<6;i++)
		exponentialTermBias(i,i) = (1+biasExponentialTermMax[i]*exp(-biasExponentialTermTimeConstants[i]*T));//*exponentialTermBias(i,i));
	const double *etaAvailableMeasurements = StateMeasurementEta->GetPortValue(T,X);

	const double* measure = EtaInputPort->GetPortValue(T,X);
	for (size_t i = 0; i<6; i++){
		Y(i) = measure[i];
		etaMeasurement[i] = Y(i);
		VbiasEstimate(i) = X[indexBiasEstimate+i];
		if (waveFiltering){
			VwaveEstimate(i) = X[indexWaveEstimate+i];
			VwaveEstimate(i+6) = X[indexWaveEstimate+i+6];
		}
		VnuEstimate(i) = X[indexNuEstimate+i];
		VetaEstimate(i) = X[indexEtaEstimate+i];
		ThrusterForcesAndMoments(i) = thrusterForces[i];
		MooringForcesAndMoments(i) = mooring[i];
	}

	J = Jmatrix(etaMeasurement);//.cast<double>();
	Jinv = J.inverse();

	if (waveFiltering){
		Eigen::Matrix<double,6,1> waveTerm = Cw * VwaveEstimate;
		for (size_t i =0; i<6; i++)
			Yhat(i) = waveTerm(i)* exp(-X[indexEtaEstimate+2]*waveNumber) +  VetaEstimate(i);
	}
	else
		Yhat = VetaEstimate;

	for (size_t i =0 ; i< 6; i++)
		Ytilde(i) = (Y(i) - Yhat(i))*etaAvailableMeasurements[i];

	Eigen::Matrix<double,6,6> temp = Eigen::Matrix<double,6,6>::Zero();
	for (size_t i = 0; i<6 ; i++){
		temp(i,i) = Dnonlinear(i,i) * abs(VnuEstimate(i));
	}
	D = temp + Dlinear;

	VwaveEstimateDot = K1*Ytilde + Aw * VwaveEstimate;
	VbiasEstimateDot = K3*exponentialTermBias*Ytilde - Tinv*VbiasEstimate;
	//TODO: sjekk om rekkefolgen er riktig i neste linje
	Eigen::Matrix<double,6,1> tempMat= K4*Ytilde;
	VnuEstimateDot = Minv * ((Jinv*tempMat) +  ThrusterForcesAndMoments - (Jinv*MooringForcesAndMoments) - (D*VnuEstimate) + (Jinv * VbiasEstimate));
	VetaEstimateDot = J*VnuEstimate + K2*Ytilde;

	
	for (size_t i =0 ; i<6 ; i++){
		XDot[indexBiasEstimate+i] = VbiasEstimateDot(i);
		XDot[indexEtaEstimate+i] = VetaEstimateDot(i);
		XDot[indexNuEstimate+i] = VnuEstimateDot(i);
		if (waveFiltering)
			XDot[indexWaveEstimate+i] = VwaveEstimateDot(i);

		accelerationBODY[i] = VnuEstimateDot(i);
		velocityNED[i] = VetaEstimateDot(i);
		velocityBODY[i] = VnuEstimate(i);
		positionNED[i] = VetaEstimate(i);
		bias[i] = VbiasEstimate(i);
		yTilde[i] = Ytilde(i);

	}
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		AccelerationBODY
//==================================================================================================
//
///	Returns the acceleration
///
/// \param [in]		T					dummy
///
/// \param [in]		X					dummy
///
/// \returns								array containing 6DOF accelerations in BODY frame
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const NonlinearPassiveObserver::AccelerationBODY(const double T, const double* const X){
		return accelerationBODY;
	}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		VelocityBODY
//==================================================================================================
//
///	Returns the velocity in BODY frame
///
/// \param [in]		T					dummy
///
/// \param [in]		X					dummy
///
/// \returns								array containing 6DOF velocities in BODY frame
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const NonlinearPassiveObserver::VelocityBODY(const double T, const double* const X){
		return velocityBODY;
	}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		VelocityNED
//==================================================================================================
//
///	Returns the velocity in NED-frame
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					dummy
/// 
/// \returns								array containing 6DOF velocity in NED
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const NonlinearPassiveObserver::VelocityNED(const double T, const double* const X){
		return velocityNED;
	}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		PositionNED
//==================================================================================================
//
///	Returns the filtered position in NED-frame
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					dummy
/// 
/// \returns								array containing filtered 6OF position in NED
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const NonlinearPassiveObserver::PositionNED(const double T, const double* const X){
		return positionNED;
	}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		PositionNEDwithWaves
//==================================================================================================
//
///	Sorts a vector of netnodes on ascending north position
/// 
/// \param [in]		T				current simulation time
///
/// \param [in]		X				current simulation state
///
/// \returns								array containing 6DOF position and attitude with wave disturbance
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const NonlinearPassiveObserver::PositionNEDwithWaves(const double T, const double* const X){
		
		Eigen::Matrix<double,6,6> etaWithWaves = Eigen::Matrix<double,6,6>::Zero();

		//VetaEstimate << X[indexEtaEstimate],X[indexEtaEstimate+1],X[indexEtaEstimate+2],X[indexEtaEstimate+3],X[indexEtaEstimate+4],X[indexEtaEstimate+5];
		
		if (!waveFiltering){
			return X+indexEtaEstimate;
		}
		else{
			for (size_t i = 0; i < 6;i++)
				positionNEDwithWaves[i] = Yhat(i);

		return positionNEDwithWaves;
		}
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		BiasEstimate
//==================================================================================================
//
///	Returns the bias estimate
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					dummy
///	
/// \returns								array containing the bias estimate
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const NonlinearPassiveObserver::BiasEstimate(const double T, const double* const X){
	return bias;
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		YError
//==================================================================================================
//
/// Returns the error between measured and estimated position. y-yHat.
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					dummy
///	
/// \returns								array containing error between measured and estimated position
///
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const NonlinearPassiveObserver::YError(const double T, const double* const X){
	return yTilde;
}
#ifdef FH_VISUALIZATION
	void RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator){} 
	void RenderUpdate(const double T, const double* const X){}
#endif
	