////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class							Nonlinear Passive Observer
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// \brief			Nonlinear Passive Observer as described in Fossen 2011
///
///
/// \author			Mats N{\aa}vik Hval
///
////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef NONLINEARPASSIVEOBSERVER_H
#define NONLINEARPASSIVEOBSERVER_H

#include "SimObject.h"
#include <string>
#include "Eigen\Dense"
#include "HelpFunctions.h"

class NonlinearPassiveObserver : public SimObject
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW
	NonlinearPassiveObserver(const string& simObjectName, ISimObjectCreator* const creator); ///< reads parameters, registers states, input/output ports and shared resources
	void OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep); ///< returns object derivatives as a function of time, states and input ports
	const double* const AccelerationBODY(const double T, const double* const X);
	const double* const VelocityBODY(const double T, const double* const X);
	const double* const VelocityNED(const double T, const double* const X);
	const double* const PositionNED(const double T, const double* const X);
	const double* const PositionNEDwithWaves(const double T, const double* const X);
	const double* const BiasEstimate(const double T, const double* const X);
	const double* const YError(const double T, const double* const X);

#ifdef FH_VISUALIZATION
	void RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator){} ///< sets up visualization resources
	void RenderUpdate(const double T, const double* const X){} ///< updates visualization resources to reflect current state
#endif
	
protected:
	ISignalPort* EtaInputPort;								///< input port. Current position and attitude
	ISignalPort* ThrusterInputPort;							///< input port. Thruster forces and moments
	ISignalPort* MooringForcesInputPort;					///< input port. Forces and moments from optional mooring lines
	ISignalPort* StateMeasurementEta;						///< input port
	//ISignalPort* StateMeasurementNu;
	//ISignalPort* StateMeasurementNuDot;

	bool waveFiltering;										///< Is the wave filter active?
	double stateEtaEstimate[6];								///< Estimate of the position and attitude
	double stateBiasEstimate[6];							///< Estimate of the slowly varying bias
	double stateWaveEstimate[12];							///< Estimate of the waves
	double etaMeasurement[6];								///< Measurements of the vessel position and attitude
	double etaError[6];										///< Estimation error. EtaMeasurement - EtaEstimate

	int indexWaveEstimate;									///< index of the wave estimate in the state vector
	int indexBiasEstimate;									///< index of the bias estimate in the state vector
	int indexNuEstimate;									///< index of nuHat in the state vector
	int indexEtaEstimate;									///< index of etaHat in the vector

	double accelerationBODY[6];								///< return array for the BODY acceleration
	double velocityBODY[6];									///< return array for the BODY velocity
	double velocityNED[6];									///< return array for the NED velocity
	double positionNED[6];									///< return array for the NED position
	double positionNEDwithWaves[6];							///< return array for the NED position with waves
	double yTilde[6];										///< return array for the error between measured and estimated position

	double waveModel_omega[6];								///< wave model parameters, omea
	double waveModel_delta[6];								///< wave model parameters, delta 
	double bias_timeConstants[6];							///< bias model parameters, timeconstants
	double linearDampingDiagonal[6];						///< diagonal elements of the linear damping
	double nonlinearDampingDiagonal[6];						///< diagonal elements of the quadratic damping
	double massMatrixDiagonal[6];							///< diagonal elements of the combined rigid body and added mass matrix
	double aK1[12];											///< Diagonal elements of K1 tuning matrix
	double aK2[6];											///< Diagonal elements of K2 tuning matrix
	double aK3[6];											///< Diagonal elements of K3 tuning matrix
	double aK4[6];											///< Diagonal elements of K4 tuning matrix
	
	///*********************
	///		Wave model matrices
	///*********************
	Eigen::Matrix<double,12,6> K1;							///< K1 tuning matrix of the wave model
	Eigen::Matrix<double,12,12> Aw;							///< Aw model matrix of the wave model
	Eigen::Matrix<double,6,12> Cw;							///< CW wave term measurement matrix of the wave model
	Eigen::Matrix<double,6,6> waveOmega;					///< wave model omega
	Eigen::Matrix<double,6,6> waveDelta;					///< Wave model delta
	double waveNumber;										///< wavenumber of the wave with the largest impact

	///*********************
	///		Bias model matrices
	///*********************
	Eigen::Matrix<double,6,6> K3;							///< K3 tuning matrix for the bias model
	Eigen::Matrix<double,6,6> Tinv;							///< Iverse bias time constant matrix
	double bias[6];											///< return array for the bias estimate
	double biasExponentialTermMax[6];						///< maxvalue of the bias exponential term. 
	double biasExponentialTermTimeConstants[6];				///< time constant of the bias exponential term

	///*********************
	///		Damping matrices
	///*********************
	Eigen::Matrix<double,6,6> Dlinear;						///< linear damping matrix
	Eigen::Matrix<double,6,6> Dnonlinear;					///< quadratic damping matrix
	Eigen::Matrix<double,6,6> D;							///< sum of linear and quadratic damping

	///*********************
	///		Other observer matrices
	///*********************
	Eigen::Matrix<double,6,6> J;							///< 6DOF transformation matrix
	Eigen::Matrix<double,6,6> Jinv;							///< inverse 6DOF transformation matrix
	Eigen::Matrix<double,6,6> K4;							///< K4 tuning matrix
	Eigen::Matrix<double,6,6> Minv;							///< inverse mass matrix
	Eigen::Matrix<double,6,6> K2;							///< K2 tuning matrix
	Eigen::Matrix<double,6,1> Yhat;							///< measurement estimate vector
	Eigen::Matrix<double,6,1> Y;							///< measurement vector
	Eigen::Matrix<double,6,1> Ytilde;						///< Estimate error vector
	Eigen::Matrix<double,6,1> VbiasEstimate;				///< bias estimate vector
	Eigen::Matrix<double,6,1> VbiasEstimateDot;				///< estimateDot vector
	Eigen::Matrix<double,6,1> VetaEstimate;					///< etaEstimate vector
	Eigen::Matrix<double,6,1> VetaEstimateDot;				///< eta dot estimate vectors
	Eigen::Matrix<double,6,1> VnuEstimate;					///< nu estimate vector
	Eigen::Matrix<double,6,1> VnuEstimateDot;				///< nudot estimate vector
	Eigen::Matrix<double,12,1> VwaveEstimate;				///<  wave estimate vector
	Eigen::Matrix<double,12,1> VwaveEstimateDot;				///< waveestimateDot vector
	Eigen::Matrix<double,6,1> ThrusterForcesAndMoments;		///< thruster forces and moments vector
	Eigen::Matrix<double,6,1> MooringForcesAndMoments;		///< mooring loads vector
};

#endif
