#include <assert.h>
#include "PIDcontroller.h"

PIDController::PIDController(const string& simObjectName, ISimObjectCreator* const creator) : ControlAlgorithm(simObjectName,creator)
{
	double Kp_array[6];
	double Ki_array[6];
	double Kd_array[6];
	
	creator->GetDoubleParamArray("ProportionalGainVector",Kp_array,6);
	creator->GetDoubleParamArray("IntegralGainVector",Ki_array,6);
	creator->GetDoubleParamArray("DerivativeGainVector",Kd_array,6);

	Kp = Eigen::Matrix<double,6,6>::Zero();
	Ki = Eigen::Matrix<double,6,6>::Zero();
	Kd = Eigen::Matrix<double,6,6>::Zero();

	for (size_t i = 0 ; i < 6; i++){
		Kp(i,i) = Kp_array[i];
		Ki(i,i) = Ki_array[i];
		Kd(i,i) = Kd_array[i];
	}
		
	double zero[6] = {0,0,0,0,0,0};
	indexErrorIntegral = creator->AddState("ErrorIntegral",6,zero);

	error = Eigen::Matrix<double,6,1>::Zero();
	errorDot = Eigen::Matrix<double,6,1>::Zero();
	errorIntegral = Eigen::Matrix<double,6,1>::Zero();
	force = Eigen::Matrix<double,6,1>::Zero();
	for (size_t i = 0; i<6;i++){
		//error(i) = 0;
		//errorDot(i) = 0;
		//errorIntegral(i) = 0;
		//force(i) = 0;
		error[i] = 0;
		forcesAndMomentsBODY[i] = 0;
	}

	creator->AddOutport("ErrorIntegral",6,PORT_FUNCTION(PIDController::ErrorIntegral));
	creator->AddOutport("TauP",6,PORT_FUNCTION(PIDController::TauProp));
	creator->AddOutport("TauI",6,PORT_FUNCTION(PIDController::TauInt));
	creator->AddOutport("TauD",6,PORT_FUNCTION(PIDController::TauDamp));

}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ForcesAndMomentsBODY
//==================================================================================================
//
///	Returns the forces and moments desired from the controller
/// 
///	
/// \param [in]		T						current simulation time
///
/// \param [in]		X						current simulation state
///
/// return									desired forces and moments in 6DOF
///
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const PIDController::ForcesAndMomentsBODY(const double T, const double* const X){

	const double* eta = inputEta->GetPortValue(T,X);

	CalculateTauP();
	CalculateTauD();
	CalculateTauI();
	Eigen::Matrix<double,6,1> forcesNED;
	for (size_t i = 0; i<6;i++)
		forcesNED(i) = a_tauP[i] + a_tauI[i] + a_tauD[i];

	force = -ControlAlgorithm::JMatrix(eta).inverse() * (forcesNED);

	for (size_t i =0;i<6;i++)
		forcesAndMomentsBODY[i] = force(i);
		
	return forcesAndMomentsBODY; 
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		OdeFnc
//==================================================================================================
//
///	Calculates the error in position and velocity, and assignes the position error to the state vector.
/// The error is defined as value - desired_value.
///	
/// \param [in]		T					current simulation time
///
/// \param [in]		X					current simulation state
///
/// \param [in,out]		XDot					derivative of the state
///
/// \param [in]		bIsMajorTimeStep		major time step flag
///
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
void PIDController::OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep){
	const double *a_eta = inputEta->GetPortValue(T,X);
	const double *a_etaDesired = inputEtaDesired->GetPortValue(T,X);
	const double *a_nu = inputNu->GetPortValue(T,X);
	const double *a_nuDesired = inputNuDesired->GetPortValue(T,X);

	for (size_t i=0;i<6;i++){
		error(i) = a_eta[i] - a_etaDesired[i];
		errorDot(i) = a_nu[i] - a_nuDesired[i];
	}

	for (size_t i = 0;i<3;i++){
		a_error[i] = error(i);
		a_error[i+3] = ShortestError(a_eta[i], a_etaDesired[i]);
	}
	
	for(size_t i = 0 ; i < 6 ; i++){
		error(i) = a_error[i];
		errorIntegral(i) = X[indexErrorIntegral+i];
		XDot[indexErrorIntegral+i] = a_error[i];
	}
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		Error
//==================================================================================================
//
///	Returns the error-array
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					dummy
///
/// \return								array of errors in position and attitude
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const PIDController::Error(const double T,const double* const X){

	return a_error;
	//return X +errorIntegralIndex;
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ErrorIntegral
//==================================================================================================
//
///	Returns the errorintegral-vector
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					current simulation state
///
/// \return								integral of the error.
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const PIDController::ErrorIntegral(const double T,const double* const X){
	
	for( size_t i =0;i<6;i++){
		a_errorIntegral[i] = X[indexErrorIntegral+i];
	}
	return a_errorIntegral;
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		TauP
//==================================================================================================
//
///	Calculates the forces in the NED frame proportional with the error
///	
////////////////////////////////////////////////////////////////////////////////////////////////////
void PIDController::CalculateTauP(){
	
	Eigen::Matrix<double,6,1> forcesP = Kp * error;
	
	for (size_t i =0;i<6;i++)
		a_tauP[i] = controllableStates[i] * forcesP(i);
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		TauI
//==================================================================================================
//
///	Calculates the forces in the NED frame proportional with the integral of the error
///	
////////////////////////////////////////////////////////////////////////////////////////////////////
void PIDController::CalculateTauI(){
	
	Eigen::Matrix<double,6,1> forcesI = Ki * errorIntegral;

	for (size_t i =0;i<6;i++)
		a_tauI[i] = controllableStates[i] *forcesI(i);
	
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		TauD
//==================================================================================================
//
///	Calculates the forces in the NED frame proporional with the derivative of the error
///	
////////////////////////////////////////////////////////////////////////////////////////////////////
void PIDController::CalculateTauD(){
	
	Eigen::Matrix<double,6,1> forcesD = Kd * errorDot;
	
	for (size_t i =0;i<6;i++)
		a_tauD[i] = controllableStates[i] * forcesD(i);
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		TauProp
//==================================================================================================
//
///	Returns the proportional term
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					dummy
///
/// \return								array of the proportional term
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const PIDController::TauProp(const double T, const double* const X){
	return a_tauP;
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		TauInt
//==================================================================================================
//
///	Returns the proportional term
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					dummy
///
/// \return								array of the integral term
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const PIDController::TauInt(const double T, const double* const X){
	return a_tauI;
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		TauProp
//==================================================================================================
//
///	Returns the proportional term
///	
/// \param [in]		T					dummy
///
/// \param [in]		X					dummy
///
/// \return								array of the derivative term
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const PIDController::TauDamp(const double T, const double* const X){
	return a_tauD;
}