/**
\class PIDController
\par User Documentation:

Simple PID controller
\par Parameters
<table >
<tr><td><b>Parameter<td>Size<td>Comment</b>
<tr><td>ProportionalGainVector<td>6<td>Coefficients for the proporional gain
<tr><td>IntegralGainVector<td>6<td>Coefficients for the integral gain
<tr><td>DerivativeGainVector<td>6<td>Coefficients for the derivative gain
</table>

\par Output ports
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>ErrorIntegral<td>6<td>Integral of the error
<tr><td>TauP<td>6<td>Proporional term of the controller
<tr><td>TauI<td>6<td>Integral term of the controller
<tr><td>TauD<td>6<td>Derivative term of the controller
</table>

*/
#ifndef PIDCONTROLLER_H
#define PIDCONTROLLER_H

#include "Eigen\Dense"
#include <sfhGeneral.h>
#include "ControlAlgorithm.h"

class PIDController : public ControlAlgorithm
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW
	PIDController(const string& simObjectName, ISimObjectCreator* const creator); ///< Reads parameters, registers states, input/output ports and shared resources
	const double* const ForcesAndMomentsBODY(const double T, const double* const X);	///< Implements the forces and moments calculation
	void OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep); ///< Returns object derivatives as a function of time, states and input ports
	const double* const Error(const double T,const double* const X);					///< Implements return of the error in position.
	const double* const ErrorIntegral(const double T,const double* const X);			///< Returns the integral of the error
	const double* const TauProp(const double T, const double* const X);					///< Returns the proportional term of the controller
	const double* const TauInt(const double T, const double* const X);					///< Returns the integral term of the controller		
	const double* const TauDamp(const double T, const double* const X);					///< Returns the derivative term of the controller
	
#ifdef FH_VISUALIZATION
	/// The rendering is not used.
	virtual void RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator){}

	/// The rendering is not used.
	virtual void RenderUpdate(const double T, const double* const X){}
#endif		

protected:
	virtual void CalculateTauP();				///< Calculate the proportional term
	virtual void CalculateTauI();				///< Calculate the integral term
	virtual void CalculateTauD();				///< Calculate the derivative term
	Eigen::Matrix<double,6,6> Kp;				///< Diagonal matrix holding the porportional controller gain
	Eigen::Matrix<double,6,6> Ki;				///< Diagonal matrix holding the porportional controller gain
	Eigen::Matrix<double,6,6> Kd;				///< Diagonal matrix holding the porportional controller gain

	Eigen::Matrix<double,6,1> error;			///< Vector of position error, for calculations
	Eigen::Matrix<double,6,1> errorDot;			///< Vector of error in velocity, for calculations
	Eigen::Matrix<double,6,1> errorIntegral;	///< Vector of error-integral, for calculations
	Eigen::Matrix<double,6,1> force;			///< Resulting forces and moments, for calculations
		
	int indexErrorIntegral;						///< Index for the error-integral-state in the state vector

	double forcesAndMomentsBODY[6];				///< Return array for body forces and moments
	double a_error[6];							///< Return array for the error
	double a_errorIntegral[6];					///< Return array for the error integral
	double integralLimits[6];					///< Array holding integral limits.
	double a_tauP[6];							///< Return array for the proportional term forces and moments
	double a_tauI[6];							///< Return array for the integral term forces and moments
	double a_tauD[6];							///< Return array for the derivative term forces and moments


#ifdef FH_VISUALIZATION
#endif

};
#endif