#include "SampledSignalWithNoise.h"
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		SampledSignalWithNoise Constructor
//==================================================================================================
///		This constructor performs all initial setup for the referencegenerator simobject.
///		Reading in parameters, setting up communication interface i.e. output ports,
///		input ports, and states, plus additional 'one time only' resource setup.
///	
/// \param [in]		simObjectName	The name of the simobject. Used primarily 
///									by superclass constructor
///
/// \param [in]		creator			Retrieve parameters. Register states, ports
///									and shared resources
///
////////////////////////////////////////////////////////////////////////////////////////////////////
SampledSignalWithNoise::SampledSignalWithNoise(const string& simObjectName, ISimObjectCreator* const creator) : SimObject(simObjectName){

	creator->GetIntParam("NumberOfSignals",&numberOfSignals);

	coloredSignals = new double[numberOfSignals];
	noiseAmplitude = new double[numberOfSignals];
	signalSamplingFreq = new double[numberOfSignals];
	signal = new double[numberOfSignals];
	count = new int[numberOfSignals];
	countReset = new int[numberOfSignals];

	creator->GetDoubleParamArray("NoiseAmplitude",noiseAmplitude,numberOfSignals);
	creator->GetDoubleParamArray("SamplingFrequency",signalSamplingFreq,numberOfSignals);

	creator->AddInport("In",numberOfSignals,&signalIn);
	creator->AddOutport("Out",numberOfSignals,PORT_FUNCTION(SampledSignalWithNoise::Signal));
	for (size_t i =0;i<numberOfSignals;i++)
		signal[i] = 0;
	
}
SampledSignalWithNoise::~SampledSignalWithNoise(){
	delete coloredSignals,noiseAmplitude,signalSamplingFreq,signal,count,countReset;
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		randomDouble
//==================================================================================================
//
///	Generates a random number between min and max input.
///	
/// \param [in]		min						minimum value of the random number
///
/// \param [in]		max						maximum value of the random number
///
/// return									random number
///
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
float SampledSignalWithNoise::randomDouble(double min, double max){

	float randomNumber = float(rand()) / RAND_MAX;

	return min + randomNumber *(max - min);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		Signal
//==================================================================================================
//
///	Returns the sampled signal with noise
/// ///	
/// \param [in]		T						dummy
///
/// \param [in]		X						dummy
///
/// \return									return the array of distorted signals
///
////////////////////////////////////////////////////////////////////////////////////////////////////

const double* const SampledSignalWithNoise::Signal(const double T, const double* const X){
	
	return signal;

}
void SampledSignalWithNoise::OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep){
	//TODO: fix such that the sampling is correct

	if (bIsMajorTimeStep){
		for (size_t i = 0;i<numberOfSignals; i++){
			double sampTime = 1/signalSamplingFreq[i];
			double modTime = fmod(T,sampTime);
			double tol = 1e-6;
			if (abs(modTime< tol) || abs(sampTime - modTime) < tol){
					signal[i] = signalIn->GetPortValue(T,X)[i] + randomDouble(-noiseAmplitude[i],noiseAmplitude[i]);
			}
			else
				int a =1;
		}
	}
}


#ifdef FH_VISUALIZATION
void SampledSignalWithNoise::RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator)
{
}

void SampledSignalWithNoise::RenderUpdate(const double T, const double* const X)
{
}
#endif