////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class							ThrusterAllocation
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// 		Thruster Allocation
///
///
/// \author			Mats N{\aa}vik Hval
///
/// \par Revision history:
///					30.04.2012 MNH: Initial version.
///
////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef THRUSTERALLOCATION_H
#define THRUSTERALLOCATION_H

#include <Eigen/Dense>
#include <Eigen/SVD>
#include <sfhGeneral.h>
#include "HelpFunctions.h"
#include "SimObject.h"

#ifndef THRUSTERSTRUCT
#define THRUSTERSTRUCT
struct ThrusterStruct{
	double positionAndAttitude[5]; ///< Position and attitude of the thruster with regards to the vehicles center of origin.
	int RPMlimits[2];				///< minimum and maximum rotations per minute per thruster
	double RPMthrustCoeffs[2];		///< coefficients describing a second order curve of the relation between RPM and thrust
	double rpm;						///< calculated rpm
	double thrust;					///< calculated thrust
};
#endif
class ThrusterAllocation : public SimObject
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW ///< enabling use of Eigen types as members
	ThrusterAllocation(const string& simObjectName, ISimObjectCreator* const creator); ///< reads parameters, registers states, input/output ports and shared resources
	const double* const RPM(const double T, const double* const X); ///< returns array of commanded RPMs
	const double* const ThrustCommanded(const double T, const double* const X); ///< returns array of corresponding tau
	void OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep); ///< returns object derivatives as a function of time, states and input ports
	

	
#ifdef FH_VISUALIZATION
	/// The rendering is not used.
	virtual void RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator){}

	/// The rendering is not used.
	virtual void RenderUpdate(const double T, const double* const X){}
#endif		

private:
	const double* const thrustFromRPM(const double* const RPM, int thrusterNumber); ///< calculate thrust from given RPM
	const double* const ThrusterAllocation::RPMfromU(const double* const ui, int thrusterNumber); ///< calculate RPM from calculated u
	ISignalPort* Tau_desired;																		///< input port for desired tau
	int numberOfThrusters;															///< number of thrusters
	//int RPMsingle;																	///< 
	double *RPMvalues;																///< array of commanded RPM-values
	//vector<int*> RPMlimits;
	vector<ThrusterStruct> thrusters;												///< vector of thrusters
	
	Eigen::MatrixXd Thr;															///< Thrust configuration matrix
	Eigen::MatrixXd TpseudoInverse;													///< pseudo inverse of the thrust configuration matrix
	Eigen::MatrixXd K;																///< thrust coefficient matrix
	Eigen::MatrixXd Kinv;															///< inverse of the thrust coefficient matrix
	Eigen::VectorXd tau;															///< tau-desired
	Eigen::VectorXd tauCommand;														///< tau-commanded
	Eigen::VectorXd u;																///< calculated effort

	//int *previousRPM;
	//bool fileRead;
	//string fileName;
	//string filePath;
	string allocationType;															///< containing type of thruster setup
	bool linearRelation;															///< flag for linear relation between rpm and thrust
	//double thrust;																	
	double tauCommanded[6];															///< return array for commanded thrust
	double controllableStates[6];													///< array holding 0 and 1, where 1 undicates uncontrollable state.

#ifdef FH_VISUALIZATION
#endif


#endif
};
