#include "WpManagementSystem.h"
#include "HelpFunctions.h"
#include <CNetCage.h>

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		WPManagementSystem constructor
//==================================================================================================
///		This constructor performs all initial setup for a reference generator simobject.
///		Reading in parameters, setting up communication interface i.e. output ports,
///		input ports, and states, plus additional 'one time only' resource setup.
///	
/// \param [in]		simObjectName	The name of the simobject. Used primarily 
///									by superclass constructor
///
/// \param [in]		creator			Retrive parameters. Register states, ports
///									and shared resources
///
////////////////////////////////////////////////////////////////////////////////////////////////////
WpManagementSystem::WpManagementSystem(const string& simObjectName, ISimObjectCreator* const creator) : GuidanceSystem(simObjectName,creator){
	
	string wpGeneratorType = "Predefined";
	creator->GetStringParam("WPGeneratorType",wpGeneratorType);
	creator->GetDoubleParam("HeadingToleranceDegrees",&angleTol);
	if (wpGeneratorType.compare("Predefined") == 0 || wpGeneratorType.compare("PreDefined") == 0 || wpGeneratorType.compare("PREDEFINED") == 0 || wpGeneratorType.compare("predefined") == 0){
		currentWaypoint = 0;
		predefinedWP = true;
		int numberOfWaypoints =0;
		creator->GetIntParam("NumberOfWaypoints",&numberOfWaypoints);
		creator->GetDoubleParam("RadiusOfAcceptance",&radiusOfAcceptance);
		for (size_t i =1 ; i< numberOfWaypoints+1; i++){
			string parameterName = "WP";
			Waypoint *wp = new Waypoint();
			double position[3];
			creator->GetDoubleParamArray(parameterName + intToString(i),position,3);
			wp->setPosition(*position,*(position+1),*(position+2));
			wp->setRadiusOfAcceptance(radiusOfAcceptance);
			wpVector.push_back(wp);
		}
		waypointIndex = 0;
		atWP = false;
		
	}
	else 
		predefinedWP = false;
	if (wpGeneratorType.compare("NetCageTraverser") == 0 || wpGeneratorType.compare("NetcageTraverser") == 0 || wpGeneratorType.compare("netcagetraverser") == 0 || wpGeneratorType.compare("NETCAGETRAVERSER") == 0|| wpGeneratorType.compare("Netcagetraverser") == 0){
		creator->GetStringParam("NetCageName",m_NetCageName);
		netcageTraversing = true;
		currentNetNode = 0;
		level = 0;
		std::string	 traversePattern = "";
		creator->GetStringParam("TraversePattern",traversePattern);
		if (traversePattern.compare("UPDOWN") == 0 || traversePattern.compare("UpDown") == 0 || traversePattern.compare("updown") == 0)
			traverseUpNdown = true;
		else
			traverseCircles = true;
		creator->GetDoubleParam("RadiusOfAcceptance",&radiusOfAcceptance);
		currentNetNode = 0;
	}
	else
		netcageTraversing = false;
}
WpManagementSystem::~WpManagementSystem(){
	//delete waypoints for the predefined case
	if (!wpVector.empty()){
		for (size_t i =0; i< wpVector.size();i++){
			delete wpVector.back();
			wpVector.pop_back();
		}
	}
	//delete netnodestructs and waypoints for the netcagetraverser case
	if (!netcage.empty()){
		for (size_t i =0; i< netcage.size();i++){
			for (size_t j =0;j< netcage.back().size();j++){
				delete netcage.back().back();
				netcage.back().pop_back();
			}
			netcage.pop_back();
		}
	}
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		FinalSetup
//==================================================================================================
///		This method registers the netcage nodes
///	
/// \param [in]		T				Current simulation time
/// \param [in]		X				Current simulation states
///									
///
/// \param [in]		creator			Retrive parameters. Register states, ports
///									and shared resources
///
////////////////////////////////////////////////////////////////////////////////////////////////////
void WpManagementSystem::FinalSetup(const double T, const double *const X, ISimObjectCreator* const creator){

		m_Environment = dynamic_cast<DeepSeaGravityWaves*>( static_cast<CEnvironment*>(creator->GetSharedResource("CSeaEnvironment") ));
	
		if (netcageTraversing){
			double position0[] = {0,0};
			double depth = m_Environment->GetSeadepth(position0);
			for (size_t i = 0;i< depth;i++){
				vector<NetNodeStruct*> depth;
				netcage.push_back(depth);
			}
	
			//std::cout << netcage.size() << std::endl;
	
			CNetCage *my_netcage = static_cast<CNetCage*>( creator->GetSharedResource(m_NetCageName+"Pointer"));
		
			std::cout << "my_netcage: " <<my_netcage << std::endl;
	
			std::vector<int> pos_idx;
			std::vector<int> vel_idx;

			my_netcage->GetNodePositionIndices(pos_idx);
			my_netcage->GetNodeVelocityIndices(vel_idx);
			//std::cout << pos_idx.size() << std::endl;
			double radiusOfAcceptance = 0.1;
			for( int i=0; i < pos_idx.size(); i++){
				if (X[pos_idx[i]+2] >= 0){
					NetNodeStruct *node = new NetNodeStruct();
				Waypoint *wp = new Waypoint(const_cast<double *const> (X+pos_idx[i]),const_cast<double *const> (X+pos_idx[i]+1),const_cast<double *const> (X+pos_idx[i]+2),radiusOfAcceptance);
					node->WP = wp;
					//netcage[int(pos_idx[i+2])].back()->WP = new WaypointTEST(X[pos_idx[i]],X[pos_idx[i]+1],X[pos_idx[i]+2],radiusOfAcceptance);
						netcage[int(X[pos_idx[i]+2])].push_back(node);
					//std::cout << i << std::endl;
				}
			}
			
			//std::cout << "Netcage size: " << netcage.size() << std::endl;
			for(size_t i =0;i<netcage.size();i++){
			//std::cout << "Nodes at depth "<< i << "-" << i+1 <<": " << netcage[i].size() << std::endl;
		}
			for (size_t i = 0; i <16;i++){
				double n = netcage.at(1).at(i)->WP->Position()[0];
				double e = netcage.at(1).at(i)->WP->Position()[1];
				double d = netcage.at(1).at(i)->WP->Position()[2];
	
			//	std::cout << n << "." << e << "." << d << std::endl;
			}
			std::vector<std::vector<NetNodeStruct*>> temp;
			for (size_t i = 0; i< netcage.size();i++){
					if (!netcage[i].empty()){
					std::vector<NetNodeStruct*> oneDepth = quicksort(netcage[i]);
					findNeighboringNodes(oneDepth);
					temp.push_back(oneDepth);				
				}
			}
			netcage = temp;
			cageDiameter = abs(netcage[0][0]->WP->Position()[0]) + abs(netcage[0][netcage[0].size()-1]->WP->Position()[0]);
		}
	//	for printing out the sorted list:
			//for (size_t i =0; i<1/*netcage.size()*/;i++){
		//	for (size_t j=0;j< netcage[i].size();j++){
		//		double x = netcage.at(i).at(j)->WP->Position()[0];
		//		double y = netcage.at(i).at(j)->WP->Position()[1];
		//		double z = netcage.at(i).at(j)->WP->Position()[2];
		//		std::cout << z << " : " << x << " : " << y << std::endl;
		//	}
		//	std::cout << "------------------------------------" << std::endl;
		//}
	
		////for (size_t i =0;i<netcage.size()-1;i++){
		//int i =0;
		//	std::cout << "Size : " << netcage[i].size() << std::endl;
		//	double *pos1 = netcage[i][0]->WP->Position();
		//	double *posR = netcage[i][0]->Right->WP->Position();
		//	double *posRL = netcage[i][0]->Right->Left->WP->Position();
		//	double *posRR = netcage[i][0]->Right->Right->WP->Position();
		//	double *posRRL = netcage[i][0]->Right->Right->Left->WP->Position();
		//	double *posRRR = netcage[i][0]->Right->Right->Right->WP->Position();
		//							
		//	std::cout << "1 - Z: " << pos1[2] << "  X: " << pos1[0] << "  Y: "<< pos1[1] << std::endl;
		//	std::cout << "R - Z: " << posR[2] << "  X: " << posR[0] << "  Y: "<< posR[1] << std::endl;
		//	std::cout << "RL - Z: " << posRL[2] << "  X: " << posRL[0] << "  Y: "<< posRL[1] << std::endl;
		//	std::cout << "RR - Z: " << posRR[2] << "  X: " << posRR[0] << "  Y: "<< posRR[1] << std::endl;
		//	std::cout << "RRL - Z: " << posRRL[2] << "  X: " << posRRL[0] << "  Y: "<< posRRL[1] << std::endl;
		//	std::cout << "RRR - Z: " << posRRR[2] << "  X: " << posRRR[0] << "  Y: "<< posRRR[1] << std::endl;

		//		
		//	std::cout<< "---------------------------------------" << std::endl;
	//	}

}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		quicksort
//==================================================================================================
//
///	Sorts a vector of netnodes on ascending north position
/// 
///	
/// \param [in]		inputVector				vector of the nodes at one depth level in the cage.
///
///
/// \returns								vector of sorted nodes
///
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
std::vector<NetNodeStruct*> WpManagementSystem::quicksort(std::vector<NetNodeStruct*> inputVector){
	std::vector<NetNodeStruct*> sorted;
	if(inputVector.size()>1){
		int pivotIndex = int(inputVector.size() /2);
		double pivot = inputVector[pivotIndex]->WP->Position()[0];
		
		std::vector<NetNodeStruct*> larger,smaller;
		for (size_t j = 0;j<inputVector.size();j++){
			if (j!= pivotIndex){
				if (inputVector[j]->WP->Position()[0] <= pivot)
					smaller.push_back(inputVector[j]);
				else
					larger.push_back(inputVector[j]);
			}
		}
		smaller = quicksort(smaller);
		larger = quicksort(larger);
		
		for (size_t i =0;i<smaller.size();i++)
			sorted.push_back(smaller[i]);
		sorted.push_back(inputVector[pivotIndex]);
		for (size_t i =0;i<larger.size();i++)
			sorted.push_back(larger[i]);
		}
	else 
		sorted = inputVector;
	return sorted;		
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		findNeighboringNodes
//==================================================================================================
//
///	finds and sets the right and left neighboring nodes of each node in the vector
/// 
///	
/// \param [in]		nodes				vector of the nodes at one depth level in the cage.
///
///
/// \returns								vector of sorted nodes
///
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
void WpManagementSystem::findNeighboringNodes(std::vector<NetNodeStruct*> nodes){
	if (nodes.size() > 2){
		double ni = 0;		double ei = 0;		//position of node i
		double n1 = 0;		double e1 = 0;		//position of node i+1
		double n2 = 0;		double e2 = 0;		//position of node i+2
		double n3 = 0;		double e3 = 0;		//position of node i+3
		double dn1 = 0;		double de1 = 0;		//difference in pos for i and i+1
		double dn2 = 0;		double de2 = 0;		//difference in pos for i and i+2
		double dn3 = 0;		double de3 = 0;		//difference in pos for i and i+3
		double norm1 = 0;			double norm2 = 0;			double norm3 = 0;

		for (size_t i =0;i<nodes.size()-3;i++){
			ni = nodes[i]->WP->Position()[0];				ei = nodes[i]->WP->Position()[1];
				
			n1 = nodes[i+1]->WP->Position()[0];				e1 = nodes[i+1]->WP->Position()[1]; 
			dn1 = ni - n1;		de1 = ei - e1;
		
			n2 = nodes[i+2]->WP->Position()[0];				e2 = nodes[i+2]->WP->Position()[1];
			dn2 = ni - n2;		de2 = ei - e2;
			
			n3 = nodes[i+3]->WP->Position()[0];				e3 = nodes[i+3]->WP->Position()[1];
			dn3 = ni - n3;		de3 = ei - e3;
		
			norm1 = sqrt(pow(dn1,2) + pow(de1,2));
			norm2 = sqrt(pow(dn2,2) + pow(de2,2));
			norm3 = sqrt(pow(dn3,2) + pow(de3,2));
	
			if (i==0){
				
				if ( e1 > ei ){
					nodes[i]->Left = nodes[i+1];		nodes[i+1]->Right = nodes[i];
					nodes[i]->Right = nodes[i+2];		nodes[i+2]->Left = nodes[i];
				}
				else{
					nodes[i]->Left = nodes[i+2];		nodes[i+2]->Right = nodes[i];
					nodes[i]->Right = nodes[i+1];		nodes[i+1]->Left = nodes[i];
				}
			}
			else{
				if (norm1 < norm2 && norm1 < norm3){
					if(nodes[i]->Left == 0){
						nodes[i]->Left = nodes[i+1];
						nodes[i+1]->Right = nodes[i];
					}
					else{
						nodes[i]->Right = nodes[i+1];
						nodes[i+1]->Left = nodes[i];
					}
				}
				else if ( norm2 < norm1 && norm2 < norm3){
					if(nodes[i]->Left == 0){
						nodes[i]->Left = nodes[i+2];
						nodes[i+2]->Right = nodes[i];
					}
					else{
						nodes[i]->Right = nodes[i+2];
						nodes[i+2]->Left = nodes[i];
					}
				}
				else{
					if(nodes[i]->Left == 0){
						nodes[i]->Left = nodes[i+3];
						nodes[i+3]->Right = nodes[i];
					}
					else{
						nodes[i]->Right = nodes[i+3];
						nodes[i+3]->Left = nodes[i];
					}
				}
			}
		}
		int i = nodes.size()-3;
		ni = nodes[i]->WP->Position()[0];				ei = nodes[i]->WP->Position()[1];
			
		n1 = nodes[i+1]->WP->Position()[0];				e1 = nodes[i+1]->WP->Position()[1]; 
		dn1 = ni - n1;		de1 = ei - e1;
		int temp = i+2;
		n2 = nodes[i+2]->WP->Position()[0];				
		e2 = nodes[i+2]->WP->Position()[1];
		dn2 = ni - n2;		de2 = ei - e2;
				
		norm1 = sqrt(pow(dn1,2) + pow(de1,2));
		norm2 = sqrt(pow(dn2,2) + pow(de2,2));
	
		if (norm1 < norm2){
			if(nodes[i]->Left == 0){
				nodes[i]->Left = nodes[i+1];
				nodes[i+1]->Right = nodes[i];
			}
			else{
				nodes[i]->Right = nodes[i+1];
				nodes[i+1]->Left = nodes[i];
			}
		}
		else{
			if(nodes[i]->Left == 0){
				nodes[i]->Left = nodes[i+2];
				nodes[i+2]->Right = nodes[i];
			}
			else{
				nodes[i]->Right = nodes[i+2];
				nodes[i+2]->Left = nodes[i];
			}
		}
		i++;
		if(nodes[i]->Right == 0){
			nodes[i]->Right = nodes[i+1];
			nodes[i+1]->Left = nodes[i];
		}
		else{
			nodes[i]->Left = nodes[i+1];
			nodes[i+1]->Right = nodes[i];
		}
				
	}
	else if (nodes.size() ==2){
		nodes[0]->Right = nodes[1];
		nodes[0]->Left =  nodes[1];
		nodes[1]->Right = nodes[0];
		nodes[1]->Left =  nodes[0];
	}
	else{
		nodes[0]->Right = nodes[0];
		nodes[0]->Left =  nodes[0];
	}
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		OdeFnc
//==================================================================================================
//
///	Calculates the desired position and changes the setpoint when inside the circle of acceptance
/// The error is defined as value - desired_value.
///	
/// \param [in]		T					current simulation time
///
/// \param [in]		X					current simulation state
///
/// \param [in]		XDot				pointer to the integrand
///
/// \param [in]		bIsMajorTimeStep	is this a major time step
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
void WpManagementSystem::OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep){

	///----------------------------------------------
	///
	///			PredefinedWP
	///
	///----------------------------------------------
	if (predefinedWP){
		const double *previousWPposition;
		if (waypointIndex==0){
			atWP = true;
			previousWPposition = etaInputPort->GetPortValue(T,X);
		}
		else
			previousWPposition = wpVector[waypointIndex-1]->Position();
		
		const double *nextWPposition;
		if (waypointIndex == wpVector.size()){
			nextWPposition = wpVector[waypointIndex-1]->Position();
		}
		else
			nextWPposition = wpVector[waypointIndex]->Position();

		if (atWP){
			
			for (size_t i =0;i<6;i++)
				startEta[i] = previousWPposition[i];
			if(waypointIndex < wpVector.size()){
				etaReference[0] = nextWPposition[0];		
				etaReference[1] = nextWPposition[1];		
				etaReference[2] = nextWPposition[2];
				etaReference[3] = 0;				
				etaReference[4] = 0;
				etaReference[5] = ShortestError(etaReference[5],atan2(nextWPposition[1] - previousWPposition[1], nextWPposition[0] - previousWPposition[0]));
			}
			else
				for (size_t i= 0;i< 6;i++)
					etaReference[i] = etaReference[i];
		atWP = false;
		if (waypointIndex < wpVector.size())
			waypointIndex++;
		}
		if (!atWP){
			const double* eta = etaInputPort->GetPortValue(T,X);
			double deltaN = eta[0] - etaReference[0];
			double deltaE = eta[1] - etaReference[1];
			double deltaD = eta[2] - etaReference[2];

			bool goodHeading = abs(ShortestError(eta[5],etaReference[5])) < angleTol/180*M_PI; 
			

			if (sqrt(pow(deltaN,2) + pow(deltaE,2) + pow(deltaD,2)) < radiusOfAcceptance && goodHeading){
				//std::cout << "punkt n{\aa}dd" << std::endl;
				
				atWP = true;
			}
		}
	}
	///----------------------------------------------
	///
	///			NetCageTraversing
	///
	///----------------------------------------------
	if(netcageTraversing){
		if (atWP){
			if (T == 0)
			currentNetNode = netcage[0][0];
			if (currentNetNode == 0){
				currentNetNode = netcage[level][0];
			}
			
			
			atWP = false;
		}
		else {
			const double* eta = etaInputPort->GetPortValue(T,X);
			double deltaN = eta[0] - currentNetNode->WP->Position()[0];
		//	std::cout << "current: "<< eta[0] << " Desired: " << etaReference[0] << " CurrentWP: " << currentNetNode->WP->Position()[0] << std::endl;
			double deltaE = eta[1] - currentNetNode->WP->Position()[1];
	//		std::cout << "current: "<< eta[1] << " Desired: " << etaReference[1] << " CurrentWP: " << currentNetNode->WP->Position()[1] << std::endl;
			double deltaD = eta[2] - currentNetNode->WP->Position()[2];
		//	std::cout << "current: "<< eta[2] << " Desired: " << etaReference[2]  << " CurrentWP: " << currentNetNode->WP->Position()[2] << "--------------------------" << std::endl;
			if (sqrt(pow(deltaN,2) + pow(deltaE,2) + pow(deltaD,2)) < radiusOfAcceptance){
				if (!currentNetNode->visited){
					currentNetNode->visited = true;
					currentNetNode = currentNetNode->Left;
				}
				else{ 
					currentNetNode = netcage[++level][0];
				}
				atWP = true;
			}
		}
		etaReference[0] = currentNetNode->WP->Position()[0];
		etaReference[1] = currentNetNode->WP->Position()[1];
		etaReference[2] = currentNetNode->WP->Position()[2];
		etaReference[3] = 0;				
		etaReference[4] = 0;
		etaReference[5] = atan2(etaReference[1], etaReference[0]);
	}
	GuidanceSystem::OdeFcn(T,X,XDot,bIsMajorTimeStep);
}