////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class							WpManagementSystem
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// \brief			Waypoint management system. 
///
///					Swithes between different waypoints and sets the referencepoint of the parent class to the next wp. Takes predefined waypoints at start up, or generates mooving waypoints based on the nodes of a netstructure.
///
/// \author			Mats N{\aa}vik Hval
///
/// \par Revision history:
///					06.06.2012 MNH: Initial version.
///
////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef WPMANAGEMENTSYSTEM_H
#define WPMANAGEMENTSYSTEM_H
#include <DeepSeaGravityWaves.h>
#include <vector>
#include "Waypoint.h"
#include "GuidanceSystem.h"
#ifndef NETNODESTRUCT
#define NETNODESTRUCT
/*! \struct NetNodeStruct
    \brief A struct holding information on the different nodes in the netstructure
*/
struct NetNodeStruct{
	Waypoint* WP;			///< pointer to the moovable waypoint
	NetNodeStruct* Left;	///< pointer to the netnode to the left
	NetNodeStruct* Right;	///< pointer to the netnode to the right
	bool visited;			///< flag to tell if the node already is visited
	bool sorted;			///< flag to tell if the node has been sorted
};
#endif

class WpManagementSystem : public GuidanceSystem
{
public:
	WpManagementSystem(const string& simObjectName, ISimObjectCreator* const creator); ///< reads parameters, registers states, input/output ports and shared resources
	~WpManagementSystem();	
	void OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep); ///< returns object derivatives as a function of time, states and input ports
	void FinalSetup(const double T, const double *const X, ISimObjectCreator* const creator);	///< sets up the nodes if cage traversing is the chosen method

#ifdef FH_VISUALIZATION
	virtual void	RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator){} ///< sets up visualization resources
	virtual void	RenderUpdate(const double T, const double* const X){} ///< updates visualization resources to reflect current state
#endif
	
protected:
	std::vector<NetNodeStruct*> WpManagementSystem::quicksort(std::vector<NetNodeStruct*> inputVector);		///< method for sorting the nodes at each level.
	void WpManagementSystem::findNeighboringNodes(std::vector<NetNodeStruct*> nodes);						///< method for find the neighboring nodes of each node.
	std::string m_NetCageName;																				///< name of the netstructure
	double radiusOfAcceptance;																				///< radius of acceptance of the node
	std::vector<Waypoint*> wpVector;																		///< vector of the predefined waypoints
	std::vector<std::vector<NetNodeStruct*>> netcage;														///< vector holding the nodes in the net structure
	DeepSeaGravityWaves *m_Environment;																		///< pointer to the current environment
	bool predefinedWP;																						///< flag for predefined wps
	bool netcageTraversing;																					///< flag for netcage traversing
	bool traverseUpNdown;																					///< flag for traversing route
	bool traverseCircles;																					///< flag for traversing route
	Waypoint *currentWaypoint;																				///< current desired wp
	int waypointIndex;																						///< index of the wp
	bool atWP;																								///< flag to tell if the WP is reached
	NetNodeStruct *currentNetNode;																			///< current desired netnode
	double cageDiameter;																					///< diameter of the cage
	int level;																								///< telling which level in the cage which is interesting
	double angleTol;																						///< tolerance for the heading error
};

#endif
