#include "FluidMemoryElement.h"
#include <sstream>
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		FluidMemoryElement Constructor
//==================================================================================================
///		This constructor performs all initial setup for a Control algorithm simobject.
///		Reading in parameters, setting up communication interface i.e. output ports,
///		input ports, and states, plus additional 'one time only' resource setup.
///	
/// \param [in]		simObjectName	The name of the simobject. Used primarily 
///									by superclass constructor
///
/// \param [in]		creator			Retrive parameters. Register states, ports
///									and shared resources
///
////////////////////////////////////////////////////////////////////////////////////////////////////
FluidMemoryElement::FluidMemoryElement(const string& simObjectName, ISimObjectCreator* const creator) : SimObject(simObjectName){

	creator->GetIntParam("OrderOfSystem",&orderOfSystem);
	creator->GetIntParamArray("DOFS",dofs,2);

	Ar = new double[orderOfSystem*orderOfSystem];
	Br = new double[orderOfSystem];
	Cr = new double[orderOfSystem];

	A = Eigen::MatrixXd::Zero(orderOfSystem,orderOfSystem);
	B = Eigen::MatrixXd::Zero(orderOfSystem,1);
	C = Eigen::MatrixXd::Zero(1,orderOfSystem);
	D = Eigen::MatrixXd::Zero(1,1);
	
	creator->GetDoubleParamArray("Ar",Ar,orderOfSystem*orderOfSystem);
	creator->GetDoubleParamArray("Br",Br,orderOfSystem);
	creator->GetDoubleParamArray("Cr",Cr,orderOfSystem);
	creator->GetDoubleParamArray("Dr",&Dr,1);

	for (size_t i =0;i<orderOfSystem;i++){
		for (size_t j=0;j<orderOfSystem;j++){
			A(i,j) = Ar[i*6+j];
		}
		B(i) = Br[i];
		C(i) = Cr[i];
	}
	//const int order = orderOfSystem;


	creator->AddInport("VesselRelativeVelocity",6,&velocity);
	//creator->AddInport("VesselPosition",6,&position);

	double zero = 0;
	for (size_t i = 0;i<orderOfSystem;i++){
		std::stringstream number;number << i;
		std::string stateName = "state"+number.str();
		int index = creator->AddState(stateName,1,&zero);
		indexState.push_back(index);
	}
	outValue = 0;
	
	creator->AddOutport("FluidMemoryElement",1, PORT_FUNCTION(FluidMemoryElement::GetValue));
}
FluidMemoryElement::~FluidMemoryElement(){
}

void FluidMemoryElement::OdeFcn(const double T, const double* const X, double* const XDot, const bool bIsMajorTimeStep){

//	double depth = position->GetPortValue(T,X)[2];


	Eigen::MatrixXd xdot = Eigen::MatrixXd::Zero(orderOfSystem,1);
	Eigen::MatrixXd x = Eigen::MatrixXd::Zero(orderOfSystem,1);
	
	Eigen::MatrixXd result = Eigen::MatrixXd::Zero(1,1);
	for (size_t i =0;i<orderOfSystem;i++){
		int index = indexState[i];
		x(i) = X[index];
	}
	double vel = velocity->GetPortValue(T,X)[dofs[1]-1];
	result = A*x+B*vel;
	//result *= exp(1.0);

	xdot = A*x + B*vel;
	for (size_t i =0;i<orderOfSystem;i++){
		XDot[indexState[i]] = xdot(i);
	}

}

const double* const FluidMemoryElement::GetValue(const double T, const double* const X){
	Eigen::MatrixXd mu = Eigen::MatrixXd::Zero(orderOfSystem,1);
	outValue =0;
	for (size_t i =0;i<orderOfSystem;i++){
		mu(i) = X[indexState[i]];
		outValue += mu(i)*C(i);
	}

	return &outValue;
}

#ifdef FH_VISUALIZATION
void			FluidMemoryElement::RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator){} ///< sets up visualization resources
void			FluidMemoryElement::RenderUpdate(const double T, const double* const X){} ///< updates visualization resources to reflect current state
#endif