#ifndef _CPOTENTIAL_DATA_H_
#define _CPOTENTIAL_DATA_H_

#include <iostream>
#include <iomanip>
#include <fstream>
#include <algorithm>
#include <string>
#include <vector>
#include <complex>
#include <map>

#include <Eigen/Eigen>
#include <Eigen/Eigenvalues>
#include <Eigen/Core>

#include "FroudeKrylov.h"
#include "WaveDrift.h"

#include "Identification/FreqDomainIdent.h"

class PotentialData
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW

	struct TimeDomainRadiation
	{
		EIGEN_MAKE_ALIGNED_OPERATOR_NEW

		TimeDomainRadiation()
		{
			Speed = -10;
			Ainf  = Eigen::Matrix< double, 6, 6 >::Zero( );
			
			for( int i=0; i <6; i++)
				for( int j=0; j <6; j++)
				{
					A[i][j] = Eigen::Matrix< double, Eigen::Dynamic, Eigen::Dynamic >::Zero( 1, 1 );;
					B[i][j] = Eigen::Matrix< double, Eigen::Dynamic, Eigen::Dynamic >::Zero( 1, 1 );;
					C[i][j] = Eigen::Matrix< double, Eigen::Dynamic, Eigen::Dynamic >::Zero( 1, 1 );;
					D[i][j] = Eigen::Matrix< double, Eigen::Dynamic, Eigen::Dynamic >::Zero( 1, 1 );;
				}
		}

		// Ainf
		Eigen::Matrix<double, 6, 6> Ainf;

		double Speed;

		// companion system matrices
		Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>	A[6][6];
		Eigen::Matrix<double, Eigen::Dynamic, 1				>	B[6][6];
		Eigen::Matrix<double, 1				, Eigen::Dynamic>	C[6][6];	
		Eigen::Matrix<double, Eigen::Dynamic, 1				>	D[6][6];
		double Direction;
	};
	
	~PotentialData()
	{
		delete Speed;
		delete Heading;
		delete Frequency;
	}

	virtual void Read( const std::string &fname ) = 0;
	virtual Eigen::Matrix<double,6,6> GetHToO() = 0;

	virtual void SetOriginWorigin() = 0;
	virtual void SetOriginGorigin() = 0;
	virtual void SetOrigin( Eigen::Matrix<double,3,1> &r ) = 0;

	Eigen::Matrix<double,6,6> GetMassRigidBody()
	{
		return GetHToO().transpose()*Mrbeig*GetHToO();
	}

	Eigen::Matrix<double,6,6> GetAinf(unsigned int speedno=0, unsigned int HeadNo=0)
	{
		//return RadiationApproximation[speedno].Ainf;
		return SpeedAndDirectionRadiationApproximation[speedno][HeadNo].Ainf;
	}

	Eigen::Matrix<double,6,6> GetMassTotal(unsigned int speedno=0, unsigned int HeadNo=0 )
	{
		return SpeedAndDirectionRadiationApproximation[speedno][HeadNo].Ainf+GetHToO().transpose()*Mrbeig*GetHToO();
	}

	Eigen::Matrix<double,6,6> GetAw(unsigned int FreqNo, unsigned int SpeedNo=0, unsigned int HeadNo=0)
	{
		//return GetHToO().transpose()*A[SpeedNo][FreqNo]*GetHToO();
		return GetHToO().transpose()*AwSpeedDir[SpeedNo][HeadNo][FreqNo]*GetHToO();
	}
	
	Eigen::Matrix<double,6,6> GetBw(unsigned int FreqNo, unsigned int SpeedNo=0, unsigned int HeadNo=0)
	{
		//return GetHToO().transpose()*B[SpeedNo][FreqNo]*GetHToO();
		return GetHToO().transpose()*BwSpeedDir[SpeedNo][HeadNo][FreqNo]*GetHToO();
	}

	virtual Eigen::Matrix<double,6,6> GetCw(unsigned int FreqNo,unsigned int SpeedNo=0, unsigned int HeadNo=0)
	{
		//return GetHToO().transpose()*C[SpeedNo][FreqNo]*GetHToO();
		return GetHToO().transpose()*CwSpeedDir[SpeedNo][HeadNo][FreqNo]*GetHToO();
	}

	Eigen::Matrix<double,1,Eigen::Dynamic> GetAij(unsigned int DOF1, unsigned int DOF2, unsigned int SpeedNo=0, unsigned int HeadNo=0)
	{
		Eigen::Matrix<double,1,Eigen::Dynamic> A = Eigen::Matrix<double,1,Eigen::Dynamic>::Zero(1,NoFrequency);

		for( int i=0; i < NoFrequency; i++)
		{
			Eigen::Matrix<double,6,6> Aw = GetAw(i,SpeedNo,HeadNo);
			A(i) = Aw(DOF1,DOF2);
		}
		return A;
	}

	Eigen::Matrix<double,1,Eigen::Dynamic> GetBij(unsigned int DOF1, unsigned int DOF2, unsigned int SpeedNo=0, unsigned int HeadNo=0)
	{
		Eigen::Matrix<double,1,Eigen::Dynamic> B = Eigen::Matrix<double,1,Eigen::Dynamic>::Zero(1,NoFrequency);

		for( int i=0; i < NoFrequency; i++)
		{
			Eigen::Matrix<double,6,6> Bw = GetBw(i,SpeedNo,HeadNo);
			B(i) = Bw(DOF1,DOF2);
		}
		return B;
	}

	Eigen::Matrix<double,1,Eigen::Dynamic> GetCij(unsigned int DOF1, unsigned int DOF2, unsigned int SpeedNo=0, unsigned int HeadNo=0)
	{
		Eigen::Matrix<double,1,Eigen::Dynamic> C = Eigen::Matrix<double,1,Eigen::Dynamic>::Zero(1,NoFrequency);

		for( int i=0; i < NoFrequency; i++)
		{
			Eigen::Matrix<double,6,6> Cw = GetCw(i,SpeedNo,HeadNo);
			C(i) = Cw(DOF1,DOF2);
		}
		return C;
	}

	Eigen::Matrix<double,1,Eigen::Dynamic> GetSpeed()
	{
		Eigen::Matrix<double,1,Eigen::Dynamic> speed = Eigen::Matrix<double,1,Eigen::Dynamic>::Zero(1,NoSpeed);
		for( int i=0; i < NoSpeed; i++ ) speed(i) = Speed[i];
		return speed;
	}

	Eigen::Matrix<double,1,Eigen::Dynamic> GetHeading()
	{
		Eigen::Matrix<double,1,Eigen::Dynamic> head = Eigen::Matrix<double,1,Eigen::Dynamic>::Zero(1,NoHeading);
		for( int i=0; i < NoHeading; i++ ) head(i) = Heading[i];
		return head;
	}

	Eigen::Matrix<double,1,Eigen::Dynamic> GetFrequency()
	{
		Eigen::Matrix<double,1,Eigen::Dynamic> freq = Eigen::Matrix<double,1,Eigen::Dynamic>::Zero(1,NoFrequency);
		for( int i=0; i < NoFrequency; i++ ) freq(i) = Frequency[i];
		return freq;
	}

	virtual Eigen::Matrix<double,3,1> GetRg()
	{
		return rg;
	}

	virtual Eigen::Matrix<double,3,1> GetRo()
	{
		return ro;
	}

	Eigen::Matrix<double,6,1>  GetFirstOrderForce(	Eigen::Matrix<double,3,1> ShipPosition, double ShipSpeed,
													double WaveDirection, double WaveAmplitude, double WaveFrequency, double WaveNumber, double WavePhase,
													double Time )
	{
		Eigen::Matrix<double,6,1> Fh = Eigen::Matrix<double,6,1>::Zero();
		Eigen::Matrix<double,6,1> Fo = Eigen::Matrix<double,6,1>::Zero();

		for( int DOF=0; DOF < 6; DOF++)
		{
			double EncounterFreq = std::abs(WaveFrequency - pow(WaveFrequency,2)/9.81*ShipSpeed*cos( WaveDirection )) ;

			std::complex<double> RAO = FroudeFrylov->GetRAOValue(DOF,WaveDirection,ShipSpeed,EncounterFreq);

			double Amplitude = std::abs(RAO);
			double Phase	 = atan2( RAO.imag(), RAO.real() );

			double ThisWaveHeight = -WaveAmplitude*sin( EncounterFreq*Time -WaveNumber*ShipPosition(0)*cos(WaveDirection) -WaveNumber*ShipPosition(1)*sin(WaveDirection) + Phase + WavePhase );
			Fh(DOF) = Amplitude*ThisWaveHeight;
		}

		Fo = GetHToO().transpose()*Fh;
		return Fo;
	}
	
	Eigen::Matrix<double,6,1> GetSecondOrderForce(Eigen::Matrix<double,3,1> ShipPosition, double ShipSpeed,
		double WaveDirection, double WaveHeight, double WaveFrequency, double WaveNumber, double WavePhase,
		double Time )
	{
		Eigen::Matrix<double,6,1> Fh = Eigen::Matrix<double,6,1>::Zero();
		Eigen::Matrix<double,6,1> Fo = Eigen::Matrix<double,6,1>::Zero();

		for( int DOF=0; DOF < 6; DOF++)
		{
			if( WaveDrift )
			{
				std::complex<double> Amplitude = WaveDrift->GetRAOValue(DOF,WaveDirection,ShipSpeed,WaveFrequency);
				Fh(DOF) = std::pow(WaveHeight,2)*Amplitude.real();			
			}
		}

		Fo = GetHToO().transpose()*Fh;
		return Fo;
	}

	void GenerateZeroSpeedTimeDomainApproximation()
	{
		if( !RadiationApproximation.empty() ) RadiationApproximation.clear();
		if( ZeroSpeedIdentificationFromCache() ) return;

		FreqDomainIdent Identification;
		Identification.SetPotentialData(this);
		Identification.IdentificationMethod(FreqDomainIdent::VECTOR_FITTING);

		SpeedAndDirectionRadiationApproximation = new TimeDomainRadiation *[1];
		SpeedAndDirectionRadiationApproximation[0] = new TimeDomainRadiation[1];
		TimeDomainRadiation zeroSpeedSystem;

		for( int i=0; i < 6; i++)
		{
			for( int j=0; j <= i; j++)
			{
				if( (i == 1  || i == 3 || i == 5) && (j==0 || j==2 || j==4) ) continue; 
				if( (j == 1  || j == 3 || j == 5) && (i==0 || i==2 || i==4) ) continue; 

				std::cout << "Identifying (i/j) = (" << i << "/" << j << ") ";
				//CPrecisionTimer T; T.Start();
				FreqDomainIdent::StateSpace SS = Identification.IdentifySS(i,j);
				zeroSpeedSystem.Ainf(i,j) = Identification.GetAinf(i,j);
				zeroSpeedSystem.A[i][j] = SS.A;
				zeroSpeedSystem.B[i][j] = SS.B;
				zeroSpeedSystem.C[i][j] = SS.C;
				zeroSpeedSystem.D[i][j] = SS.D;

				zeroSpeedSystem.Ainf(j,i) = zeroSpeedSystem.Ainf(i,j);
				zeroSpeedSystem.A[j][i] = zeroSpeedSystem.A[i][j];
				zeroSpeedSystem.B[j][i] = zeroSpeedSystem.B[i][j];
				zeroSpeedSystem.C[j][i] = zeroSpeedSystem.C[i][j];
				zeroSpeedSystem.D[j][i] = zeroSpeedSystem.D[i][j];
				std::cout << "identified new data - " /*<< T.Stop() << " seconds" */<< std::endl;
			}
		}
		SpeedAndDirectionRadiationApproximation[0][0]= zeroSpeedSystem;
		ZeroSpeedIdentificaitonToCache();
	}

	void GenerateTimeDomainApproximationWithDirection()
	{
		if( IdentificationWithDirectionFromCache() ) return;

		FreqDomainIdent Identification;
		Identification.SetPotentialData(this);
		Identification.IdentificationMethod(FreqDomainIdent::VECTOR_FITTING);

		SpeedAndDirectionRadiationApproximation = new TimeDomainRadiation *[NoSpeed];

		for (int speed=0; speed < NoSpeed; speed++ )
		{
			SpeedAndDirectionRadiationApproximation[speed] = new TimeDomainRadiation[NoHeading];
			TimeDomainRadiation speedDirectionSystem;

			for( int head=0; head < NoHeading; head++)
			{
				for( int i=0; i < 6; i++)
				{
					for( int j=0; j <= i; j++)
					{
						if( (i == 1  || i == 3 || i == 5) && (j==0 || j==2 || j==4) ) continue; 
						if( (j == 1  || j == 3 || j == 5) && (i==0 || i==2 || i==4) ) continue; 

						std::cout << "Identifying (i/j) = ("<<speed<<","<<head<<")(" << i << "/" << j << ") ";
						//CPrecisionTimer T; T.Start();
						FreqDomainIdent::StateSpace SS = Identification.IdentifySS(i,j,speed,head);
						speedDirectionSystem.Ainf(i,j) = Identification.GetAinf(i,j,speed,head);
						speedDirectionSystem.A[i][j] = SS.A;
						speedDirectionSystem.B[i][j] = SS.B;
						speedDirectionSystem.C[i][j] = SS.C;
						speedDirectionSystem.D[i][j] = SS.D;

						speedDirectionSystem.Ainf(j,i) = speedDirectionSystem.Ainf(i,j);
						speedDirectionSystem.A[j][i] = speedDirectionSystem.A[i][j];
						speedDirectionSystem.B[j][i] = speedDirectionSystem.B[i][j];
						speedDirectionSystem.C[j][i] = speedDirectionSystem.C[i][j];
						speedDirectionSystem.D[j][i] = speedDirectionSystem.D[i][j];
						std::cout << "identified new data - " /*<< T.Stop() << " seconds"*/ << std::endl;
					}
				}
				SpeedAndDirectionRadiationApproximation[speed][head] = speedDirectionSystem;
			}
		}
		IdentificaitonWithDirectionToCache();
	}


	void PrintShip()
	{
		Eigen::Matrix<double,6,6> M = GetMassRigidBody();
		
		std::cout << "%Ship information form file: " << fname << std::endl;
		std::cout << "%\tLpp:\t" << Lpp << std::endl;
		std::cout << "%\tB:\t" << Bredth << std::endl;
		std::cout << "%\tT:\t" << Draught << std::endl;
		std::cout << "%\tr_cg\t" << LCG<<"\t,\t0\t,\t"<<VCG<< " (from AP and base line)" << std::endl;
		std::cout << "%\tr_cb\t" << LCB<<"\t,\t0\t,\t"<<VCB<< " (from AP and base line)" << std::endl;

		std::cout << "%\tPotential data set information: " << std::endl;
		std::cout << "%\t\tNumber of speeds:\t" <<NoSpeed<< std::endl;
		std::cout << "speed = [\t\t" ;
		for( int i=0; i < NoSpeed; i++ ) std::cout <<  Speed[i] << "\t";
		std::cout << "];" << std::endl;

		std::cout << "%\t\tNumber of headings:\t" <<NoHeading<< std::endl;
		std::cout << "head =[\t\t" ;
		for( int i=0; i < NoHeading; i++ ) std::cout <<  Heading[i] << "\t";
		std::cout << "];" << std::endl;

		std::cout << "%\t\tNumber of frequencies:\t" <<NoFrequency<< std::endl;
		std::cout << "freq =[\t\t" ;
		for( int i=0; i < NoFrequency; i++ ) std::cout <<  Frequency[i] << "\t";
		std::cout << "];" << std::endl;
		
		for( int i=0; i < NoSpeed; i++ )
		{
			std::cout << "%Total mass matrix for Speed " << i << " V="<<Speed[i] << " m/s" << std::endl;

			Eigen::Matrix<double,6,6> Ainf = GetAinf(i);//FIXME: Direction

			Eigen::Matrix<double,6,6> Mtotal = M + Ainf;
			std::cout << std::scientific;
			for( int j=0; j < 6; j++)
			{
				if( j== 0) std::cout << "M{" <<i+1<<"}\t=[";
				std::cout <<std::setprecision(4) <<"\t"<< Mtotal(j,0) << "\t," << Mtotal(j,1) << "\t," << Mtotal(j,2) << "\t," << Mtotal(j,3) << "\t," << Mtotal(j,4) << "\t,"<<Mtotal(j,5);
				if( j== 5) std::cout <<"];";
				std::cout <<";" <<std::endl;
			}

		}


		for( int i=0; i < NoSpeed; i++)
		{
			std::cout << "%Fluid memory model for Speed " << i << " V="<<Speed[i] << " m/s" << std::endl;
			auto mod = RadiationApproximation[i];

			for( int j=0; j < 6; j++ )
			{
				for( int k=0; k < 6; k++ )
				{
					int num_states = mod.A[k][j].cols();

					if( num_states == 1 )
					{
						std::cout << "A{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}" <<"=\t"  << "[" << mod.A[j][k](0,0) << "];"<< std::endl;
						std::cout << std::endl;
						std::cout << "B{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}" <<" =\t"  << "[" << mod.B[j][k](0,0) << "];"<< std::endl;
						std::cout << std::endl;
						std::cout << "C{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}" <<" =\t"  << "[" << mod.C[j][k](0,0) << "];"<< std::endl;
						std::cout << std::endl;
						std::cout << "D{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}" <<" =\t"  << "[" << mod.D[j][k](0,0) << "];"<< std::endl;
					}
					else if (num_states == 2 )
					{
						std::cout << "A{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}"<<" =\t"  << "[" << mod.A[j][k](0,0) << "\t,"<<mod.A[j][k](0,1)<<"\t;" << mod.A[j][k](1,0) << "\t,"<<mod.A[j][k](1,1)<<"\t;"<< "];"<< std::endl;
						std::cout << std::endl;
						std::cout << "B{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}"<<" =\t"  << "[" << mod.B[j][k](0,0) << "\t;" << mod.B[j][k](1,0) <<"];"<< std::endl;
						std::cout << std::endl;
						std::cout << "C{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}"<<" =\t"  << "[" << mod.C[j][k](0,0) << "\t," << mod.C[j][k](0,1) <<"];"<< std::endl;
						std::cout << std::endl;
						std::cout << "D{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}"<<" =\t"  << "[" << mod.D[j][k](0,0) <<"];"<< std::endl;
					}
					else 
					{

						std::cout << "A{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}"<<" =\t"  << "[";
						for( int s=0; s < num_states; s++)
						{
							if(s!=0) std::cout <<"   \t ";
							for( int c=0; c < num_states; c++)
							{
								std::cout << mod.A[j][k](s,c);
								if( c < (num_states-1) ) std::cout << "\t,";
							}
							if( s < (num_states-1) ) std::cout << std::endl;
						}
						std::cout << "];"<< std::endl;
						
						std::cout << std::endl;
						std::cout << "B{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}"<<" =\t"  << "[";
						for( int s=0; s < num_states; s++)
						{
							if( s != 0 )std::cout <<"   \t ";
							std::cout << mod.B[j][k](s,0);
							if( s < (num_states-1) ) std::cout << ";" <<std::endl;
						}
						std::cout << "];"<< std::endl;				
						std::cout << std::endl;

						std::cout << "C{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}"<<" =\t"  << "[";
						for( int s=0; s < num_states; s++)
						{
							std::cout << mod.C[j][k](0,s);
							if( s < (num_states-1) ) std::cout << ",";
						}
						std::cout << "];"<< std::endl;				

						std::cout << std::endl;
						std::cout << "D{"<<i+1<<"}{"<<j+1<<"}{"<<k+1<<"}"<<" =\t"  << "[" << mod.D[j][k](0,0) << "];"<< std::endl;
					}
				}
			}
		}
	}

	int GetTimeDomainStateSize( int DOF1, int DOF2, int SpeedNo=0,unsigned int HeadNo=0 )
	{
		return SpeedAndDirectionRadiationApproximation[SpeedNo][HeadNo].A[DOF1][DOF2].cols();
	}

	void ZeroSpeedIdentificaitonToCache()
	{
		std::string cachebase = fname;
		std::stringstream ss; ss << cachebase << ".identcache0";  

		std::fstream cache(ss.str(),std::ios::out|std::ios::binary);

		int a_size[2];
		int b_size[2];
		int c_size[2];
		int d_size[2];

		for( int i=0; i < 6; i++)
		{
			for( int j=0; j < 6; j++)
			{
				a_size[0] = SpeedAndDirectionRadiationApproximation[0][0].A[i][j].rows();
				a_size[1] = SpeedAndDirectionRadiationApproximation[0][0].A[i][j].cols();

				cache.write( reinterpret_cast<char*>(a_size), sizeof(a_size) );
				cache.write( reinterpret_cast<char*>(SpeedAndDirectionRadiationApproximation[0][0].A[i][j].data()),sizeof(double)*a_size[0]*a_size[1] );

				b_size[0] = SpeedAndDirectionRadiationApproximation[0][0].B[i][j].rows();
				b_size[1] = SpeedAndDirectionRadiationApproximation[0][0].B[i][j].cols();

				cache.write( reinterpret_cast<char*>(b_size), sizeof(b_size) );
				cache.write( reinterpret_cast<char*>(SpeedAndDirectionRadiationApproximation[0][0].B[i][j].data()),sizeof(double)*b_size[0]*b_size[1] );


				c_size[0] = SpeedAndDirectionRadiationApproximation[0][0].C[i][j].rows();
				c_size[1] = SpeedAndDirectionRadiationApproximation[0][0].C[i][j].cols();

				cache.write( reinterpret_cast<char*>(c_size), sizeof(c_size) );
				cache.write( reinterpret_cast<char*>(SpeedAndDirectionRadiationApproximation[0][0].C[i][j].data()),sizeof(double)*c_size[0]*c_size[1] );

				d_size[0] = SpeedAndDirectionRadiationApproximation[0][0].D[i][j].rows();
				d_size[1] = SpeedAndDirectionRadiationApproximation[0][0].D[i][j].cols();
				cache.write( reinterpret_cast<char*>(d_size), sizeof(d_size) );
				cache.write( reinterpret_cast<char*>(SpeedAndDirectionRadiationApproximation[0][0].D[i][j].data()),sizeof(double)*d_size[0]*d_size[1] );

				//std::cout << "Caching speed: " << MySpeed << " ("<<i<<","<<j<<") - " << " wrote A[ "<<a_size[0]<<" x "<<a_size[1]<<" ]"<<std::endl;
			}
		}

		cache.write( reinterpret_cast<char*>(SpeedAndDirectionRadiationApproximation[0][0].Ainf.data()),sizeof(double)*6*6 );
		cache.close();
	}

	bool ZeroSpeedIdentificationFromCache()
	{
		std::string cachebase = fname;
		std::stringstream ss; ss << cachebase <<".identcache0"; 
		std::fstream cache(ss.str(),std::ios::in|std::ios::binary);
		if( !cache.good() ) return false;

		int a_size[2];
		int b_size[2];
		int c_size[2];
		int d_size[2];

		double *A_tmp;
		double *B_tmp;
		double *C_tmp;
		double *D_tmp;
		double Ainf_tmp[6*6];
		SpeedAndDirectionRadiationApproximation = new TimeDomainRadiation *[1];
		SpeedAndDirectionRadiationApproximation[0] = new TimeDomainRadiation[1]; 
		TimeDomainRadiation radiationModel;
		for( int i=0; i < 6; i++)
		{
			for( int j=0; j < 6; j++)
			{
				cache.read( reinterpret_cast<char*>(a_size), sizeof(a_size) );
				A_tmp = new double[a_size[0]*a_size[1]];

				cache.read( reinterpret_cast<char*>(A_tmp),sizeof(double)*a_size[0]*a_size[1] );


				cache.read( reinterpret_cast<char*>(b_size), sizeof(b_size) );
				B_tmp = new double[b_size[0]*b_size[1]];

				cache.read( reinterpret_cast<char*>(B_tmp),sizeof(double)*b_size[0]*b_size[1] );

				cache.read( reinterpret_cast<char*>(c_size), sizeof(c_size) );
				C_tmp = new double[c_size[0]*c_size[1]];

				cache.read( reinterpret_cast<char*>(C_tmp),sizeof(double)*c_size[0]*c_size[1] );


				cache.read( reinterpret_cast<char*>(d_size), sizeof(d_size) );
				D_tmp = new double[d_size[0]*d_size[1]];

				cache.read( reinterpret_cast<char*>(D_tmp),sizeof(double)*d_size[0]*d_size[1] );

				//std::cout << "Recovered speed: " << MySpeed << " ("<<i<<","<<j<<") - " << " wrote A["<<a_size[0]<<"x"<<a_size[1]<<"]"<<std::endl;

				radiationModel.A[i][j] = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Zero(a_size[0],a_size[1]);
				radiationModel.B[i][j] = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Zero(b_size[0],b_size[1]);
				radiationModel.C[i][j] = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Zero(c_size[0],c_size[1]);
				radiationModel.D[i][j] = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Zero(d_size[0],d_size[1]);

				for( int k=0; k < a_size[0]; k++)
					for(int l=0; l < a_size[1]; l++)
					{
						assert(a_size[0]*a_size[1] > (l+k*a_size[1]) );
						radiationModel.A[i][j](k,l) = A_tmp[k+l*a_size[0]];
					}
				for( int k=0; k < b_size[0]; k++)
					for(int l=0; l < b_size[1]; l++)
					{
						assert(b_size[0]*b_size[1] > (l+k*b_size[1]) );
						radiationModel.B[i][j](k,l) = B_tmp[k+l*b_size[0]];
					}
					
				for( int k=0; k < c_size[0]; k++)
					for(int l=0; l < c_size[1]; l++)
					{
						assert(c_size[0]*c_size[1] > (l+k*c_size[1]) );
						radiationModel.C[i][j](k,l) = C_tmp[k+l*c_size[0]];
					}
				for( int k=0; k < d_size[0]; k++)
					for(int l=0; l < d_size[1]; l++)
					{
						assert(d_size[0]*d_size[1] > (l+k*d_size[1]) );
						radiationModel.D[i][j](k,l) = D_tmp[k+l*d_size[0]];
					}
				delete []A_tmp;
				delete []B_tmp;
				delete []C_tmp;
				delete []D_tmp;
			}
		}

		cache.read( reinterpret_cast<char*>(Ainf_tmp),sizeof(double)*6*6);

		for(int i=0; i < 6; i++)
			for(int j=0; j < 6; j++)
				radiationModel.Ainf(i,j)=Ainf_tmp[j+i*6];

		radiationModel.Speed = 0;
		SpeedAndDirectionRadiationApproximation[0][0] = radiationModel;
		cache.close();
		return true;
	}

	void IdentificaitonWithDirectionToCache()
	{
		std::string cachebase = fname;
		std::stringstream ss; ss << cachebase << ".identcache3";  

		std::fstream cache(ss.str(),std::ios::out|std::ios::binary);

		int a_size[2];
		int b_size[2];
		int c_size[2];
		int d_size[2];

		cache.write( reinterpret_cast<char*>(&NoSpeed),sizeof(unsigned int) );
		cache.write( reinterpret_cast<char*>(&NoHeading),sizeof(unsigned int) );


		for( int SpeedIdx=0; SpeedIdx < NoSpeed; SpeedIdx++)
		{
			double MySpeed = Speed[SpeedIdx];

			cache.write( reinterpret_cast<char*>(&MySpeed),sizeof(double) );

			for( int HeadIdx=0; HeadIdx < NoHeading; HeadIdx++)
			{
				double MyHeading = Heading[HeadIdx];

				cache.write( reinterpret_cast<char*>(&MyHeading),sizeof(double) );

				for( int i=0; i < 6; i++)
				{
					for( int j=0; j < 6; j++)
					{
						a_size[0] = SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].A[i][j].rows();
						a_size[1] = SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].A[i][j].cols();


						cache.write( reinterpret_cast<char*>(a_size), sizeof(a_size) );
						cache.write( reinterpret_cast<char*>(SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].A[i][j].data()),sizeof(double)*a_size[0]*a_size[1] );

						b_size[0] = SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].B[i][j].rows();
						b_size[1] = SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].B[i][j].cols();

						cache.write( reinterpret_cast<char*>(b_size), sizeof(b_size) );
						cache.write( reinterpret_cast<char*>(SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].B[i][j].data()),sizeof(double)*b_size[0]*b_size[1] );


						c_size[0] = SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].C[i][j].rows();
						c_size[1] = SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].C[i][j].cols();

						cache.write( reinterpret_cast<char*>(c_size), sizeof(c_size) );
						cache.write( reinterpret_cast<char*>(SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].C[i][j].data()),sizeof(double)*c_size[0]*c_size[1] );

						d_size[0] = SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].D[i][j].rows();
						d_size[1] = SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].D[i][j].cols();
						cache.write( reinterpret_cast<char*>(d_size), sizeof(d_size) );
						cache.write( reinterpret_cast<char*>(SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].D[i][j].data()),sizeof(double)*d_size[0]*d_size[1] );

						//std::cout << "Caching speed: " << MySpeed << " ("<<i<<","<<j<<") - " << " wrote A[ "<<a_size[0]<<" x "<<a_size[1]<<" ]"<<std::endl;
					}
				}
				cache.write( reinterpret_cast<char*>(SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx].Ainf.data()),sizeof(double)*6*6 );
			}
		}
		cache.close();
	}

	bool IdentificationWithDirectionFromCache()
	{
		std::string cachebase = fname;
		std::stringstream ss; ss << cachebase <<".identcache3"; 
		std::fstream cache(ss.str().c_str(),std::ios::in|std::ios::binary);
		if( !cache.good() ) return false;

		int a_size[2];
		int b_size[2];
		int c_size[2];
		int d_size[2];

		double *A_tmp;
		double *B_tmp;
		double *C_tmp;
		double *D_tmp;
		double Ainf_tmp[6*6];

		cache.read( reinterpret_cast<char*>(&NoSpeed),sizeof(unsigned int) );
		cache.read( reinterpret_cast<char*>(&NoHeading),sizeof(unsigned int) );

		SpeedAndDirectionRadiationApproximation = new TimeDomainRadiation *[NoSpeed];

		for( int SpeedIdx=0; SpeedIdx < NoSpeed; SpeedIdx++)
		{
			SpeedAndDirectionRadiationApproximation[SpeedIdx] = new TimeDomainRadiation[NoHeading];
			double MySpeed;

			cache.read( reinterpret_cast<char*>(&MySpeed),sizeof(double) );

			for( int HeadIdx=0; HeadIdx < NoHeading; HeadIdx++)
			{
				double MyHeading ;

				cache.read( reinterpret_cast<char*>(&MyHeading),sizeof(double) );

				TimeDomainRadiation radiationModel;
				for( int i=0; i < 6; i++)
				{
					for( int j=0; j < 6; j++)
					{
						cache.read( reinterpret_cast<char*>(a_size), sizeof(a_size) );
						A_tmp = new double[a_size[0]*a_size[1]];

						cache.read( reinterpret_cast<char*>(A_tmp),sizeof(double)*a_size[0]*a_size[1] );


						cache.read( reinterpret_cast<char*>(b_size), sizeof(b_size) );
						B_tmp = new double[b_size[0]*b_size[1]];

						cache.read( reinterpret_cast<char*>(B_tmp),sizeof(double)*b_size[0]*b_size[1] );

						cache.read( reinterpret_cast<char*>(c_size), sizeof(c_size) );
						C_tmp = new double[c_size[0]*c_size[1]];

						cache.read( reinterpret_cast<char*>(C_tmp),sizeof(double)*c_size[0]*c_size[1] );


						cache.read( reinterpret_cast<char*>(d_size), sizeof(d_size) );
						D_tmp = new double[d_size[0]*d_size[1]];

						cache.read( reinterpret_cast<char*>(D_tmp),sizeof(double)*d_size[0]*d_size[1] );

						//std::cout << "Recovered speed: " << MySpeed << " ("<<i<<","<<j<<") - " << " wrote A["<<a_size[0]<<"x"<<a_size[1]<<"]"<<std::endl;

						radiationModel.A[i][j] = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Zero(a_size[0],a_size[1]);
						radiationModel.B[i][j] = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Zero(b_size[0],b_size[1]);
						radiationModel.C[i][j] = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Zero(c_size[0],c_size[1]);
						radiationModel.D[i][j] = Eigen::Matrix<double, Eigen::Dynamic, Eigen::Dynamic>::Zero(d_size[0],d_size[1]);

						for( int k=0; k < a_size[0]; k++)
							for(int l=0; l < a_size[1]; l++)
							{
								assert(a_size[0]*a_size[1] > (l+k*a_size[1]) );
								radiationModel.A[i][j](k,l) = A_tmp[k+l*a_size[0]];
							}
							for( int k=0; k < b_size[0]; k++)
								for(int l=0; l < b_size[1]; l++)
								{
									assert(b_size[0]*b_size[1] > (l+k*b_size[1]) );
									radiationModel.B[i][j](k,l) = B_tmp[k+l*b_size[0]];
								}

								for( int k=0; k < c_size[0]; k++)
									for(int l=0; l < c_size[1]; l++)
									{
										assert(c_size[0]*c_size[1] > (l+k*c_size[1]) );
										radiationModel.C[i][j](k,l) = C_tmp[k+l*c_size[0]];
									}
									for( int k=0; k < d_size[0]; k++)
										for(int l=0; l < d_size[1]; l++)
										{
											assert(d_size[0]*d_size[1] > (l+k*d_size[1]) );
											radiationModel.D[i][j](k,l) = D_tmp[k+l*d_size[0]];
										}
										delete []A_tmp;
										delete []B_tmp;
										delete []C_tmp;
										delete []D_tmp;
					}
				}

				cache.read( reinterpret_cast<char*>(Ainf_tmp),sizeof(double)*6*6);

				for(int i=0; i < 6; i++)
					for(int j=0; j < 6; j++)
						radiationModel.Ainf(i,j)=Ainf_tmp[j+i*6];

				radiationModel.Speed = MySpeed;
				radiationModel.Direction = MyHeading;
				SpeedAndDirectionRadiationApproximation[SpeedIdx][HeadIdx]=radiationModel;
			}

		}
		cache.close();
		return true;
	}



	void RadiationDerivative(int DOF1, int DOF2, double *derivative, const double *const radiation_state, const double *ship_state, unsigned int SpeedNo=0, unsigned int HeadNo=0 )
	{
		TimeDomainRadiation& radiationModel = SpeedAndDirectionRadiationApproximation[SpeedNo][HeadNo];

		int state_size = radiationModel.A[DOF1][DOF2].cols();
		Eigen::Matrix<double,Eigen::Dynamic,1> X = Eigen::Matrix<double,Eigen::Dynamic,1>::Zero(state_size,1);
		for( int i=0; i < state_size; i++) X(i) = radiation_state[i];
		Eigen::Matrix<double,Eigen::Dynamic,1> dotX = radiationModel.A[DOF1][DOF2]*X + radiationModel.B[DOF1][DOF2]*ship_state[DOF2];
		for( int i=0; i < state_size; i++) derivative[i] = dotX(i);
	}

	double RadiationOutput(int DOF1, int DOF2, const double *const radiation_state, const double * const ship_state, unsigned int SpeedNo=0, unsigned int HeadNo=0 )
	{
		TimeDomainRadiation& radiationModel = SpeedAndDirectionRadiationApproximation[SpeedNo][HeadNo];

		int state_size = radiationModel.A[DOF1][DOF2].cols();
		Eigen::Matrix<double,Eigen::Dynamic,1> X = Eigen::Matrix<double,Eigen::Dynamic,1>::Zero(state_size,1);
		for( int i=0; i < state_size; i++) X(i) = radiation_state[i];
		Eigen::Matrix<double,Eigen::Dynamic,1> Y = radiationModel.C[DOF1][DOF2]*X + radiationModel.D[DOF1][DOF2]*ship_state[DOF2];
		return Y.value();
	}

protected:
	std::string fname;

	double Lpp,Bredth,Draught;

	double LCG,VCG;
	double LCB,VCB;


	unsigned int NoSpeed;	
	unsigned int NoHeading;
	unsigned int NoDOF;
	unsigned int NoFrequency;

	double *Speed;
	double *Heading;
	double *Frequency;

	std::map<int,TimeDomainRadiation,std::less<int>,Eigen::aligned_allocator<std::pair<int,TimeDomainRadiation> > > RadiationApproximation;

	TimeDomainRadiation **SpeedAndDirectionRadiationApproximation;

	FroudeKrylovForce	*FroudeFrylov;
	WaveDriftForce		*WaveDrift;

	Eigen::Matrix<double,6,6> Mrbeig;
	Eigen::Matrix<double,6,6> **A;
	Eigen::Matrix<double,6,6> **B;
	Eigen::Matrix<double,6,6> **C;
	Eigen::Matrix<double,6,6> **Aadd;
	Eigen::Matrix<double,6,6> **Badd;
	Eigen::Matrix<double,6,6> **Cadd;
	
	Eigen::Matrix<double,6,6> ***AwSpeedDir;
	Eigen::Matrix<double,6,6> ***BwSpeedDir;
	Eigen::Matrix<double,6,6> ***CwSpeedDir;
	Eigen::Matrix<double,6,6> ***AaddwSpeedDir;
	Eigen::Matrix<double,6,6> ***BaddwSpeedDir;
	Eigen::Matrix<double,6,6> ***CaddwSpeedDir;


	Eigen::Matrix<double,3,1> rg;
	Eigen::Matrix<double,3,1> ro;
	std::map<std::string,double> param;
};
#endif
