#ifndef POTENTIALVESSEL_H
#define POTENTIALVESSEL_H
#include "SimObject.h"
#include "Eigen\StdVector"
#include "FKForce.h"
#include "MeanDriftForce.h"
#include "ViscousForces.h"
#include <fstream>
#include <string>
#include <DeepSeaGravityWaves.h>
#include "SyntheticWave.h"
//#include <sfhGeneral.h>
//#include "WamitReader.h"
//#include "fhConstants.h"
//#include <Eigen/StdVector>

/**
\class PotentialVessel
\par User Documentation: TODO: correct this

Simulation model of a vessel in 6 degrees of freedom

\f{eqnarray*}{
\dot{\boldsymbol{\eta}}	&=& \mathbf{J}_\Theta(\boldsymbol{\eta}\boldsymbol{\nu} \\
\mathbf{M}\dot{\boldsymbol{\nu}}	&=& \sum \vec{F}
\f}

, where
- \f$\boldsymbol{\eta}\f$ is the position and attitude of the vessel in the NED frame.
- \f$\mathbf{J}_\Theta\f$ is the 6 DOF transformation matrix from body frame to NED-frame.
- \f$\boldsymbol{\nu}\f$ is the translational and rotational velocity of the vessel in the body frame
- \f$M\f$ is the mass of the vessel.
- \f$\sum \vec{F}\f$ is the sum of the forces and moments acting on the vessel.


\par Parameters
<table >
<tr><td><b>Parameter<td>Default<td>Comment</b>
<tr><td>fileName<td>-<td>filename of the vessel corresponding to the wamit output files.
</table>

\par Input ports
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>Waveforces<td>6<td>Wave induced forces and moments.
<tr><td>ViscousForces<td>6<td>Viscous forces and moments due to body velocity in a fluid.
<tr><td>FluidMemoryForce<td>6<td>Fluid memory force according to Fossen.
<tr><td>UmbilicalForce<td>6<td>Forces and moments induced by the umbilical.
<tr><td>ThrusterForce<td>6<td>Forces and moments induced by the thrusters.
<tr><td>Current<td>2<td>Current velocity
</table>

\par Output ports
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>Position<td>3<td>The position.
<tr><td>Attitude<td>3<td>The attitude
<tr><td>Velocity<td>3<td>The translational velocity.
<tr><td>RotationalVelocity<td>3<td>The rotational velocity.
<tr><td>Acceleration<td>3<td>The translational acceleration.
<tr><td>RotationalAcceleration<td>3<td>The rotational acceleration.
</table>

\par States
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>Position<td>6<td>The position.
<tr><td>Velocity<td>6<td>The velocity.
<tr><td>Acceleration<td>6<td>The acceleration.
</table>

*/


/**
\brief	Potential vessel superclass

\author	Mats Nvik Hval


\par Revision history:
	02.05.2012 MNH: Initial version
*/
class PotentialVessel : public SimObject
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW
	/**
	\brief	Reads parameters, registers states, input/output ports and shared resources

	The constructor performs all initial setup for a vessel simobject.
	Reading in parameters, setting up communication interface i.e. output ports,
	input ports, and states, plus additional 'one time only' resource setup.

	Reads in "fileName" parameter and tests for validity. 
	Registers communication interface; a force input ports, and "Velocity"/"Position"/"Acceleration" output ports.
	Registers states; "Velocity"/"Position"/"Acceleration"

	\param[in]	simObjectName	->	The name of the simobject
	\param[in]	creator			->	Retrieve parameters. Register states, ports
									and shared resources
	*/
	PotentialVessel(const string& simObjectName, ISimObjectCreator* const creator);
	
	void FinalSetup( const double T, const double *const X, ISimObjectCreator* const creator );


	/**
	\brief Computes object derivatives as a function of time, states and input ports

	Returns state derivatives. Velocity as derivative of position, 
	(external_force+gravity)/mass as derivative of velocity

	\param[in]	T				-> Current simulation time
	\param[in]	X				-> Current simulation state
	\param[out]	XDot			-> State derivatives
	\param[in]	IsMajorTimeStep	-> Is this a major time step?
	*/
	void OdeFcn(const double T , const double* const X,	double* const XDot,	const bool IsMajorTimeStep) =0; 

	
	/**
	\brief Output port. Returns vessel position
	
	See PortDefs.h

	\param[in]	T	-> Current simulation time
	\param[in]	X	-> Current simulation state
	\return			-> Address of current position
	*/
	const double* const Position(const double T, const double* const X); 
	
	/**
	\brief Output port. Returns vessel velocity
	
	See PortDefs.h

	\param[in]	T	-> Current simulation time
	\param[in]	X	-> Current simulation state
	\return			-> Address of current velocity
	*/
	const double* const VelocityBODY(const double T, const double* const X);
	const double* const NuDot(const double T, const double* const X); 
	const double* const VelocityPerturbation(const double T, const double* const X); 

	const double* const FroudeForce(const double T, const double* const X); 

	const double* const MeanWaveForce(const double T, const double* const X); 
	const double* const CurrentVel(const double T, const double* const X); 

#ifdef FH_VISUALIZATION
	/**
	\brief Sets up visualization resources

	Creates a spherical mesh and registers at visualization system
		
	\param[in]	ogreRoot	-> Register visualization resources	
	\param[in]	creator		-> Retrieve parameters.
	*/
	void RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator);

	/**
	\brief Updates visualization resources to reflect current state
		
	Updates position of mesh

	\param[in]	T	-> Current simulation time
	\param[in]	X	-> Current simulation state
	*/
	void RenderUpdate(const double T, const double* const X);
#endif

private:
	
	void initializeMembers();
	void setEnvironment();

protected:
	bool waveForces;
	bool syntheticWaveForces;
	SyntheticWave *syntWave;

	void RigidBodyCoriolisForces(const double T, const double* const X);
	void AddedMassCoriolisForces(const double T, const double* const X);
	void FluidMemoryForces(const double T, const double* const X);
	void HydroStaticRestoringForces(const double T, const double* const X);
	void FroudeKrylovForces(const double T, const double* const X);
	void MeanDriftWaveForces(const double T, const double* const X);
	void VaryingCurrentDirectionForces(const double T, const double* const X);
	void ThrusterForces(const double T, const double* const X);

	Eigen::Matrix<double,6,1> v_rigidBodyCoriolisForce;		/// Rigid body coriolis forces
	Eigen::Matrix<double,6,1> v_addedMassCoriolisForce;		/// Added mass coriolis forces
	Eigen::Matrix<double,6,1> v_fluidMemoryForces;			/// Fluid memory effect forces
	Eigen::Matrix<double,6,1> v_restoringForce;				/// Hydrostatic restoring forces
	Eigen::Matrix<double,6,1> v_FKForces;					/// Excitation forces from the waves
	Eigen::Matrix<double,6,1> v_MeanDriftForces;			/// Mean drift forces from the waves
	Eigen::Matrix<double,6,1> v_varyingCurrentDirectionForce;	/// Added mass forces due to changing current heading	
	Eigen::Matrix<double,6,1> v_thrusterForces;				/// Forces due to the thrusters
	Eigen::Matrix<double,6,1> v_SUMforces;					/// Sum of all the forces on the vessel.
	Eigen::Matrix<double,6,1> v_sumWOwaveFrequencyForces;	/// Sum of all the forces on the vessel except the wave frequency forces.
	Eigen::Matrix<double,6,1> v_otherForces;

	Eigen::Matrix<double,6,1> currentVelocity;
	Eigen::Matrix<double,6,1> nuRelative;
	Eigen::Matrix<double,3,1> nuRelative1;
	Eigen::Matrix<double,3,1> nuRelative2;
	
	Eigen::Matrix<double,6,1> nu;
	Eigen::Matrix<double,6,1> nuDot;
	Eigen::Matrix<double,6,1> eta;
	Eigen::Matrix<double,6,1> etaDot;
	Eigen::Matrix<double,6,1> nuDotMean;

		
	Eigen::Matrix<double,6,6> rigidBodyMassMatrix;			/// Rigid body mass matrix of the vessel
	Eigen::Matrix<double,6,6> rigidBodyCoriolisMatrix;		/// Rigid body coriolis matrix of the vessel
	Eigen::Matrix<double,6,6> addedmassCoriolisMatrix;		/// Added mass coriolis matrix of the vessel
	Eigen::Matrix<double,6,6> zeroFrequencyAddedMass;		/// Added mass matrix of the vessel at infinite frequency
	Eigen::Matrix<double,6,6> varyingCurrentHeadingMatrix;	/// 
	Eigen::Matrix<double,6,6> mInv;							/// inverse of the sum of rigidBodyMassMatrix and zeroFrequencyAddedMass
	
	Eigen::Matrix3d skewMatrixCG;							/// Skew symmetric matrix based on CG.

	double vesselMainDimensions[3];							/// Main dimension of the vessel
	double VOLUME;											/// Volume of the submerged body
	double RHO;												/// Water density
	double GRAVITY;											/// Gravity constant
	double CG[3];											/// Center of gravity
	double CB[3];											/// Center of bouyancy
	double W;												/// Gravity force [N]
	double B;												/// Bouyancy force [N]

	vector<double> potentialPeriods;						/// Containing the periods used by the potential analysis software

	std::vector<Eigen::Matrix<double,6,6>, Eigen::aligned_allocator< Eigen::Matrix< double, 6, 6 > >> potentialA;	/// 6DOF added mass pr frequency
	std::vector<Eigen::Matrix<double,6,6>, Eigen::aligned_allocator< Eigen::Matrix< double, 6, 6 > >> potentialB;	/// 6DOF potentialDamping pr frequency
	double		potentialC[6][6];							/// 6x6 linearized restoring matrix+
	//vector<double[6][6]> sd;
	
	FKForce			*FroudeKrylov;							/// Pointer to a froude Krylov wave force object
	MeanDriftForce	*WaveDrift;								/// Pointer to a mean wave drift force object
	vector<double> currentSpeed;							/// Vector of the speed of the current
	vector<double> currentHeading;							/// Vector of heading of the current
	vector<double> currentLayerDepth;						///	Vector of the height of the different current layers 

	DeepSeaGravityWaves *environment;						/// Pointer to the environment simobject

	double seaDepth;

	ISignalPort* m_ThrusterForces;							/// Input port. Forces and momemts from the thrusters
	ISignalPort* m_OtherForces;

	int indexEta;											/// The index of the position state in the state array
	int indexNu;											/// The index of the velocity state in the state array
	int indexNuDot;											/// The index of the acceleration state in the state array
	int indexMeanNu;										/// The index of the mean velocity, that is the velocity if the WF forces are neglected.

	std::string potentialFileName;							/// String holding the file name of the potential file
	std::string potentialFileLocation;						/// String holding the location of the potential files

	///***********************************
	///		Arrays holding return values
	///***********************************
	
	double a_position[6];
	double a_velocity[6];
	double a_acceleration[6];
	double a_sumForces[6];
	double a_currentVel[6];

	double firstOrderForces[6];
	double secondOrderForces[6];
	double nuPerturbation[6];
	
#ifdef FH_VISUALIZATION

public:
		enum RenderType
	{
		BOX,
		MESH,
	};

	void DrawVessel(Ogre::SceneNode* renderNode, Ogre::SceneManager* sceneMgr);
private:
	RenderType RenderAs;
	Ogre::SceneNode*	m_Vessel;					/// The node to which the rendered entity is attached.
	std::string meshFileName;
	std::string meshFileLocation;

#endif
};
#endif