#include "SyntheticWave.h"
#include <iostream>
#include "HelpFunctions.h"
#include <Eigen/Core>
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		SyntheticWave Constructor
//==================================================================================================
///		This constructor registers the wave frequency and heading, and calculates the wave number and perturbation amplitudes
///	
////////////////////////////////////////////////////////////////////////////////////////////////////
SyntheticWave::SyntheticWave(double waveFrequency, double waveHeading, double * amplitudeScale,bool perturbationForces):heading(waveHeading), frequency(waveFrequency),perturbationLoads(perturbationForces){
	waveNumber = pow(frequency,2) / 9.81;

	double wavePer = 2*M_PI/frequency;
	forceScale[0] = amplitudeScale[0];
	forceScale[1] = amplitudeScale[1];

	amplitude[0] = 9.81*M_PI/pow(frequency,2);
	amplitude[1] = 9.81*M_PI/ (7*pow(frequency,2));
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ForcePerturbationBODY
//==================================================================================================
//
///	Returns the perturbation calculated in the BODY frame
///	
/// /param [in]  eta	the position and attitude of the vehicle
/// /param [in] T		Current simulation time
/// \return				perturbation in body frame
////////////////////////////////////////////////////////////////////////////////////////////////////
double* SyntheticWave::ForcePerturbationBODY(const double* eta, const double T){ 

	double *PosPert = PositionPerturbation(eta,T);
	
	Eigen::Matrix<double,6,6> rotmatNED2BODY = Jmatrix(eta);
	Eigen::Matrix<double,6,1> wavePerturbationNED(PosPert);
	Eigen::Matrix<double,6,1> wavePerturbationBODYframe = rotmatNED2BODY * wavePerturbationNED;
	wavePerturbationBODYframe(0) *= forceScale[0];
	wavePerturbationBODYframe(1) *= forceScale[1];
	for (size_t i =0;i<6;i++){
		forcePerturbation[i] = wavePerturbationBODYframe(i);
	}

	return forcePerturbation;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		PositionPerturbation
//==================================================================================================
//
///	Returns the perturbation calculated in the NED frame
///	
/// /param [in]  eta	the position and attitude of the vehicle
/// /param [in] T		Current simulation time
/// \return				perturbation in NED frame
////////////////////////////////////////////////////////////////////////////////////////////////////
double* SyntheticWave::PositionPerturbation(const double* eta, const double T){ 
	
	double depthScale = exp(-waveNumber*eta[2]);
	double waveHeave = amplitude[1] * sin(waveNumber*eta[2] - frequency*T)*depthScale;
	double waveSurge = amplitude[0] * sin(waveNumber*eta[0] - frequency*T)*depthScale;

	Eigen::Matrix<double,3,3> rotmatWAVE2NED = RotationMatrix(Eigen::Vector3d(0,0,heading));
	Eigen::Matrix<double,3,1> wavePerturbationWAVEframe(waveSurge,0,waveHeave);
	Eigen::Matrix<double,3,1> wavePerturbationNEDframe = rotmatWAVE2NED * wavePerturbationWAVEframe;

	for (size_t i = 0;i<3;i++){
		positionPerturbation[i] = wavePerturbationNEDframe(i);
		positionPerturbation[i+3] = 0;
	}
	return positionPerturbation;

}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		VelocityPerturbation
//==================================================================================================
//
///	Returns the velocity perturbation calculated in the BODY frame
///	
/// /param [in]  eta	the position and attitude of the vehicle
/// /param [in] T		Current simulation time
/// \return				perturbation
////////////////////////////////////////////////////////////////////////////////////////////////////
double* SyntheticWave::VelocityPerturbation(const double* eta, const double T){ 
	
	double depthScale = exp(-waveNumber*eta[2]);
	double waveHeave = amplitude[1] * -frequency*cos(waveNumber*eta[2] - frequency*T)*depthScale;
	double waveSurge = amplitude[0] * -frequency*cos(waveNumber*eta[0] - frequency*T)*depthScale;

	Eigen::Matrix<double,3,3> rotmatWAVE2NED = RotationMatrix(Eigen::Vector3d(0,0,heading));
	Eigen::Matrix<double,3,1> wavePerturbationWAVEframe(waveSurge,0,waveHeave);
	Eigen::Matrix<double,3,1> wavePerturbationNEDframe = rotmatWAVE2NED * wavePerturbationWAVEframe;
	Eigen::Matrix<double,3,3> rotmatNED2BODY = RotationMatrix(Eigen::Vector3d(eta[3],eta[4],eta[5]));
	Eigen::Matrix<double,3,1> wavePerturbationBODYframe = rotmatNED2BODY * wavePerturbationNEDframe;
	for (size_t i = 0;i<3;i++){
		velocityPerturbation[i] = wavePerturbationBODYframe(i);
		velocityPerturbation[i+3] = 0;
	}
	return velocityPerturbation;

}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		AccelerationPerturbation
//==================================================================================================
//
///	Returns the acceleration perturbation calculated in the BODY frame
///	
/// /param [in]  eta	the position and attitude of the vehicle
/// /param [in] T		Current simulation time
/// \return				perturbation
////////////////////////////////////////////////////////////////////////////////////////////////////
double* SyntheticWave::AccelerationPerturbation(const double* eta, const double T){ 
	
	double depthScale = exp(-waveNumber*eta[2]);
	double waveHeave = amplitude[1] * -pow(frequency,2)*sin(waveNumber*eta[2] - frequency*T)*depthScale;
	double waveSurge = amplitude[0] * -pow(frequency,2)*sin(waveNumber*eta[0] - frequency*T)*depthScale;

	Eigen::Matrix<double,3,3> rotmatWAVE2NED = RotationMatrix(Eigen::Vector3d(0,0,heading));
	Eigen::Matrix<double,3,1> wavePerturbationWAVEframe(waveSurge,0,waveHeave);
	Eigen::Matrix<double,3,1> wavePerturbationNEDframe = rotmatWAVE2NED * wavePerturbationWAVEframe;
	Eigen::Matrix<double,3,3> rotmatNED2BODY = RotationMatrix(Eigen::Vector3d(eta[3],eta[4],eta[5]));
	Eigen::Matrix<double,3,1> wavePerturbationBODYframe = rotmatNED2BODY * wavePerturbationNEDframe;
	for (size_t i = 0;i<3;i++){
		accelerationPerturbation[i] = wavePerturbationBODYframe(i);
		accelerationPerturbation[i+3] = 0;
	}
	return accelerationPerturbation;

}
bool SyntheticWave::ForceBool(){

	return perturbationLoads;
}