////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class							SyntheticWave
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// \brief			Simple synthetic wave object
///
///			A simple sine wave with an amplitude in surge and an amplitude in heave with regards to the origin. Assuming linear wave theory, the amplitudes are found from the the frequency under the assumption of a breaking wave.
///			The amplitudes can be scaled in surge and heave such that the sine wave as an alternativ can be used as wave forces in body frame. Assuming infinite water, the amplitudes decreases as the vessel depth increses.
///
/// \author			Mats N{\aa}vik Hval
/// \par Revision history:
///					06.06.2011 MNH: Initial version.
///
////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef SYNTHETICWAVE_H
#define SYNTHETICWAVE_H

class SyntheticWave
{
public:
	SyntheticWave(double waveFrequency, double waveHeading, double * amplitudeScale,bool perturbationForces);	///< Constructor setting up the wave
	double* ForcePerturbationBODY(const double* eta, const double T); ///< Returns the force perturbation in the BODY frame
	double* PositionPerturbation(const double* eta, const double T); ///< Returns the position perturbation in the NED-frame
	double* VelocityPerturbation(const double* eta, const double T); ///< Returns the velocity perturbation in the BODY-frame
	double* AccelerationPerturbation(const double* eta, const double T); ///< Returns the acceleration perturbation in the BODY-frame

	bool	ForceBool(); ///< Returns if the perturbations are loads
	
private: 
	double positionPerturbation[6];			///< NED perturbation
	double velocityPerturbation[6];			///< BODY velocity perturbation
	double accelerationPerturbation[6];		///< BODY acceleration perturbation
	double forcePerturbation[6];			///< BODY perturbation
	double heading;						///< Heading of the wave
	double frequency;					///< Waypoint position as an array
	double amplitude[2];				///< Amplitude of the wave found from breaking wave assumption
	double forceScale[2];				///< Scaling coefficients for the perturbation amplitudes
	double *positionE;					///< Waypoint position as an array
	double *positionD;					///< Waypoint position as an array
	double waveNumber;					///< Wave number assuming infinite depth
	bool perturbationLoads;

};
#endif

