////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class							Thruster
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// \author			Mats Nvik Hval
///
///	Thrusterobject, fixed or rotatable. Fixedpitch.
///
/// \par Revision history:
///					06.05.2012 MNH: Initial version.
///
////////////////////////////////////////////////////////////////////////////////////////////////////
#include "Thruster.h"
using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		Thruster Constructor
//==================================================================================================
///		This constructor performs all initial setup for the thruster simobject.
///		Reading in parameters, setting up communication interface i.e. output ports,
///		input ports, and states, plus additional 'one time only' resource setup.
///	
/// \param [in]		simObjectName	The name of the simobject. Used primarily 
///									by superclass constructor
///
/// \param [in]		creator			Retrieve parameters. Register states, ports
///									and shared resources
///
////////////////////////////////////////////////////////////////////////////////////////////////////
Thruster::Thruster(const string& simObjectName, ISimObjectCreator* const creator) : SimObject(simObjectName)
{
	RPM2thrust = false;
	RPM2thrustCoeff = false;
	RPM2torque = false;
	RPM2torqueCoeff = false;
	torqueControlled = false;
	speedControlled = false;

	creator->GetStringParam("BasisForControl",basisForControl);
	if (basisForControl.compare("speed") == 0){
		speedControlled = true;
		creator->AddInport	("RPMinput",	1,&RPMinput);
		}
	else if (basisForControl.compare("torque") == 0){
		torqueControlled = true;
		creator->AddInport	("TorqueInput",	1,&torqueInput);
		}
	else
		creator->ReportParameterError("BasisForControl", "Must be 'speed' or 'torque', written in lowercase .");
	
	double zero3[] = {0, 0, 0};
	double zero2[] = {0, 0};
	
		creator->GetDoubleParam("PropellerReverseThrustEfficiency",&propellerReverseThrustEfficiency,0.5);
		creator->GetDoubleParam("PropellerReverseTorqueEfficiency",&propellerReverseTorqueEfficiency,0.5);
		creator->GetStringParam("BasisForThrustCalculation",basisForThrustCalculation);
	
		if (basisForThrustCalculation.compare("RPM_TO_THRUST") == 0){
			
			creator->GetDoubleParamArray("RPM_TO_THRUST_Q1",curveCoeff_n2T_Q1,3);
			n2T.push_back(curveCoeff_n2T_Q1);
			creator->GetDoubleParamArray("RPM_TO_THRUST_Q2",curveCoeff_n2T_Q2,3,curveCoeff_n2T_Q1);
			n2T.push_back(curveCoeff_n2T_Q2);
			creator->GetDoubleParamArray("RPM_TO_THRUST_Q3",curveCoeff_n2T_Q3,3,curveCoeff_n2T_Q1);
			n2T.push_back(curveCoeff_n2T_Q3);
			creator->GetDoubleParamArray("RPM_TO_THRUST_Q4",curveCoeff_n2T_Q4,3,curveCoeff_n2T_Q3);
			n2T.push_back(curveCoeff_n2T_Q4);

			creator->GetDoubleParam("PropellerReverseThrustEfficiency",&propellerReverseThrustEfficiency,0.5);
			for (unsigned int i = 1;i<4;i++){
				if(n2T[i][0] == 0 && n2T[i][1] == 0 && n2T[i][2] == 0) 
					for (unsigned int j = 0;j<3;j++)
						n2T[i][j] = n2T[0][j]*propellerReverseThrustEfficiency;
			}
			RPM2thrust = true;
		}
		else if (basisForThrustCalculation.compare("RPM_TO_THRUST_COEFFICIENT") == 0){
			creator->GetDoubleParamArray("RPM_TO_THRUST_COEFFICIENT_Q1",curveCoeff_n2CT_Q1,3);
			n2CT.push_back(curveCoeff_n2CT_Q1);
			creator->GetDoubleParamArray("RPM_TO_THRUST_COEFFICIENT_Q2",curveCoeff_n2CT_Q2,3,curveCoeff_n2CT_Q1);
			n2CT.push_back(curveCoeff_n2CT_Q2);
			creator->GetDoubleParamArray("RPM_TO_THRUST_COEFFICIENT_Q3",curveCoeff_n2CT_Q3,3,curveCoeff_n2CT_Q1);
			n2CT.push_back(curveCoeff_n2CT_Q3);
			creator->GetDoubleParamArray("RPM_TO_THRUST_COEFFICIENT_Q4",curveCoeff_n2CT_Q4,3,curveCoeff_n2CT_Q3);
			n2CT.push_back(curveCoeff_n2CT_Q4);
			
			for (unsigned int i = 1;i<4;i++){
				if(n2CT[i][0] == 0 && n2CT[i][1] == 0 && n2CT[i][2] == 0) 
					for (unsigned int j = 0;j<3;j++)
						n2CT[i][j] = n2CT[0][j]*propellerReverseThrustEfficiency;
			}
			
			RPM2thrustCoeff = true;
		}
		else {
		creator->ReportParameterError("BasisForThrustCalculation", "Must be RPM_TO_THRUST or RPM_TO_THRUST_COEFFICIENT.");		
		}


		creator->GetStringParam("BasisForTorqueCalculation",basisForTorqueCalculation);
		if (basisForTorqueCalculation.compare("RPM_TO_TORQUE") == 0){
			creator->GetDoubleParamArray("RPM_TO_TORQUE_Q1",curveCoeff_n2Q_Q1,3);
			n2Q.push_back(curveCoeff_n2Q_Q1);
			creator->GetDoubleParamArray("RPM_TO_TORQUE_Q2",curveCoeff_n2Q_Q2,3,curveCoeff_n2Q_Q1);
			n2Q.push_back(curveCoeff_n2Q_Q2);
			creator->GetDoubleParamArray("RPM_TO_TORQUE_Q3",curveCoeff_n2Q_Q3,3,curveCoeff_n2Q_Q1);
			n2Q.push_back(curveCoeff_n2Q_Q3);
			creator->GetDoubleParamArray("RPM_TO_TORQUE_Q4",curveCoeff_n2Q_Q4,3,curveCoeff_n2Q_Q3);
			n2Q.push_back(curveCoeff_n2Q_Q3);
			
			for (unsigned int i = 1;i<4;i++){
				if(n2Q[i][0] == 0 && n2Q[i][1] == 0 && n2Q[i][2] == 0) 
					for (unsigned int j = 0;j<3;j++)
						n2Q[i][j] = n2Q[0][j]*propellerReverseTorqueEfficiency;
			}
			RPM2torque = true;
		}
		else if (basisForTorqueCalculation.compare("RPM_TO_TORQUE_COEFFICIENT") == 0){
			creator->GetDoubleParamArray("RPM_TO_TORQUE_COEFFICIENT_Q1",curveCoeff_n2CQ_Q1,3);
			n2CQ.push_back(curveCoeff_n2CQ_Q1);
			creator->GetDoubleParamArray("RPM_TO_TORQUE_COEFFICIENT_Q2",curveCoeff_n2CQ_Q2,3,curveCoeff_n2CQ_Q1);
			n2CQ.push_back(curveCoeff_n2CQ_Q2);
			creator->GetDoubleParamArray("RPM_TO_TORQUE_COEFFICIENT_Q3",curveCoeff_n2CQ_Q3,3,curveCoeff_n2CQ_Q1);
			n2CQ.push_back(curveCoeff_n2CQ_Q3);
			creator->GetDoubleParamArray("RPM_TO_TORQUE_COEFFICIENT_Q4",curveCoeff_n2CQ_Q4,3,curveCoeff_n2CQ_Q3);
			n2CQ.push_back(curveCoeff_n2CQ_Q4);
			
			for (unsigned int i = 1;i<4;i++){
				if(n2CQ[i][0] == 0 && n2CQ[i][1] == 0 && n2CQ[i][2] == 0) 
					for (unsigned int j = 0;j<3;j++)
						n2CQ[i][j] = n2CQ[0][j]*propellerReverseTorqueEfficiency;
			}
			
			RPM2torqueCoeff = true;
		}
		else {
		creator->ReportParameterError("BasisForTorqueCalculation", "Must be RPM_TO_TORQUE or RPM_TO_TORQUE_COEFFICIENT.");	
		}
	
	creator->GetDoubleParamArray("PositionWRT_CenterOfGravity",positionWRTcg,3,zero3);		/// Position of the thruster given in body coordinates with regard to CG
	creator->GetDoubleParamArray("Attitude",attitude,2,zero2);								/// Attitude of thruster given as spherical coordinates, phi and theta. Given in degrees where phi = 0 is along the vessel body x-axis in the x-y-plane and theta=0 is between the vessel body x-y-plane and the vector.
	double temp;
	for (size_t i = 0;i<2;i++){
		temp = attitude[i] * M_PI / 180;
		attitude[i] = temp;
	}
	creator->GetStringParam("ThrusterType",thrusterType,"Fixed");
	if (thrusterType.compare("Rotatable") == 0) {
		creator->GetDoubleParamArray("ThrusterRotation_MinMax_Theta",minMaxRotationTheta,2);	
		minMaxRotationTheta[0] = minMaxRotationTheta[0] *M_PI / 180;
		minMaxRotationTheta[1] = minMaxRotationTheta[1] *M_PI / 180;
		/// In case of rotatable thruster, maximum angle phi
	}

	creator->GetDoubleParam("PropellerDiameter",&propellerDiameter);
	//creator->GetDoubleParam("PropellerAdvanceNumber",&propellerAdvanceNumber,0);
	/*creator->GetDoubleParam("WakeFractionNumber",&wakeFractionNumber,0);
	creator->GetDoubleParam("OpenWaterEfficiency",&efficiencyOpenWater,1);				
	creator->GetDoubleParam("RelativeRotativeEfficiency",&efficiencyRelativeRotative,1);
	creator->GetDoubleParam("MechanicalEfficiency",&efficiencyMechanical,1);
	creator->GetDoubleParam("HullEfficiency",&efficiencyHull,1);

	creator->GetDoubleParam("ThrustDeductionFactor",&thrustDeductionFactor,0);
	*/
	
	creator->GetDoubleParam("Rho",&rho,1000);
	creator->GetIntParam("MinRPM",&minRPM,-10000);
	creator->GetIntParam("MaxRPM",&maxRPM,10000);
	creator->GetDoubleParam("ThrustlossFactor",&thrusterEfficiency);
	thrusterEfficiency = 1-thrusterEfficiency;

//	creator->AddInport	("RPM",	1,&RPM);
	creator->AddInport ("VesselVelocityBODY",6,&bodyVelocity);
	creator->AddInport ("VesselPositionNED",6,&vesselPosition);
	
	creator->AddOutport("ForcesNED",6,PORT_FUNCTION(Thruster::ForcesNED)); // Output port. Force
	creator->AddOutport("ForcesBODY",6,PORT_FUNCTION(Thruster::ForcesBODY)); // Output port. Force
	creator->AddOutport("TorqueLoad",1,PORT_FUNCTION(Thruster::TorqueLoad)); // Output port. Force
	creator->AddOutport("ThrustAxial",1,PORT_FUNCTION(Thruster::ThrustAxial)); // Output port. Force
	creator->AddOutport("RPM",1,PORT_FUNCTION(Thruster::RPM)); // Output port. Force
	
	double zero = 0;
	indexRPMstate = creator->AddState("RPM",1,&zero);
	
	creator->GetDoubleParam("RPM_timeconstant",&RPMtimeconstant);
	
	iRPM = 0;
	deltaTheta = 0;
	
}
	
void Thruster::FinalSetup(const double T,const double* const X,  ISimObjectCreator* const creator){
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ForcesBODY
//==================================================================================================
//
///	Returns the forces and moments generated by the thruster on the vessel in the body frame
///	
/// \param [in]		T						current simulation time
///
/// \param [in]		X						current simulation state
///
/// \param [out]	forcesBODYframe			forces and moments generated by the thruster
///
////////////////////////////////////////////////////////////////////////////////////////////////////

const double* const Thruster::ForcesBODY(const double T, const double* const X){
	
	const double* const thrustArray = ThrustAxial(T,X);
	
	Eigen::Matrix3d rotMat1 = RotationMatrix(Eigen::Vector3d(0,attitude[1],attitude[0]+deltaTheta));
	Eigen::Vector3d thrustVector;
	thrustVector << thrustArray[0],0,0;
	
	Eigen::Vector3d thrustLocal = rotMat1 * thrustVector;

	double thrust[6];
	for (size_t i =0;i<3;i++)
		thrust[i] = thrustLocal(i);

	double a_thrustLocal[3];

	a_thrustLocal[0] = thrustVector(0) * cos(attitude[0]+deltaTheta) * sin(attitude[1]);
	a_thrustLocal[1] = thrustVector(0) * sin(attitude[0]+deltaTheta) * sin(attitude[1]);
	a_thrustLocal[2] = thrustVector(0) * cos(attitude[1]); 

	forcesBODYframe[0] = a_thrustLocal[0];
	forcesBODYframe[1] = a_thrustLocal[1];
	forcesBODYframe[2] = a_thrustLocal[2];
	forcesBODYframe[3] = -forcesBODYframe[1]*positionWRTcg[2] + forcesBODYframe[2]*positionWRTcg[1];
	forcesBODYframe[4] = -forcesBODYframe[2]*positionWRTcg[0] + forcesBODYframe[0]*positionWRTcg[2];
	forcesBODYframe[5] = -forcesBODYframe[0]*positionWRTcg[1] + forcesBODYframe[1]*positionWRTcg[0];
	/*
	Thr(0,i) = cos(thrusters.back().positionAndAttitude[3]*M_PI/180)*sin(thrusters.back().positionAndAttitude[4]*M_PI/180);
	Thr(1,i) = sin(thrusters.back().positionAndAttitude[3]*M_PI/180)*sin(thrusters.back().positionAndAttitude[4]*M_PI/180);
	Thr(2,i) = cos(thrusters.back().positionAndAttitude[4]*M_PI/180);
	Thr(3,i) = Thr(2,i)*thrusters.back().positionAndAttitude[1] - Thr(1,i)*thrusters.back().positionAndAttitude[2];
	Thr(4,i) = Thr(0,i)*thrusters.back().positionAndAttitude[2] - Thr(2,i)*thrusters.back().positionAndAttitude[0];
	Thr(5,i) = Thr(1,i)*thrusters.back().positionAndAttitude[0] - Thr(0,i)*thrusters.back().positionAndAttitude[1];*/


	return forcesBODYframe;
	
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ForcesNED
//==================================================================================================
//
///	Returns the thrust forces in the NED frame 
///	
/// \param [in]		T					current simulation time
///
/// \param [in]		X					current simulation state
///
/// \param [out]	thrust				Thrust in the local body x-y-z-system
///
////////////////////////////////////////////////////////////////////////////////////////////////////
const double* const Thruster::ForcesNED(const double T, const double* const X){

	const double* const a_thrustBODY = ForcesBODY(T,X);
	Eigen::Matrix<double,6,1> v_thrustBody;
	for (size_t i = 0;i<6;i++)
		v_thrustBody(i) = a_thrustBODY[i];

	const double* eta = vesselPosition->GetPortValue(T,X);

	Eigen::Matrix<double,6,6> JMat = Jmatrix(eta);
	
	Eigen::Matrix<double,6,1> forcesNED = JMat*v_thrustBody;

	for (size_t i =0;i<6;i++)
		forcesNEDframe[i] = forcesNED(i);
	
return forcesNEDframe;
}

const double* const Thruster::ThrustAxial(const double T, const double* const X){
	/// TODO :  implement for 2. and 4. quadrant, Implement ability for torque control

	if (speedControlled){
		if (iRPM >= 0)
			quadrantOfOperation = 1;
		else
			quadrantOfOperation = 3;
		if (RPM2thrust) {
			double coeff1 = n2T[quadrantOfOperation][0];
			double coeff2 = n2T[quadrantOfOperation][1];
			double coeff3 = n2T[quadrantOfOperation][2];
			int positiveRPM = (iRPM>=0)?1:-1;

			thrust = thrusterEfficiency * positiveRPM *  coeff1*pow(double(iRPM),2) + coeff2*iRPM;// + n2T[quadrantOfOperation][2];
		}
		if (RPM2thrustCoeff) {
			double CT = (n2CT[quadrantOfOperation][0]*pow(double(iRPM),2) + n2CT[quadrantOfOperation][1]*iRPM);// + n2T[quadrantOfOperation][2]);
			thrust = (iRPM>=0)?1:-1 * CT * rho * pow(double(iRPM),2) * pow(propellerDiameter,4);
		}
	}
	else if (torqueControlled){
	}
	else{
		thrust = 0;
	}
return &thrust;
}

const double* const Thruster::TorqueLoad(const double T, const double* const X){

	if (RPM2torque) {
		torque = n2Q[quadrantOfOperation][0]*pow(double(iRPM),2) + n2Q[quadrantOfOperation][1]*iRPM + n2Q[quadrantOfOperation][2];
	}
	if (RPM2torqueCoeff) {
		double CQ = (n2CQ[quadrantOfOperation][0]*pow(double(iRPM),2) + n2CQ[quadrantOfOperation][1]*iRPM + n2CQ[quadrantOfOperation][2]);
		torque = CQ * rho * pow(double(iRPM),2) * pow(propellerDiameter,5);
	}
	else torque = 0;
return &torque;
}

const double* const Thruster::RPM(const double T, const double* const X){
	return X+indexRPMstate;
}
void Thruster::OdeFcn(const double T , const double* const X,	double* const XDot,	const bool IsMajorTimeStep){
	//TODO : implement check for inflow velocity to decide quadrant of operation
	if ((iRPM) > maxRPM) {
		iRPM = maxRPM;
	}
	if ((iRPM) < minRPM) {
		iRPM = minRPM;
	}

	
	if (iRPM >=0){
		quadrantOfOperation = 1;
	}
	else{
		quadrantOfOperation = 3;
		}


	if (torqueControlled){
		torque = torqueInput->GetPortValue(T,X)[0];
		complex<double> sqrtTerm = sqrt(pow(n2Q[quadrantOfOperation][1],2) - 4*n2Q[quadrantOfOperation][0]*(n2Q[quadrantOfOperation][2]-torque));
		iRPM = (int)(-n2Q[quadrantOfOperation][1] + real(sqrtTerm))*(quadrantOfOperation==1 || quadrantOfOperation==2)?1:-1;
	}
	if(speedControlled)
		iRPM = (int)RPMinput->GetPortValue(T,X)[0];

	XDot[indexRPMstate] = -1/RPMtimeconstant * ( X[indexRPMstate] - iRPM);
}


#ifdef FH_VISUALIZATION
void Thruster::RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator)
{
}
void Thruster::RenderUpdate(const double T, const double* const X)
{
}
#endif