#ifndef THRUSTER
#define THRUSTER

#include "SimObject.h"
#include "Eigen\Core"
#include <string>
#include <math.h>
#include <assert.h>
#include "HelpFunctions.h"
#include <sfhGeneral.h>

class Thruster : public SimObject
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW
	/**
	\brief	Reads parameters, registers states, input/output ports and shared resources

	The constructor performs all initial setup for a vessel simobject.
	Reading in parameters, setting up communication interface i.e. output ports,
	input ports, and states, plus additional 'one time only' resource setup.

	Reads in "fileName" parameter and tests for validity. 
	Registers communication interface; a force input ports, and "Velocity"/"Position"/"Acceleration" output ports.
	Registers states; "Velocity"/"Position"/"Acceleration"

	\param[in]	simObjectName	->	The name of the simobject
	\param[in]	creator			->	Retrieve parameters. Register states, ports
									and shared resources
	*/
	Thruster(const string& simObjectName, ISimObjectCreator* const creator);
	//const double * const Forces();
	
	void FinalSetup( const double T, const double *const X, ISimObjectCreator* const creator );

	/**
	\brief Computes object derivatives as a function of time, states and input ports

	Returns state derivatives. Velocity as derivative of position, 
	(external_force+gravity)/mass as derivative of velocity

	\param[in]	T				-> Current simulation time
	\param[in]	X				-> Current simulation state
	\param[out]	XDot			-> State derivatives
	\param[in]	IsMajorTimeStep	-> Is this a major time step?
	*/
	void OdeFcn(const double T , const double* const X,	double* const XDot,	const bool IsMajorTimeStep); 

#ifdef FH_VISUALIZATION
	/**
	\brief Sets up visualization resources

	Creates a spherical mesh and registers at visualization system
		
	\param[in]	ogreRoot	-> Register visualization resources	
	\param[in]	creator		-> Retrieve parameters.
	*/
	void RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator);

	/**
	\brief Updates visualization resources to reflect current state
		
	Updates position of mesh

	\param[in]	T	-> Current simulation time
	\param[in]	X	-> Current simulation state
	*/
	void RenderUpdate(const double T, const double* const X);
#endif

	const double* const ForcesNED(const double T, const double* const X); 
	const double* const ForcesBODY(const double T, const double* const X); 
	const double* const TorqueLoad(const double T, const double* const X); 
	const double* const RPM(const double T, const double* const X); 
	const double* const ThrustAxial(const double T, const double* const X); 


	
protected:


private:
	//const double* const ThrustLocal(const double T, const double* const X); 
//	const double* const ThrustAxial(const double T, const double* const X); 

	ISignalPort* RPMinput;						/// RPM control input to the thruster
	ISignalPort* torqueInput;					/// Torque control input to the thruster
	ISignalPort* bodyVelocity;					/// Vessel 6DOF velocities input
	ISignalPort* vesselPosition;				/// Position and attitude of the vessel(6DOF)
	double iRPM;									
	unsigned int quadrantOfOperation;			/// For multiple-quadrant operation.
	int deltaTheta;								/// index of the rotation of the thruster for rotatable thrusters. 
	int indexRPMstate;

	string thrusterType;
	string basisForThrustCalculation;
	string basisForTorqueCalculation;
	
	double curveCoeff_n2T_Q1[3];
	double curveCoeff_n2T_Q2[3];
	double curveCoeff_n2T_Q3[3];
	double curveCoeff_n2T_Q4[3];

	vector<double*> n2T;
	
	double curveCoeff_n2CT_Q1[3];
	double curveCoeff_n2CT_Q2[3];
	double curveCoeff_n2CT_Q3[3];
	double curveCoeff_n2CT_Q4[3];

	vector<double*> n2CT;

	double curveCoeff_n2Q_Q1[3];
	double curveCoeff_n2Q_Q2[3];
	double curveCoeff_n2Q_Q3[3];
	double curveCoeff_n2Q_Q4[3];
	
	vector<double*> n2Q;

	double curveCoeff_n2CQ_Q1[3];
	double curveCoeff_n2CQ_Q2[3];
	double curveCoeff_n2CQ_Q3[3];
	double curveCoeff_n2CQ_Q4[3];
		
	vector<double*> n2CQ;

	bool RPM2thrust;
	bool RPM2thrustCoeff;
	bool RPM2torque;
	bool RPM2torqueCoeff;
		
	double propellerDiameter;				
	
	int maxRPM;	
	int minRPM;
	double rho;
	double RPMtimeconstant;

	double thrusterEfficiency;				
	double propellerReverseThrustEfficiency;
	double propellerReverseTorqueEfficiency;
		
	double positionWRTcg[3];					/// Position of the thruster given in body coordinates with regard to CG
	double attitude[2];							/// Attitude of thruster given as spherical coordinates, theta and phi. Theta = 0 is along the vessel body x-axis in the x-y-plane and phi=0 is along the vessel body negative z-axis (up).
	double minMaxRotationTheta[2];				/// In case of rotatable thruster, minimum and maximum angle theta,
	double forcesBODYframe[6];
	double forcesNEDframe[6];
	double thrust;
	double torque;

	string basisForControl;
	bool speedControlled;
	bool torqueControlled;


#ifdef FH_VISUALIZATION
	Ogre::SceneNode*	m_RenderNode;			/// The node to which the rendered entity is attached.
	string				meshModel;				/// Name of Ogre material used for visualization
#endif
};
#endif