#include "ViscousForce.h"
using namespace Eigen;
using namespace std;

////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class							ViscousForces
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// \brief			Generates the linear and quadratic viscous forces
///
///					
/// \author			Mats Nvik Hval
///
///	\par
///					01.03.2012 MNH: Initial version
///
////////////////////////////////////////////////////////////////////////////////////////////////////


ViscousForce::ViscousForce(const string& simObjectName, ISimObjectCreator* const creator):SimObject(simObjectName)
{
	double zeroArray[] = {0,0,0,0,0,0,0};
	creator->GetDoubleParamArray("xLinear",xLinear,6,zeroArray);
	creator->GetDoubleParamArray("yLinear",yLinear,6,zeroArray);
	creator->GetDoubleParamArray("zLinear",zLinear,6,zeroArray);
	creator->GetDoubleParamArray("pLinear",pLinear,6,zeroArray);
	creator->GetDoubleParamArray("qLinear",qLinear,6,zeroArray);
	creator->GetDoubleParamArray("rLinear",rLinear,6,zeroArray);

	creator->GetDoubleParamArray("xQuadratic",xQuadratic,6,zeroArray);
	creator->GetDoubleParamArray("yQuadratic",yQuadratic,6,zeroArray);
	creator->GetDoubleParamArray("zQuadratic",zQuadratic,6,zeroArray);
	creator->GetDoubleParamArray("pQuadratic",pQuadratic,6,zeroArray);
	creator->GetDoubleParamArray("qQuadratic",qQuadratic,6,zeroArray);
	creator->GetDoubleParamArray("rQuadratic",rQuadratic,6,zeroArray);

	creator->AddOutport	("LinearDamping"	,	6, PORT_FUNCTION(ViscousForce::LinearDamping));
	creator->AddOutport	("QuadraticDamping"	,	6, PORT_FUNCTION(ViscousForce::QuadraticDamping));
	creator->AddOutport ("ViscousDamping"	,	6, PORT_FUNCTION(ViscousForce::ViscousDamping));
	
	quadraticDampingMatrix.row(0) << xQuadratic[0], xQuadratic[1], xQuadratic[2], xQuadratic[3], xQuadratic[4], xQuadratic[5];
	quadraticDampingMatrix.row(1) << yQuadratic[0], yQuadratic[1], yQuadratic[2], yQuadratic[3], yQuadratic[4], yQuadratic[5];
	quadraticDampingMatrix.row(2) << zQuadratic[0], zQuadratic[1], zQuadratic[2], zQuadratic[3], zQuadratic[4], zQuadratic[5];
	quadraticDampingMatrix.row(3) << pQuadratic[0], pQuadratic[1], pQuadratic[2], pQuadratic[3], pQuadratic[4], pQuadratic[5];
	quadraticDampingMatrix.row(4) << qQuadratic[0], qQuadratic[1], qQuadratic[2], qQuadratic[3], qQuadratic[4], qQuadratic[5];
	quadraticDampingMatrix.row(5) << rQuadratic[0], rQuadratic[1], rQuadratic[2], rQuadratic[3], rQuadratic[4], rQuadratic[5];

	linearDampingMatrix.row(0) << xLinear[0], xLinear[1], xLinear[2], xLinear[3], xLinear[4], xLinear[5];
	linearDampingMatrix.row(1) << yLinear[0], yLinear[1], yLinear[2], yLinear[3], yLinear[4], yLinear[5];
	linearDampingMatrix.row(2) << zLinear[0], zLinear[1], zLinear[2], zLinear[3], zLinear[4], zLinear[5];
	linearDampingMatrix.row(3) << pLinear[0], pLinear[1], pLinear[2], pLinear[3], pLinear[4], pLinear[5];
	linearDampingMatrix.row(4) << qLinear[0], qLinear[1], qLinear[2], qLinear[3], qLinear[4], qLinear[5];
	linearDampingMatrix.row(5) << rLinear[0], rLinear[1], rLinear[2], rLinear[3], rLinear[4], rLinear[5];
}

void ViscousForce::FinalSetup( const double T, const double *const X, ISimObjectCreator* const creator )
{
	

}

const double* const ViscousForce::LinearDamping(const double T, const double* const X)
{
	const double *nuRelative;
	nuRelative = m_nuRelative->GetPortValue(T,X);

	nu << nuRelative[0],nuRelative[1],nuRelative[2],nuRelative[3],nuRelative[4],nuRelative[5];

	v_linearForces = linearDampingMatrix*nu;

	for (size_t i = 0;i<6;i++)
		linearForces[i] = v_linearForces[i];
	
	return linearForces;
}
const double* const ViscousForce::QuadraticDamping(const double T, const double* const X)
{
	const double *nuRelative;
	nuRelative = m_nuRelative->GetPortValue(T,X);

	nu << nuRelative[0],nuRelative[1],nuRelative[2],nuRelative[3],nuRelative[4],nuRelative[5];

	Matrix<float,6,6> quadraticAbsNu;
		
	for (size_t i =0;i<6;i++)
		for (size_t j=0;j<6;j++)
			quadraticAbsNu(i,j) = quadraticDampingMatrix(i,j)*abs(nu(i));
	
	v_quadraticForces = quadraticAbsNu*nu;

	for (size_t i = 0;i<6;i++)
		quadraticForces[i] = v_quadraticForces[i];
	
	return quadraticForces;

}
const double* const ViscousForce::ViscousDamping(const double T, const double* const X)
{
	const double* const linear = LinearDamping(T,X);
	const double* const quadratic = QuadraticDamping(T,X);

	for (size_t i = 1;i<6;i++)
		viscousForces[i] = linear[i] + quadratic[i];

	return viscousForces;
	
}
void ViscousForce::OdeFcn(const double T , const double* const X,	double* const XDot,	const bool IsMajorTimeStep)
{
}

//#ifdef FH_VISUALIZATION
//	/**
//	\brief Sets up visualization resources
//
//	Creates a spherical mesh and registers at visualization system
//		
//	\param[in]	ogreRoot	-> Register visualization resources	
//	\param[in]	creator		-> Retrieve parameters.
//	*/
//	void RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator)
//	{
//	}
//
//	/**
//	\brief Updates visualization resources to reflect current state
//		
//	Updates position of mesh
//
//	\param[in]	T	-> Current simulation time
//	\param[in]	X	-> Current simulation state
//	*/
//	void RenderUpdate(const double T, const double* const X){
//	}
//#endif