#ifndef VISCOUSFORCE_H
#define VISCOUSFORCE_H
#include <ISimObjectCreator.h>
#include <CFhSimMgr.h>
#include <string>
#include <list>
#include <iostream>
#include <locale>
#include "Eigen\Core"
#include "SimObject.h"
#include <ISimObjectCreator.h>
#include <string>

////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class							ViscousForces
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// \brief			Generates the linear and quadratic viscous forces
///
///					
/// \author			Mats Nvik Hval
///
///	\par
///					01.03.2012 MNH: Initial version
///
////////////////////////////////////////////////////////////////////////////////////////////////////

class ViscousForce : public SimObject
{
public:
	/**
	\brief	Reads parameters, registers states, input/output ports and shared resources

	The constructor performs all initial setup for viscousForces simobject.
	Reading in parameters, setting up communication interface i.e. output ports,
	input ports, and states, plus additional 'one time only' resource setup.

	Registers communication interface; a relative velocity input port, and "Forces" output port.
	
	\param[in]	simObjectName	->	The name of the simobject
	\param[in]	creator			->	Retrieve parameters. Register states, ports
									and shared resources
	*/
	ViscousForce(const string &simObjectName, ISimObjectCreator* const creator);
	
	void FinalSetup( const double T, const double *const X, ISimObjectCreator* const creator );
	void OdeFcn(const double T , const double* const X,	double* const XDot,	const bool IsMajorTimeStep); 

	const double* const LinearDamping(const double T, const double* const X);
	const double* const QuadraticDamping(const double T, const double* const X);
	const double* const ViscousDamping(const double T, const double* const X);

	//Eigen::Matrix3f RotationMatrix(Eigen::Vector3f inputVector);
	
	/**
	\brief Computes object derivatives as a function of time, states and input ports

	Returns state derivatives. Velocity as derivative of position, 
	(external_force+gravity)/mass as derivative of velocity

	\param[in]	T				-> Current simulation time
	\param[in]	X				-> Current simulation state
	\param[out]	XDot			-> State derivatives
	\param[in]	IsMajorTimeStep	-> Is this a major time step?
	*/

#ifdef FH_VISUALIZATION
	/**
	\brief Sets up visualization resources

	Creates a spherical mesh and registers at visualization system
		
	\param[in]	ogreRoot	-> Register visualization resources	
	\param[in]	creator		-> Retrieve parameters.
	*/
	void RenderInit(Ogre::Root* const ogreRoot, ISimObjectCreator* const creator){}

	/**
	\brief Updates visualization resources to reflect current state
		
	Updates position of mesh

	\param[in]	T	-> Current simulation time
	\param[in]	X	-> Current simulation state
	*/
	void RenderUpdate(const double T, const double* const X){}
#endif

	/**
	\brief Output port. Returns current position
	
	See PortDefs.h

	\param[in]	T	-> Current simulation time
	\param[in]	X	-> Current simulation state
	\return			-> Address of current position
	*/
	const double* const Position(const double T, const double* const X); 
	
	/**
	\brief Output port. Returns current velocity
	
	See PortDefs.h

	\param[in]	T	-> Current simulation time
	\param[in]	X	-> Current simulation state
	\return			-> Address of current velocity
	*/
	const double* const Velocity(const double T, const double* const X);

protected:
	Eigen::Matrix<float,6,6> quadraticDampingMatrix;	/// Rigid body mass matrix of the vessel
	Eigen::Matrix<float,6,6> linearDampingMatrix;		/// Rigid body coriolis matrix of the vessel

	
	Eigen::Matrix<float,6,1> v_linearForces;			/// vector with the linear forces
	Eigen::Matrix<float,6,1> v_quadraticForces;		/// vector with the quadratic forces
	Eigen::Matrix<float,6,1> v_viscousForces;			/// vector with the sum of linear and quadratic forces

	
	Eigen::Matrix<float,6,1> nu;						/// Rigid body velocity vector
	Eigen::Vector3f nu1;								/// First three elements of nu
	Eigen::Vector3f nu2;								/// Last three elements of nu
	
	ISignalPort* m_nuRelative;							/// Input port. Wave forces and moments acting on the vessel
		
	//int m_PositionIndex;								/// The index of the position state in the state array

	double xLinear[6];									/// the elements in the first row in the 6DOF linear damping matrix
	double yLinear[6];									/// the elements in the second row in the 6DOF linear damping matrix
	double zLinear[6];									/// the elements in the third row in the 6DOF linear damping matrix
	double pLinear[6];									/// the elements in the forth row in the 6DOF linear damping matrix
	double qLinear[6];									/// the elements in the fifth row in the 6DOF linear damping matrix
	double rLinear[6];									/// the elements in the sixth row in the 6DOF linear damping matrix
									
	double xQuadratic[6];								/// the elements in the first row in the 6DOF quadratic damping matrix
	double yQuadratic[6];								/// the elements in the second row in the 6DOF quadratic damping matrix
	double zQuadratic[6];								/// the elements in the third row in the 6DOF quadratic damping matrix
	double pQuadratic[6];								/// the elements in the forth row in the 6DOF quadratic damping matrix
	double qQuadratic[6];								/// the elements in the fifth row in the 6DOF quadratic damping matrix
	double rQuadratic[6];								/// the elements in the sixth row in the 6DOF quadratic damping matrix

	double linearForces[6];								/// vector with the linear forces
	double quadraticForces[6];							/// vector with the quadratic forces
	double viscousForces[6];							/// vector with the sum of linear and quadratic forces

	

};
#endif