#include "ViscousForces.h"
using namespace Eigen;
using namespace std;
#include "CNetCage.h"
#include "DeepSeaGravityWaves.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class							ViscousForces
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// \brief			Generates the linear and quadratic viscous forces
///
///					
/// \author			Mats Nvik Hval
///
///	\par
///					01.03.2012 MNH: Initial version
///
////////////////////////////////////////////////////////////////////////////////////////////////////



ViscousForces::ViscousForces(const double* const linearDampingExponentialCoefficients)
{
	initializeMembers();	
	linearDampingExponentialCoeff_k = linearDampingExponentialCoefficients[0];
	linearDampingExponentialCoeff_p = linearDampingExponentialCoefficients[1];
	linearDampingExponentialCoeff_r = linearDampingExponentialCoefficients[2];
	
}
ViscousForces::ViscousForces(){
	initializeMembers();
	linearDampingExponentialCoeff_k = 0.1;
	linearDampingExponentialCoeff_p = 15;
	linearDampingExponentialCoeff_r = 3;
}

void ViscousForces::initializeMembers(){
	quadraticDampingMatrix = Matrix<double,6,6>::Zero();
	linearDampingMatrix = Matrix<double,6,6>::Zero();
	nu = Matrix<double,6,1>::Zero();
}

void ViscousForces::setDiagonalDampingMatrix(const double* const linearDiagonal, const double* const quadraticDiagonal){

	for (size_t i = 0; i<6;i++){
		linearDampingMatrix(i,i) = linearDiagonal[i];
		quadraticDampingMatrix(i,i) = quadraticDiagonal[i];
	}
}


void ViscousForces::setQuadraticDampingMatrix(const double* const xQuadraticDamping,const double* const yQuadraticDamping,const double* const zQuadraticDamping,const double* const pQuadraticDamping,const double* const qQuadraticDamping,const double* const rQuadraticDamping){

	quadraticDampingMatrix.row(0) << xQuadraticDamping[0], xQuadraticDamping[1], xQuadraticDamping[2], xQuadraticDamping[3], xQuadraticDamping[4], xQuadraticDamping[5];
	quadraticDampingMatrix.row(1) << yQuadraticDamping[0], yQuadraticDamping[1], yQuadraticDamping[2], yQuadraticDamping[3], yQuadraticDamping[4], yQuadraticDamping[5];
	quadraticDampingMatrix.row(2) << zQuadraticDamping[0], zQuadraticDamping[1], zQuadraticDamping[2], zQuadraticDamping[3], zQuadraticDamping[4], zQuadraticDamping[5];
	quadraticDampingMatrix.row(3) << pQuadraticDamping[0], pQuadraticDamping[1], pQuadraticDamping[2], pQuadraticDamping[3], pQuadraticDamping[4], pQuadraticDamping[5];
	quadraticDampingMatrix.row(4) << qQuadraticDamping[0], qQuadraticDamping[1], qQuadraticDamping[2], qQuadraticDamping[3], qQuadraticDamping[4], qQuadraticDamping[5];
	quadraticDampingMatrix.row(5) << rQuadraticDamping[0], rQuadraticDamping[1], rQuadraticDamping[2], rQuadraticDamping[3], rQuadraticDamping[4], rQuadraticDamping[5];
}
void ViscousForces::setLinearDampingMatrix(const double* const xLinearDamping,const double* const yLinearDamping,const double* const zLinearDamping,const double* const pLinearDamping,const double* const qLinearDamping,const double* const rLinearDamping){
	linearDampingMatrix.row(0) << xLinearDamping[0], xLinearDamping[1], xLinearDamping[2], xLinearDamping[3], xLinearDamping[4], xLinearDamping[5];
	linearDampingMatrix.row(1) << yLinearDamping[0], yLinearDamping[1], yLinearDamping[2], yLinearDamping[3], yLinearDamping[4], yLinearDamping[5];
	linearDampingMatrix.row(2) << zLinearDamping[0], zLinearDamping[1], zLinearDamping[2], zLinearDamping[3], zLinearDamping[4], zLinearDamping[5];
	linearDampingMatrix.row(3) << pLinearDamping[0], pLinearDamping[1], pLinearDamping[2], pLinearDamping[3], pLinearDamping[4], pLinearDamping[5];
	linearDampingMatrix.row(4) << qLinearDamping[0], qLinearDamping[1], qLinearDamping[2], qLinearDamping[3], qLinearDamping[4], qLinearDamping[5];
	linearDampingMatrix.row(5) << rLinearDamping[0], rLinearDamping[1], rLinearDamping[2], rLinearDamping[3], rLinearDamping[4], rLinearDamping[5];
}

const double* const ViscousForces::LinearDamping(const double* const VesselVelocity)
{
	nu << VesselVelocity[0] ,VesselVelocity[1], VesselVelocity[2], VesselVelocity[3], VesselVelocity[4], VesselVelocity[5];

	Matrix<double,6,1> linearScaling = Matrix<double,6,1>::Zero();
	linearScaling(0) = 0.2+0.8*exp(-linearDampingExponentialCoeff_k* pow(linearDampingExponentialCoeff_p *abs(VesselVelocity[0]),linearDampingExponentialCoeff_r));
	linearScaling(1) = 0.2+0.8*exp(-linearDampingExponentialCoeff_k* pow(linearDampingExponentialCoeff_p *abs(VesselVelocity[1]),linearDampingExponentialCoeff_r));
	linearScaling(2) = 0.2+0.8*exp(-linearDampingExponentialCoeff_k* pow(linearDampingExponentialCoeff_p *abs(VesselVelocity[2]),linearDampingExponentialCoeff_r));
	linearScaling(3) = 1;
	linearScaling(4) = 1;
	linearScaling(5) = 1;
	
	v_linearForces = linearDampingMatrix*nu;

	
	for (size_t i = 0;i<6;i++)
		linearForces[i] = v_linearForces[i] * linearScaling(i);
	
	return linearForces;
}
const double* const ViscousForces::QuadraticDamping(const double* const VesselVelocity )
{
	nu << VesselVelocity[0],VesselVelocity[1],VesselVelocity[2],VesselVelocity[3],VesselVelocity[4],VesselVelocity[5];

	Matrix<double,6,6> quadraticAbsNu;
		
	for (size_t i =0;i<6;i++)
		for (size_t j=0;j<6;j++)
			quadraticAbsNu(i,j) = quadraticDampingMatrix(i,j)*abs(nu(i));
	
	v_quadraticForces = quadraticAbsNu*nu;

	for (size_t i = 0;i<6;i++)
		quadraticForces[i] = v_quadraticForces[i];
	
	return quadraticForces;

}
const double* const ViscousForces::getViscousForces(const double* const VesselVelocity)
{
	const double* const linear = LinearDamping(VesselVelocity);
	const double* const quadratic = QuadraticDamping(VesselVelocity);

	for (size_t i = 0;i<6;i++){
		viscousForces[i] = linear[i] + quadratic[i];
	}
	return viscousForces;
}