////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class							WamitReader
////////////////////////////////////////////////////////////////////////////////////////////////////
//
/// \brief			Reads the following output and input files from WAMIT:
///
///					*.1			Frequency dependent added mass and potential damping. 
///					*.2			Wave excitation forces from Haskind Relation.
///					*.3			Wave excitation forces from diffraction potential
///					*.8			Mean drift forces and moments from momentum
///					*.9			Mean drift forces and moments from pressure
///					*.out		Gravity, water density, waterdepth, volume, CoB, CoG, restoring coefficients
///
///					*.frc		Rigid body mass matrix
///					*.cfg		Information about the force file
///
///					
/// \author			Mats Nvik Hval
///
///	\par
///					20.02.2012 MNH: Initial version
///
////////////////////////////////////////////////////////////////////////////////////////////////////
#include "WamitReader.h"

using namespace std;
using namespace Eigen;

//////////////////////////////////////////////////////////////////////////////////////////////////////
////==================================================================================================
//// FUNCTION NAME:		TokenizeString
////==================================================================================================
////
/////	Takes a string and returns the elements/words in a string vector removing the white spaces
/////	
///// \param [in]		&line			string containing the line to tokenize
/////
///// \param [in]		&tokens			an empty vector for the tokens in "line".
/////
//////////////////////////////////////////////////////////////////////////////////////////////////////
//void WamitReader::TokenizeString(const string &line,vector<string> &tokens)
//{
//	tokens.clear();
//	string::size_type lastPos = line.find_first_not_of(" ",0);
//	string::size_type pos = line.find_first_of(" ",lastPos);
//
//	while (string::npos != pos || string::npos !=lastPos){
//		tokens.push_back(line.substr(lastPos,pos-lastPos));
//		lastPos = line.find_first_not_of(" ",pos);
//		pos = line.find_first_of(" ",lastPos);
//	}
//}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ReadFile
//==================================================================================================
//
///	Read the WAMIT output-file containing the added mass and potential damping. Converts the matrices from WAMIT-reference frame to NED-frame.
///	
/// \param [in]		wavePeriods			pointer to a vector containing the wave periods
///
/// \param [in]		addedMass			pointer to a vector containing the frequency dependent added mass matrices
///
/// \param [in]		potentialDamping	pointer to a vector containing the frequency dependent potential damping matrices
///
/// \param [in]		fileName			string containing the filename of the output file. i.e. "vessel" in "vessel.1"
///
/// \param [in]		ULEN				characteristic length of the vessel, found from the geometry file used in WAMIT.
///
/// \param [in]		rho					density of the fluid surrounding the vessel
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
void WamitReader::AddedMassAndDamping(vector<double> *wavePeriods,vector<Eigen::Matrix<float,6,6>, Eigen::aligned_allocator<Eigen::Matrix<float,6,6>>> *addedMass, vector<Eigen::Matrix<float,6,6>, Eigen::aligned_allocator<Eigen::Matrix<float,6,6>>> *potentialDamping, string fileName, const float ULEN, const float RHO)
	{
	wavePeriods->clear();
	addedMass->clear();
	potentialDamping->clear();
	Matrix<float,6,1> scaleDiag;
	scaleDiag << 1,-1,-1,1,-1,-1; 
	Matrix<float,6,6> WAMIT2NED;
	WAMIT2NED = scaleDiag.asDiagonal(); // transformation matrix for  converting from WAMIT reference frame to NED frame
	file.open(fileName+".1");
	int nPer = 0;
	int indexOfFirstWhiteSpace;
	string line;
	vector<double> vec_line; // vector containing the numbers read from current line
	string str_value;
	while ( !file.eof() ){
		getline(file,line);
		if (line.length() >5) {
			while ( !line.empty() && line.length() >1){ 
				while( line.at(0) == ' '){ line.erase(line.begin() ); }//remove white spaces in the beginning of the string.
				indexOfFirstWhiteSpace = line.find(' '); 
				str_value = line.substr(0,indexOfFirstWhiteSpace);
				line.erase(0,indexOfFirstWhiteSpace);
				vec_line.push_back(atof((const char*) str_value.c_str()));
			}
			if (vec_line[1] == vec_line[2] && int(vec_line[1]) == 1) {
					if (!addedMass->empty()){
						Matrix<float,6,6> transposedMatrix = addedMass->back().transpose();
						addedMass->back() = 0.5 * (addedMass->back() + transposedMatrix);
					}
					wavePeriods->push_back(vec_line[0]);
					Matrix<float,6,6> temp,temp2;
					temp.setZero();
					temp2.setZero();
					addedMass->push_back(temp);
					potentialDamping->push_back(temp2);
					nPer++;
				}
			if (vec_line[1] >3){
				if (vec_line[2] > 3){
					vec_line[3] *= pow(ULEN,5)*RHO;
					if (vec_line.size() == 5) vec_line[4] *= pow(ULEN,5)*RHO*vec_line[0];
				}
				else {
					vec_line[3] *= pow(ULEN,4)*RHO;
					if (vec_line.size() == 5) vec_line[4] *= pow(ULEN,4)*RHO*vec_line[0];
				}
			}
			else {
				if (vec_line[2] > 3){
					vec_line[3] *= pow(ULEN,4)*RHO;
					if (vec_line.size() == 5) vec_line[4] *= pow(ULEN,4)*RHO*vec_line[0];
				}
				else {
					vec_line[3] *= pow(ULEN,3)*RHO;
					if (vec_line.size() == 5) vec_line[4] *= pow(ULEN,3)*RHO*vec_line[0];
				}
			}
			addedMass->back()(int(vec_line[1])-1,int(vec_line[2])-1) = float(vec_line[3]);
			if (vec_line.size() == 5) potentialDamping->back()(int(vec_line[1])-1,int(vec_line[2])-1) = float(vec_line[4]);
			vec_line.clear();
		}
	}
	file.close();

	for (int i = 0;i<nPer;i++){
		Eigen::Matrix<float,6,6> temp = addedMass->at(i);
		addedMass->at(i) = WAMIT2NED*temp*WAMIT2NED; //transform added mass matrix to NED-frame
		temp = potentialDamping->at(i);
		potentialDamping->at(i)= WAMIT2NED*temp*WAMIT2NED; // transform potential damping matrix to NED frame
	}
}
////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ReadFile
//==================================================================================================
//
///	Read the WAMIT output-file containing the RAO data calculated by WAMIT. It is converted into NED-frame
///	
/// \param [in]		wavePeriods			pointer to a vector containing the wave periods
///
/// \param [in]		waveHeadings		pointer to a vector containing the wave headings
///
/// \param [in]		RAOdata				pointer to a vector containing a vector with frequency and wave heading dependent RAO data
///
/// \param [in]		fileName			string containing the filename of the output file. i.e. "vessel" in "vessel.2"
///
/// \param [in]		ULEN				characteristic length of the vessel, found from the geometry file used in WAMIT.
///
/// \param [in]		GRAVITY				gravity, found from the geometry file used in WAMIT.
///
/// \param [in]		RHO					density of the fluid surrounding the vessel
///
/// \param [in]		wamitOption			string indicating which WAMIT option the file is a result for
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////
void WamitReader::RAOdata(vector<double> *wavePeriods,vector<double> *waveHeadings,vector<vector<MatrixXf>> *RAOdata, string fileName, const float ULEN, const float RHO, const string wamitOption, const float GRAVITY){
	wavePeriods->clear();
	waveHeadings->clear();
	RAOdata->clear();
//	fileName = fileName.append("." + wamitOption);
	
	file.open(fileName+"."+wamitOption);
	int nPer = 0;
	int nHeading = 0;
	string line;
	vector<string> tokens;
	vector<double> floatTokens; // vector containing the values in current line
	
	while ( file.good() ){
		getline(file,line);
		TokenizeString(line,tokens);
		if( !tokens.empty()) {	for (size_t i=0;i<tokens.size();i++) floatTokens.push_back(float(atof(tokens.at(i).c_str())));
		if (!floatTokens.empty())
			if (atoi(wamitOption.c_str()) == 2 || atoi(wamitOption.c_str()) == 3){
				if (floatTokens[2] >3){
					floatTokens[3] *= pow(ULEN,3)*RHO*GRAVITY;
					}
				else 
					floatTokens[3] *= pow(ULEN,2)*RHO*GRAVITY;
			}
			if(atoi(wamitOption.c_str()) == 4){
				if (floatTokens[2] <= 3){
					floatTokens[3] /= ULEN;
				}
			}
			//cout << floatTokens[2]<< endl;
			if (find(wavePeriods->begin(),wavePeriods->end(),floatTokens[0]) == wavePeriods->end()){//nPer==0 || wavePeriods->back() != floatTokens[0]){
				wavePeriods->push_back(floatTokens[0]);
				nPer++;
			}
			if (find(waveHeadings->begin(),waveHeadings->end(),floatTokens[1]) == waveHeadings->end()) {
				vector<MatrixXf> raoMatrixVector;
				MatrixXf matrix;
				matrix = MatrixXf::Zero(6,4);
				raoMatrixVector.push_back(matrix);
				RAOdata->push_back(raoMatrixVector);
				nHeading++;
				waveHeadings->push_back(floatTokens[1]);

			}
			//RAOdata->at(nPer).at(nHeading)(int(floatTokens[2]),0) = float(floatTokens[3]);
			if(floatTokens[2] > 3){
				RAOdata->back().back()(int(floatTokens[2])-1,0) = float(floatTokens[3]);	//force amplitude
			}
			else 
				RAOdata->back().back()(int(floatTokens[2])-1,0) = float(floatTokens[3]);	//force amplitude

			if (floatTokens[2] == 3 ||	floatTokens[2] == 4 ||	floatTokens[2] == 5 )
				RAOdata->back().back()(int(floatTokens[2])-1,1) = float(floatTokens[4])+float(180.0);				//force phase
			else
				RAOdata->back().back()(int(floatTokens[2])-1,1) = float(floatTokens[4]);							//force phase
			
			if(int(floatTokens[2]>3)){
				RAOdata->back().back()(int(floatTokens[2])-1,2) = float(floatTokens[5])*RHO*pow(ULEN,3)*GRAVITY;	//force amplitude real part
				RAOdata->back().back()(int(floatTokens[2])-1,3) = float(floatTokens[6])*RHO*pow(ULEN,3)*GRAVITY;	//force amplitude real part
			}
			else {
				RAOdata->back().back()(int(floatTokens[2])-1,2) = float(floatTokens[5])*RHO*pow(ULEN,2)*GRAVITY;	//force amplitude imaginary part
				RAOdata->back().back()(int(floatTokens[2])-1,3) = float(floatTokens[6])*RHO*pow(ULEN,2)*GRAVITY;	//force amplitude imaginary part
			}

			//if (nHeading == 0 || waveHeadings->back() != floatTokens[1]){
			//	//cout << waveHeadings->size() << endl;
			//	waveHeadings->push_back(floatTokens[1]);
			//	//cout << waveHeadings->at(nHeading) << endl;
			//	nHeading++;
			//}
		}
		floatTokens.clear();

	}
	file.close();
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ReadFile
//==================================================================================================
//
///	Read the WAMIT output-file "fileName.out" and get the data for the input arguments
///	
/// \param [in]		fileName			string containing the filename of the output file. i.e. "vessel" in "vessel.2"  [string]
///
/// \param [in]		*ULEN				characteristic length of the vessel, defined from the geometry file used in WAMIT [float]
///
/// \param [in]		*GRAVIY				gravity, found from the geometry file used in WAMIT [float]
///
/// \param [in]		*RHO				density of the fluid surrounding the vessel [float]
///
/// \param [in]		*WATERDEPTH			waterdepth for the wamit-analyzis
///
/// \param [in]		XBODY[]				the position and heading of the body fixed 
///
/// \param [in]		VOLUMES[]			string indicating which WAMIT option the file is a result for
///
/// \param [in]		COB[]				string indicating which WAMIT option the file is a result for 
///
/// \param [in]		wamitCmatrix		string indicating which WAMIT option the file is a result for
///
/// \param [in]		COG[]				string indicating which WAMIT option the file is a result for
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////

void WamitReader::VesselParameters(string fileName, float *ULEN, float *RHO, float *GRAVITY, float *WATERDEPTH, float XBODY[], float *VOLUME, float COB[], Matrix<float,6,6> &wamitCmatrix, float COG[]){
	Matrix<float,6,1> scaleDiag;
	scaleDiag << 1,-1,-1,1,-1,-1; 
	const Matrix<float,6,6> WAMIT2NED = scaleDiag.asDiagonal(); // transformation matrix for  converting from WAMIT reference frame to NED frame
	float VOLUMES[3];
	wamitCmatrix.setZero(); // = MatrixXf::Zero(6,6);
	
	file.open(fileName.append(".out"));
	string line;
	vector<string> tokens;
	while ( file.good() ){
		line.clear();
		getline(file,line);
		
		if(line.find("Gravity:",0) != -1){
			TokenizeString(line,tokens);
			*GRAVITY = float(atof(tokens.at(1).c_str()));
			*ULEN = float(atof(tokens.at(4).c_str()));
		}
		if(line.find("Water depth",0) != -1){
			TokenizeString(line,tokens);
			*WATERDEPTH = float(atof(tokens.at(2).c_str()));
			*RHO = float(atof(tokens.at(5).c_str()));
		}
		if(line.find("XBODY",0) != -1){
			TokenizeString(line,tokens);
			XBODY[0] = float(atof(tokens.at(2).c_str()));
			XBODY[1] = -float(atof(tokens.at(5).c_str()));
			XBODY[2] = -float(atof(tokens.at(8).c_str()));
			XBODY[3] = -float(atof(tokens.at(11).c_str()));
		}
		if(line.find("Volumes",0) != -1){
			TokenizeString(line,tokens);
			VOLUMES[0] = float(atof(tokens.at(2).c_str()));
			VOLUMES[1] = float(atof(tokens.at(3).c_str()));
			VOLUMES[2] = float(atof(tokens.at(4).c_str()));
		}
		if(line.find("Center of Buoyancy",0) != -1){
			TokenizeString(line,tokens);
			COB[0] = float(atof(tokens.at(4).c_str()));
			COB[1] = -float(atof(tokens.at(5).c_str()));
			COB[2] = -float(atof(tokens.at(6).c_str()));
		}
		if(line.find("C(3,3)") != -1){
			TokenizeString(line,tokens);
			wamitCmatrix(2,2) = float(atof(tokens.at(1).c_str()))*(*RHO)*(*GRAVITY)*pow(*ULEN,2);
			wamitCmatrix(2,3) = float(atof(tokens.at(2).c_str()))*(*RHO)*(*GRAVITY)*pow(*ULEN,3);
			wamitCmatrix(2,4) = float(atof(tokens.at(3).c_str()))*(*RHO)*(*GRAVITY)*pow(*ULEN,3);
		}
		if(line.find("C(4,4)") != -1){
			TokenizeString(line,tokens);
			wamitCmatrix(3,3) = float(atof(tokens.at(1).c_str()))*(*RHO)*(*GRAVITY)*pow(*ULEN,4);
			wamitCmatrix(3,4) = float(atof(tokens.at(2).c_str()))*(*RHO)*(*GRAVITY)*pow(*ULEN,4);
			wamitCmatrix(3,5) = float(atof(tokens.at(3).c_str()))*(*RHO)*(*GRAVITY)*pow(*ULEN,4);
		}
		if(line.find("C(5,5)") != -1){
				TokenizeString(line,tokens);
			wamitCmatrix(4,4) = float(atof(tokens.at(1).c_str()))*(*RHO)*(*GRAVITY)*pow(*ULEN,4);
			wamitCmatrix(4,5) = float(atof(tokens.at(2).c_str()))*(*RHO)*(*GRAVITY)*pow(*ULEN,4);
		}
		if(line.find("Center of Gravity") != -1){
				TokenizeString(line,tokens);
			COG[0] = float(atof(tokens.at(4).c_str()));			
			COG[1] = -float(atof(tokens.at(5).c_str()));
			COG[2] = -float(atof(tokens.at(6).c_str()));
		}
		tokens.clear();
	}
	file.close();
	*VOLUME = (VOLUMES[0] + VOLUMES[1] + VOLUMES[2]) /3;
	wamitCmatrix = WAMIT2NED*wamitCmatrix*WAMIT2NED;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ReadFile
//==================================================================================================
//
///	Read the mean drift WAMIT output-file. The input arguments are given values. 
///	
///	Mean drift forces and moments can be found from meanDriftData by using PER,BETA1,BETA2 and DOF as
///	
///	meanDriftData.at(index of Period).at(index of BETA1).at(index of BETA2)(matrix index1,matrix index2)
///	
///	where the matrices have the following layout:
///---------
///	Mod(F1)		Pha(F1)		Re(F1)		Im(F1)
///	  .			  .			  .			  .
///	  .			  .			  .			  .
///	Mod(Fn)		Pha(Fn)		Re(Fn)		Im(Fn)
///---------
///	
/// \param [in]		fileName			string containing the filename of the output file. i.e. "vessel" in "vessel.2"  [string]
///
/// \param [in]		&secondWaveHeadings	vector containing the second waveheadings [float]
///
/// \param [in]		*meanDriftData		Matrix containing the mean wave drift data
///
/// \param [in]		ULEN				The characteristic length of the vessel [float]
///
/// \param [in]		RHO					Density of the fluid [float]
///
/// \param [in]		GRAVITY				the position and heading of the body fixed [float]
///
///	\param [in]		wamitOption			Mean drift loads from momentum (8) or pressure (9) [string]
/// 
////////////////////////////////////////////////////////////////////////////////////////////////////

void WamitReader::MeanDrift( string fileName ,vector<float> *secondWaveHeadings,vector<vector<vector<MatrixXf>>> *meanDriftData,const float ULEN, const float RHO,const float GRAVITY, const string wamitOption)
{
	
	file.open(fileName.append("."+wamitOption));
	int intWamitOption = atoi(wamitOption.c_str());
	int nHeading = 0,nHeading2 = 0,nPer = 0;
	string line;
	vector<float> periods,waveHeadings;
	vector<string> tokens;
	vector<float> floatTokens;
	while ( file.good() ){
		line.clear();
		tokens.clear();
		floatTokens.clear();
		MatrixXf meanDriftDataContent;
		getline(file,line);
		TokenizeString(line,tokens);
		if( !tokens.empty()) {
		for (size_t i=0;i<tokens.size();i++) floatTokens.push_back(float(atof(tokens.at(i).c_str())));
			if (int(floatTokens.at(3)) == 1)
			{
				if (find(secondWaveHeadings->begin(),secondWaveHeadings->end(),floatTokens[2]) == secondWaveHeadings->end())
				{
					secondWaveHeadings->push_back(floatTokens.at(2));
					nHeading2++;
				}
				if  (find(periods.begin(),periods.end(),floatTokens[2]) == periods.end()) //(periods.empty() || periods.back() != floatTokens.at(0)) 
				{
					periods.push_back(floatTokens.at(0));	
					nPer++;
				}
				if  (find(waveHeadings.begin(),waveHeadings.end(),floatTokens[2]) == waveHeadings.end()) //(waveHeadings.empty() || waveHeadings.back() != floatTokens.at(1)) 
				{
					waveHeadings.push_back(floatTokens.at(1));	
					nHeading++;
				}
				if (intWamitOption == 8)
				{
					meanDriftDataContent = MatrixXf::Zero(3,4);
				}
				if (intWamitOption == 9)
				{
					meanDriftDataContent = MatrixXf::Zero(9,4);
				}
				vector<MatrixXf> beta2;
				vector<vector<MatrixXf>> beta1;
				beta2.push_back(meanDriftDataContent);
				beta1.push_back(beta2);
				meanDriftData->push_back(beta1);
			}
			int dof = int(floatTokens.at(3));
			if (dof > 0 && dof <4){
				meanDriftData->back().back().back()(dof-1,0) = floatTokens.at(4)*GRAVITY*RHO*ULEN;
				meanDriftData->back().back().back()(dof-1,1) = floatTokens.at(5); 
				meanDriftData->back().back().back()(dof-1,2) = floatTokens.at(6)*GRAVITY*RHO*ULEN; 
				meanDriftData->back().back().back()(dof-1,3) = floatTokens.at(7)*GRAVITY*RHO*ULEN; 
			}
			if ( floatTokens.at(3) >3){	//surge and sway on the output file
				if (intWamitOption == 9){
					meanDriftData->back().back().back()(dof-1,0) = floatTokens.at(4)*GRAVITY*RHO*pow(ULEN,2); 
					meanDriftData->back().back().back()(dof-1,1) = floatTokens.at(5); 
					meanDriftData->back().back().back()(dof-1,2) = floatTokens.at(6)*GRAVITY*RHO*pow(ULEN,2); 
					meanDriftData->back().back().back()(dof-1,3) = floatTokens.at(7)*GRAVITY*RHO*pow(ULEN,2); 
				}
				else{ //yaw in the output file
					meanDriftData->back().back().back()(2,0) = floatTokens.at(4)*GRAVITY*RHO*pow(ULEN,2); 
					meanDriftData->back().back().back()(2,1) = floatTokens.at(5); 
					meanDriftData->back().back().back()(2,2) = floatTokens.at(6)*GRAVITY*RHO*pow(ULEN,2); 
					meanDriftData->back().back().back()(2,3) = floatTokens.at(7)*GRAVITY*RHO*pow(ULEN,2); 
				}
			}
		}
	}
	file.close();
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//==================================================================================================
// FUNCTION NAME:		ReadMassMatrix
//==================================================================================================
//
///	Reads the rigid body mass matrix or the matrix of the body radii of gyration from WAMIT input-file.  
///	If the radii of gyration matrix is given as input to WAMIT (Force Control File Alternative form 1), 
///	the rigid body mass matrix is calculated according to p4-4 in the WAMIT v6.4 Manual.
///	
/// In case of Force Control File Alternative Form 2, the rigid body mass matrix is read directly.
/// 
/// \param [in]		fileName					string containing the filename of the output file. i.e. "vessel" in "vessel.frc"  [string]
///
/// \param [in]		&rigidBodyMassMatrix		vector containing the second waveheadings [float]
///
////////////////////////////////////////////////////////////////////////////////////////////////////

void WamitReader::MassMatrix(const string fileName, const float RHO, const float VOLUME,const float COG[], const float COB[], Matrix<float,6,6> &rigidBodyMassMatrix)
	{

		float mass = VOLUME*RHO;
		Matrix<float,6,1> scaleDiag(6);
		scaleDiag << 1,-1,-1,1,-1,-1; 
		const Matrix<float,6,6> WAMIT2NED = scaleDiag.asDiagonal(); // transformation matrix for  converting from WAMIT reference frame to NED frame
		//cout << WAMIT2NED << endl;
		rigidBodyMassMatrix.setZero();// = MatrixXf::Zero(6,6);
		Matrix3f radiiOfGyration = Matrix3f::Zero(3,3);
		bool frcAlternativeTwo = false;
		bool yaw = false;
		string line;
		vector<string> tokens;
		int lineNumber=0;
		ifstream file(fileName + ".cfg");
		//vector<string> tokens;

		while(file.good()){
			getline(file,line);
			if (line.find("IALTFRC") != -1)
				for (size_t i = 0;i<line.length();i++)
					if (line[i] == '2') {
						frcAlternativeTwo = true; 
					}
		}
		file.close();
		file.open(fileName +".frc");
		int dof = 0;
		while (file.good())
		{
			getline(file,line);
			lineNumber++;
			if (frcAlternativeTwo && lineNumber > 4 && 6-dof > 0){
				TokenizeString(line,tokens);
				if (tokens.size() >= 6){
					rigidBodyMassMatrix(dof,0) = float(atof(tokens.at(0).c_str()));
					rigidBodyMassMatrix(dof,1) = float(atof(tokens.at(1).c_str()));
					rigidBodyMassMatrix(dof,2) = float(atof(tokens.at(2).c_str()));
					rigidBodyMassMatrix(dof,3) = float(atof(tokens.at(3).c_str()));
					rigidBodyMassMatrix(dof,4) = float(atof(tokens.at(4).c_str()));
					rigidBodyMassMatrix(dof,5) = float(atof(tokens.at(5).c_str()));
					dof++;
				}
			}
			if(!frcAlternativeTwo) {
				TokenizeString(line,tokens);
				if(lineNumber >2 && tokens.size() >=3 && tokens.size() < 6){
					radiiOfGyration(dof,0) = float(atof(tokens.at(0).c_str()));
					radiiOfGyration(dof,1) = float(atof(tokens.at(1).c_str()));
					radiiOfGyration(dof,2) = float(atof(tokens.at(2).c_str()));
					dof++;
				}
			}
		}
		file.close();
		if (!frcAlternativeTwo){
			COG = COB;
			rigidBodyMassMatrix(0,0) = mass;
			rigidBodyMassMatrix(1,1) = mass;
			rigidBodyMassMatrix(2,2) = mass;
		
			rigidBodyMassMatrix(0,4) = mass*COG[2];
			rigidBodyMassMatrix(0,5) = -mass*COG[1];
			rigidBodyMassMatrix(1,3) = -mass*COG[2];
			rigidBodyMassMatrix(1,5) = mass*COG[1];
			rigidBodyMassMatrix(2,3) = mass*COG[1];
			rigidBodyMassMatrix(2,4) = -mass*COG[0];
		
			rigidBodyMassMatrix(3,1) = rigidBodyMassMatrix(1,3);
			rigidBodyMassMatrix(3,2) = rigidBodyMassMatrix(2,3);
			rigidBodyMassMatrix(4,0) = rigidBodyMassMatrix(0,4);
			rigidBodyMassMatrix(4,2) = rigidBodyMassMatrix(2,4);
			rigidBodyMassMatrix(5,0) = rigidBodyMassMatrix(0,5);
			rigidBodyMassMatrix(5,1) = rigidBodyMassMatrix(1,5);

			rigidBodyMassMatrix(3,3) = mass*radiiOfGyration(0,0)*abs(radiiOfGyration(0,0));
			rigidBodyMassMatrix(3,4) = mass*radiiOfGyration(0,1)*abs(radiiOfGyration(0,1));
			rigidBodyMassMatrix(3,5) = mass*radiiOfGyration(0,2)*abs(radiiOfGyration(0,2));
			rigidBodyMassMatrix(4,3) = mass*radiiOfGyration(1,0)*abs(radiiOfGyration(1,0));
			rigidBodyMassMatrix(4,4) = mass*radiiOfGyration(1,1)*abs(radiiOfGyration(1,1));
			rigidBodyMassMatrix(4,5) = mass*radiiOfGyration(1,2)*abs(radiiOfGyration(1,2));
			rigidBodyMassMatrix(5,3) = mass*radiiOfGyration(2,0)*abs(radiiOfGyration(2,0));
			rigidBodyMassMatrix(5,4) = mass*radiiOfGyration(2,1)*abs(radiiOfGyration(2,1));
			rigidBodyMassMatrix(5,5) = mass*radiiOfGyration(2,2)*abs(radiiOfGyration(2,2));
		}
	
		// transform from WAMIT-axes to NED-frame.
		rigidBodyMassMatrix = WAMIT2NED*rigidBodyMassMatrix*WAMIT2NED;
	
		return;
}