#include "PotentialVessel.h"
#include "ViscousForces.h"
#include "SyntheticWave.h"
#ifndef WAMITVESSEL_H
#define WAMITVESSEL_H

/** TODO:  Edit this one
\class WamitVessel
\par User Documentation:

Simulation model of a vessel in 6 degrees of freedom

\f{eqnarray*}{
\dot{\boldsymbol{\eta}}	&=& \mathbf{J}_\Theta(\boldsymbol{\eta}\boldsymbol{\nu} \\
\mathbf{M}\dot{\boldsymbol{\nu}}	&=& \sum \vec{F}
\f}

, where
- \f$\boldsymbol{\eta}\f$ is the position and attitude of the vessel in the NED frame.
- \f$\mathbf{J}_\Theta\f$ is the 6 DOF transformation matrix from body frame to NED-frame.
- \f$\boldsymbol{\nu}\f$ is the translational and rotational velocity of the vessel in the body frame
- \f$M\f$ is the mass of the vessel.
- \f$\sum \vec{F}\f$ is the sum of the forces and moments acting on the vessel.


\par Parameters
<table >
<tr><td><b>Parameter<td>Default<td>Comment</b>
<tr><td>fileName<td>-<td>filename of the vessel corresponding to the wamit output files.
</table>

\par Input ports
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>Waveforces<td>6<td>Wave induced forces and moments.
<tr><td>ViscousForces<td>6<td>Viscous forces and moments due to body velocity in a fluid.
<tr><td>FluidMemoryForce<td>6<td>Fluid memory force according to Fossen.
<tr><td>UmbilicalForce<td>6<td>Forces and moments induced by the umbilical.
<tr><td>ThrusterForce<td>6<td>Forces and moments induced by the thrusters.
<tr><td>Current<td>2<td>Current velocity
</table>

\par Output ports
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>Position<td>3<td>The position.
<tr><td>Attitude<td>3<td>The attitude
<tr><td>Velocity<td>3<td>The translational velocity.
<tr><td>RotationalVelocity<td>3<td>The rotational velocity.
<tr><td>Acceleration<td>3<td>The translational acceleration.
<tr><td>RotationalAcceleration<td>3<td>The rotational acceleration.
</table>

\par States
<table>
<tr><td><b>Tag<td>Size<td>Comment</b>
<tr><td>Position<td>6<td>The position.
<tr><td>Velocity<td>6<td>The velocity.
<tr><td>Acceleration<td>6<td>The acceleration.
</table>

*/


/**
\brief	Vessel generated from WAMIT-results

\author	Mats Nvik Hval

Vessel generated from WAMIT-results including visualization


\par Revision history:
	20.02.2012 MNH: Initial version
*/
class WamitVessel : public PotentialVessel
{
public:
	EIGEN_MAKE_ALIGNED_OPERATOR_NEW
	/**
	\brief	Reads parameters, registers states, input/output ports and shared resources

	The constructor performs all initial setup for a vessel simobject.
	Reading in parameters, setting up communication interface i.e. output ports,
	input ports, and states, plus additional 'one time only' resource setup.

	Reads in "fileName" parameter and tests for validity. 
	Registers communication interface; a force input ports, and "Velocity"/"Position"/"Acceleration" output ports.
	Registers states; "Velocity"/"Position"/"Acceleration"

	\param[in]	simObjectName	->	The name of the simobject
	\param[in]	creator			->	Retrieve parameters. Register states, ports
									and shared resources
	*/
	WamitVessel(const string& simObjectName, ISimObjectCreator* const creator);
	~WamitVessel();
	/**
	\brief Computes object derivatives as a function of time, states and input ports

	Returns state derivatives. Velocity as derivative of position, 
	(external_force+gravity)/mass as derivative of velocity

	\param[in]	T				-> Current simulation time
	\param[in]	X				-> Current simulation state
	\param[out]	XDot			-> State derivatives
	\param[in]	IsMajorTimeStep	-> Is this a major time step?
	*/
	void OdeFcn(const double T , const double* const X,	double* const XDot,	const bool IsMajorTimeStep); 

protected:
	ViscousForces* viscousForce;	
	void getViscousForces(const double T,const double* const X);
	Eigen::Matrix<double,6,1> v_viscousForces;



private:
	void Read();
	

	
};
#endif