#include "WamitVessel.h"

using namespace Eigen;
using namespace std;

WamitVessel::WamitVessel(const string& simObjectName, ISimObjectCreator* const creator):SimObject(simObjectName)
{

	creator->GetStringParam("PotentialFileName",potentialFileName);
	if(potentialFileName.length() == 0)
		creator->ReportParameterError("PotentialFileName", "Must correspond to the output files from WAMIT.");
	
	creator->GetStringParam("PotentialFileLocation",potentialFileLocation);
	if(potentialFileLocation.length() == 0)
		creator->ReportParameterError("PotentialFileLocation", "The directory for the WAMIT files is incorrect.");

	double zero[]= {0,0,0};
	creator->GetDoubleParamArray("VesselMainDimensions",vesselMainDimensions,3,zero);
	//creator->GetDoubleParamArray("VesselMainDimensions",vesselMainDimensions,3,zero);

	
	creator->AddOutport	("Position"	,		 6, PORT_FUNCTION(WamitVessel::Position));
	creator->AddOutport	("Velocity"	,		 6, PORT_FUNCTION(WamitVessel::Velocity));
	
	creator->AddInport	("WaveForces"	,	 6, &m_WaveForces);
	creator->AddInport	("ViscousForces",	 6, &m_ViscousForces);
	creator->AddInport	("FluidMemoryForces",6, &m_FluidMemoryForces);
	creator->AddInport	("ThrusterForces",	 6, &m_ThrusterForces);

	//creator->AddInport  ("StatePosition",	6,&m_initialPosition);
	//creator->AddInport  ("StateVelocity",	6,&m_initialVelocity);
	//creator->AddInport	("ThrusterForces",	 6, &m_ThrusterForces);

	m_PositionIndex		= creator->AddState("Position",		6);
	m_VelocityIndex		= creator->AddState("Velocity",		6);
	m_AccelerationIndex = creator->AddState("Acceleration",	6);
	///test

}

void WamitVessel::FinalSetup( const double T, const double *const X, ISimObjectCreator* const creator )
{
	/*m_Environment = static_cast<Environment*>( creator->GetSharedResource("CSeaEnvironment") );

	if( m_Environment )
	{
		m_Environment->GetWaveAmplitudes	( WaveHeight );
		m_Environment->GetWaveFrequencies	( WaveFreq );
		m_Environment->GetWavePhaseAngles	( WavePhase );
		m_Environment->GetWaveDirections	( WaveDirection );
		m_Environment->GetWaveNumbers		( WaveNumber );
	}


	if( WaveFreq.size() == 0)
	{
		WaveHeight.push_back(0);
		WaveFreq.push_back(0);
		WavePhase.push_back(0);
		WaveDirection.push_back(0);
		WaveNumber.push_back(0);
	}
	*/
		
	/// read current velocity;


	
	/// read parameters of the vessel
	float ULEN,XBODY[3],WATERDEPTH;
	WamitReader wamitreader;
	Matrix<float,6,6> wamitCmatrix;
	
	wamitreader.VesselParameters(fileLocation+fileName,&ULEN,&RHO,&GRAVITY,&WATERDEPTH,XBODY,&VOLUME,CoB,wamitCmatrix,CoG);
	wamitreader.MassMatrix(fileLocation+fileName,RHO, VOLUME, CoG, CoB, rigidBodyMassMatrix);

	/// Calculate the skew symmetric matrix of the center of gravity
	Vector3f centerOfGravity;
	centerOfGravity << CoG[0], CoG[1], CoG[2];
	Smtrx(&smtrxOfCG,(double)CoG[0], (double)CoG[1], (double)CoG[2]);
		
	/// read the added mass matrix for the infinite frequency.
	vector<double> wavePeriods;
	
	std::vector<Eigen::MatrixXf> addedmass;
	std::vector<Eigen::MatrixXf> potentialdamping;
	
	wamitreader.AddedMassAndDamping(&wavePeriods,&addedmass,&potentialdamping,fileLocation+fileName,ULEN,RHO);
	//for (int i = 0; i< 21;i++)
	//	std::cout << wavePeriods[i] << endl;// << addedmass << endl << potentialdamping << endl << fileLocation+fileName<< endl << ULEN << endl<< RHO;
	//
	int indexOfInfFrequency;
	for (size_t i = 0 ; i< wavePeriods.size();i++){
		if (wavePeriods.at(i) == -1 || (wavePeriods.at(i) > 1.01 *-1 && wavePeriods.at(i) < 1.01*-1))
		{
			indexOfInfFrequency = i;
			infFrequencyAddedMass = addedmass.at(indexOfInfFrequency);
		}
	}
	if(wavePeriods[indexOfInfFrequency] != -1)
		creator->ReportParameterError("WAMIT", "WAMIT has to be run with infinity as one of the frequencies");
}

void WamitVessel::OdeFcn(const double T, const double* const X, double* const XDot, const bool IsMajorTimeStep){
	const double* waveForces = m_WaveForces->GetPortValue(T,X);
	const double* viscousForces = m_ViscousForces->GetPortValue(T,X);
	const double* fluidMemoryForces = m_FluidMemoryForces->GetPortValue(T,X);
	//const double* umbilicalForces = m_UmbilicalForce->GetPortValue(T,X);
	const double* thrusterForces = m_ThrusterForces->GetPortValue(T,X);

	VectorXf v_waveForces(*waveForces);
	VectorXf v_viscousForces(*viscousForces);
	VectorXf v_fluidMemoryForces(*fluidMemoryForces);
	//VectorXf v_umbilicalForce(*umbilicalForces);
	VectorXf v_thrusterForces(*thrusterForces);

	////////// Hydrostatic restoring forces //////////
	// second term in B is to prevent the vessel to lift of at the surface.
	W = rigidBodyMassMatrix(0,0)*GRAVITY;
	B = RHO*GRAVITY*VOLUME - RHO*GRAVITY*(X[m_PositionIndex+2]+CoB[2])*vesselMainDimensions[0]*vesselMainDimensions[1];

	double phi = X[m_PositionIndex+3];		//< roll angle
	double theta=X[m_PositionIndex+4];		//< pitch angle
	double psi = X[m_PositionIndex+5];		//< yaw angle

	v_restoringForce(0) =  (W-B) * sin(theta);
	v_restoringForce(1) = -(W-B) * cos(theta) * sin(phi);
	v_restoringForce(2) = -(W-B) * cos(theta) * cos(phi);
	v_restoringForce(3) = -(CoG[1]*W - CoB[1]*B) * cos(theta) * cos(phi) + (CoG[2]*W - CoB[2]*B) * cos(theta) * sin(phi);
	v_restoringForce(4) =  (CoG[2]*W - CoB[2]*B) * sin(theta)			 + (CoG[0]*W - CoB[0]*B) * cos(theta) * cos(phi);
	v_restoringForce(5) = -(CoG[0]*W - CoB[0]*B) * cos(theta) * sin(phi) - (CoG[1]*W - CoB[1]*B) * sin(theta);
		
	//////////       Coriolis and centripetal forces     //////////
	//rigid body part, C_RB
	rigidBodyCoriolisMatrix.setZero();// = MatrixXf::Zero(6,6);
	
	Matrix3f sMatrixNu1;
	Matrix3f sMatrixNu2;
	Vector3f nu1;
	nu1 << X[m_PositionIndex],   X[m_PositionIndex+1], X[m_PositionIndex+2];
	Vector3f nu2;
	nu1 << X[m_PositionIndex+3],   X[m_PositionIndex+4], X[m_PositionIndex+5];
	
	Smtrx(&sMatrixNu1,X[m_PositionIndex],   X[m_PositionIndex+1], X[m_PositionIndex+2]);
	Smtrx(&sMatrixNu2,X[m_PositionIndex+3], X[m_PositionIndex+4], X[m_PositionIndex+5]);

	nuRelative << nu1(0)-currentN, nu1(2)-currentE, nu1(2)-currentD, nu2(0), nu2(1), nu2(2); 

	rigidBodyCoriolisMatrix.topRightCorner(3,3) = -rigidBodyMassMatrix(0,0) * sMatrixNu1 - rigidBodyMassMatrix(0,0) * sMatrixNu2 * smtrxOfCG;
	
	rigidBodyCoriolisMatrix.bottomLeftCorner(3,3) = -rigidBodyMassMatrix(0,0) * sMatrixNu1 - rigidBodyMassMatrix(0,0) * sMatrixNu2 * smtrxOfCG;

	Matrix3f inertiaBouyancy = rigidBodyMassMatrix.bottomRightCorner(3,3) - rigidBodyMassMatrix(0,0)*(smtrxOfCG*smtrxOfCG);

	Vector3f inertiaBouyancyNu2 = inertiaBouyancy*nu2;
	Matrix3f skewMatrixInertiaBouyancyNu2;
	Smtrx(&skewMatrixInertiaBouyancyNu2,inertiaBouyancyNu2[0],inertiaBouyancyNu2[1],inertiaBouyancyNu2[2]);
	rigidBodyCoriolisMatrix.bottomRightCorner(3,3) = - skewMatrixInertiaBouyancyNu2;
	
	///added mass part, C_A
	addedmassCoriolisMatrix.setZero();//= MatrixXf::Zero();

	//Eigen::Matrix3f skewMatrix;

	//Smtrx(infFrequencyAddedMass.topLeftCorner(3,3)*nu1 + infFrequencyAddedMass.topRightCorner(3,3)*nu2);

	addedmassCoriolisMatrix.topRightCorner(3,3) = -Smtrx(infFrequencyAddedMass.topLeftCorner(3,3)*nu1 + infFrequencyAddedMass.topRightCorner(3,3)*nu2);
	addedmassCoriolisMatrix.bottomLeftCorner(3,3) = addedmassCoriolisMatrix.topRightCorner(3,3); 
	addedmassCoriolisMatrix.bottomRightCorner(3,3) = -Smtrx(infFrequencyAddedMass.bottomLeftCorner(3,3)*nu1 + infFrequencyAddedMass.bottomRightCorner(3,3)*nu2);
		
	//////////       Forces due to current and vessel rotational speed  A_c    //////////
	///varyingCurrentMatrix
	/// Calculate the A_c matrix as in Refsnes to compensate for varying current

	varyingCurrentMatrix.setZero();
	VectorXf nu_current = VectorXf::Zero(6);
	MatrixXf currentRot = MatrixXf::Zero(6,6);

	currentRot.topLeftCorner(3,3) = RotationMatrix(nu2).transpose();
	nu_current = currentRot*currentVelocity;
	
	varyingCurrentMatrix.topRightCorner(3,3) << 0, infFrequencyAddedMass(0,0)*nu_current(3), -infFrequencyAddedMass(0,0)*nu_current(2), -infFrequencyAddedMass(1,1)*nu_current(2),0,infFrequencyAddedMass(1,1)*nu_current(0),infFrequencyAddedMass(2,2)*nu_current(1), -infFrequencyAddedMass(2,2)*nu_current(0), 0;
	varyingCurrentMatrix.bottomRightCorner(3,3) << -infFrequencyAddedMass(3,1)*nu_current(2), 0, infFrequencyAddedMass(3,1)*nu_current(0), 0, infFrequencyAddedMass(4,0)*nu_current(2), -infFrequencyAddedMass(4,0)*nu_current(1),0,0,0;

	nuDot = rigidBodyMassMatrix.inverse()*(v_waveForces - v_viscousForces - v_fluidMemoryForces + v_thrusterForces + infFrequencyAddedMass*nuRelative - addedmassCoriolisMatrix*nuRelative - rigidBodyCoriolisMatrix*nu);

	for(int i = 0;i < 6;i++){
		XDot[m_PositionIndex +i] = X[m_VelocityIndex + i];
		XDot[m_VelocityIndex +i] = nuDot(i);
	}
}

const double* const WamitVessel::Position(const double T, const double* const X){
	return X + m_PositionIndex;
}


const double* const WamitVessel::Velocity(const double T, const double* const X){
	return X + m_VelocityIndex;
}

#ifdef FH_VISUALIZATION
void WamitVessel::RenderInit( Ogre::Root* const ogreRoot, ISimObjectCreator* const creator ){
	creator->LogDebug("Entering WamitVessel::RenderInit().\n");

	creator->GetStringParam("meshModel",meshModel);
	if (meshModel.length() <= 0)
		creator->ReportParameterError("meshModel","To use visualization a mesh model has to be given.");

	Ogre::SceneManager* sceneMgr = ogreRoot->getSceneManager("main");

	Ogre::ManualObject* manual = sceneMgr->createManualObject("Object" + m_SimObjectName);
	
	creator->LogDebug("Exiting WamitVessel::RenderInit().\n");
}

void WamitVessel::RenderUpdate(const double T, const double* const X){
	m_RenderNode->setPosition(Ogre::Vector3(X[m_PositionIndex + 0],X[m_PositionIndex + 1],X[m_PositionIndex + 2]));
}
#endif
